//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "boids.h"

#if !defined AVOID_RENDERING

#include "TransProg/surface.h"
#include "TransProg/r32Vector.h"

#include "TransGl/viewer.h"
#include "TransGl/glDraw.h"

#define COLOR_COUNT (3*3*3)

typedef struct _RenderingData
{
real32_t colors[COLOR_COUNT][4];
BoundingBox groundBbox;
GlShape groundShape;
BoundingBox rockBbox;
GlShape rockShape;
BoundingBox sharkBbox;
GlShape sharkShape;
BoundingBox fishBbox;
GlShape fishShape;
PtrVector viewers;
int viewerCount;
} RenderingData;

static RenderingData rd;

typedef struct _ThreadViewerData
{
real64_t xSeen,ySeen,zSeen;
PtrVector seen;
} ThreadViewerData;

typedef struct _MyViewer
{
Viewer v;
bool close;
intptr_t seenCount;
real64_t xSeen,ySeen,zSeen;
ThreadViewerData **tvd[2];
} MyViewer;

static
void
addViewer(ViewerCallback callback)
{
MyViewer *mv=Memory_alloc(1,MyViewer);
int w=600;
int h=600;
int sw=GlView_screenWidth()-w;
int sh=GlView_screenHeight()-h;
int x=sw-((rd.viewerCount*50)%sw);
int y=(rd.viewerCount*50)%sh;
Viewer_init(&mv->v,x,y,w,h,false,callback,(void *)0);
char buffer[0x100];
sprintf(buffer,"Viewer %d",++rd.viewerCount);
GlView_setTitle(Viewer_view(&mv->v),buffer);
SpatialBasis_yaw(Viewer_viewPoint(&mv->v),M_PI/6.0);
SpatialBasis_pitch(Viewer_viewPoint(&mv->v),M_PI/6.0);
SpatialBasis_translate(Viewer_viewPoint(&mv->v),-LIMIT_SIZE,0.0,0.0);
Perspective_setNearDistance(Viewer_perspective(&mv->v),0.01*LIMIT_SIZE);
Perspective_setFarDistance(Viewer_perspective(&mv->v),100*LIMIT_SIZE);
mv->close=false;
mv->seenCount=0;
mv->xSeen=0.0;
mv->ySeen=0.0;
mv->zSeen=0.0;
mv->tvd[0]=Memory_alloc(Platform_cpuCtxCount(),ThreadViewerData *);
mv->tvd[1]=Memory_alloc(Platform_cpuCtxCount(),ThreadViewerData *);
for(int i=0,i_end=Platform_cpuCtxCount();i<i_end;++i)
  {
  ThreadViewerData *tvd=Memory_alloc(1,ThreadViewerData);
  PtrVector_init(&tvd->seen);
  tvd->xSeen=0.0;
  tvd->ySeen=0.0;
  tvd->zSeen=0.0;
  mv->tvd[0][i]=tvd;
  tvd=Memory_alloc(1,ThreadViewerData);
  PtrVector_init(&tvd->seen);
  tvd->xSeen=0.0;
  tvd->ySeen=0.0;
  tvd->zSeen=0.0;
  mv->tvd[1][i]=tvd;
  }
PtrVector_push(&rd.viewers,mv);
}

static
void
viewerCB(Viewer *v)
{
MyViewer *mv=STRUCT_OF_MEMBER(v,MyViewer,v);
GlView *view=Viewer_view(v);
const GfxState *state=GlView_state(view);
real64_t d=mv->seenCount ? 1.0/mv->seenCount : 0.0;
Viewer_setRefDistanceFromPoint(v,d*mv->xSeen,d*mv->ySeen,d*mv->zSeen);
switch(GfxState_event(state))
  {
  case GFX_KEY_EVT:
    {
    if(GfxState_pressed(state))
      {
      if(!strcmp(GfxState_key(state),"v"))
        {
        addViewer(&viewerCB);
        }
      else if(!strcmp(GfxState_key(state),"F11"))
        {
        Perspective_setStereoEyeSeparation(Viewer_perspective(v),
                    Perspective_stereoEyeSeparation(Viewer_perspective(v))*
                    (GfxState_modifiers(state)&GFX_SHIFT_MDF ? 1.1 : 1.0/1.1));
        GlView_postRedraw(view);
        }
      else if(!strcmp(GfxState_key(state),"F12"))
        {
        Perspective_setStereoViewDistance(Viewer_perspective(v),
                    Perspective_stereoViewDistance(Viewer_perspective(v))*
                    (GfxState_modifiers(state)&GFX_SHIFT_MDF ? 1.1 : 1.0/1.1));
        GlView_postRedraw(view);
        }
      else if(!strcmp(GfxState_key(state),"\x1B"))
        {
        mv->close=true;
        }
      else if(!strcmp(GfxState_key(state),"q"))
        {
        simu.mustQuit=true;
        }
      }
    break;
    }
  case GFX_CLOSE_EVT:
    {
    mv->close=true;
    break;
    }
  case GFX_DRAW_EVT:
    {
    mv->xSeen=mv->ySeen=mv->zSeen=0.0;
    mv->seenCount=0;
    const real32_t *projMtx=Viewer_projectionMatrix(v);
    const real32_t *viewMtx=Viewer_viewMatrix(v);
    real32_t headLight[3]={0.0f,0.0f,0.0f};
    real32_t groundRgba[4]={0.6f,0.5f,0.1f,1.0f};
    real32_t ambient=0.2f;
    real32_t specular=0.75f;
    real32_t shininess=20.0f;
    const real32_t *fog=(const real32_t *)0;
    GlShape_draw(&rd.groundShape,
                 groundRgba,ambient,specular,shininess,
                 headLight,fog,0,projMtx,viewMtx);
    for(intptr_t i=0,i_end=PtrVector_size(&simu.rocks);i<i_end;++i)
      {
      const Rock *rock=(const Rock *)PtrVector_get(&simu.rocks,i);
      if(Perspective_boxVisibility(Viewer_perspective(v),Viewer_viewPoint(v),
                                   &rd.rockBbox,&rock->location,
                                   (SpatialTransform *)0))
        {
        ++mv->seenCount;
        real64_t xc=BoundingBox_centerX(&rd.rockBbox);
        real64_t yc=BoundingBox_centerY(&rd.rockBbox);
        real64_t zc=BoundingBox_centerZ(&rd.rockBbox);
        SpatialBasis_localToGlobalPosition(&rock->location,&xc,&yc,&zc);
        mv->xSeen+=xc;
        mv->ySeen+=yc;
        mv->zSeen+=zc;
        real32_t modelViewMtx[16];
        Gfx_computeModelViewMatrix(modelViewMtx,
                                   viewMtx,rock->location.matrix);
        real32_t rockRgba[4]={0.6f,0.3f,0.3f,1.0f};
        GlShape_draw(&rd.rockShape,
                     rockRgba,ambient,specular,shininess,
                     headLight,fog,0,projMtx,modelViewMtx);
        }
      }
    int fromState2=Scheduler_fromState2(&simu.sched);
    for(intptr_t i=0,i_end=PtrVector_size(&simu.sharks);i<i_end;++i)
      {
      const Shark *shark=(const Shark *)PtrVector_get(&simu.sharks,i);
#if defined ASYNC_BOIDS
      const SpatialBasis *location=&shark->location;
#else
      const SpatialBasis *location=&shark->location[fromState2];
#endif
      if(Perspective_boxVisibility(Viewer_perspective(v),Viewer_viewPoint(v),
                                   &rd.sharkBbox,location,
                                   (SpatialTransform *)0))
        {
        ++mv->seenCount;
        real64_t xc=BoundingBox_centerX(&rd.sharkBbox);
        real64_t yc=BoundingBox_centerY(&rd.sharkBbox);
        real64_t zc=BoundingBox_centerZ(&rd.sharkBbox);
        SpatialBasis_localToGlobalPosition(location,&xc,&yc,&zc);
        mv->xSeen+=xc;
        mv->ySeen+=yc;
        mv->zSeen+=zc;
        real32_t modelViewMtx[16];
        Gfx_computeModelViewMatrix(modelViewMtx,
                                   viewMtx,location->matrix);
#if defined SHOW_AFFINITY
        int sharkColor=Activity_cpuCtx(&shark->act);
#else
        int sharkColor=shark->school;
#endif
        GlShape_draw(&rd.sharkShape,
                     rd.colors[sharkColor%COLOR_COUNT],
                     ambient,specular,shininess,
                     headLight,fog,0,projMtx,modelViewMtx);
        }
      }
    for(int i=0,i_end=Platform_cpuCtxCount();i<i_end;++i)
      {
      ThreadViewerData *tvd=mv->tvd[fromState2][i];
      mv->xSeen+=tvd->xSeen;
      mv->ySeen+=tvd->ySeen;
      mv->zSeen+=tvd->zSeen;
      PtrVector *seen=&tvd->seen;
      intptr_t seenCount=PtrVector_size(seen);
      mv->seenCount+=seenCount;
      for(int j=0;j<seenCount;++j)
        {
        Fish *fish=(Fish *)PtrVector_get(seen,j);
#if defined ASYNC_BOIDS
        const SpatialBasis *location=&fish->location;
#else
        const SpatialBasis *location=&fish->location[fromState2];
#endif
        real32_t modelViewMtx[16];
        Gfx_computeModelViewMatrix(modelViewMtx,
                                   viewMtx,location->matrix);
#if defined SHOW_AFFINITY
        int fishColor=i;
#else
        int fishColor=fish->school;
#endif
        GlShape_draw(&rd.fishShape,
                     rd.colors[fishColor%COLOR_COUNT],
                     ambient,specular,shininess,
                     headLight,fog,0,projMtx,modelViewMtx);
        }
      }
    break;
    }
  default:
    {
    break;
    }
  }
}

void
Rendering_showFish(Fish *f)
{
int toState2=Scheduler_toState2(&simu.sched);
if(simu.needsRedraw[toState2])
  {
  int cpuCtx=Activity_cpuCtx(&f->act);
  for(intptr_t i=0,i_end=PtrVector_size(&rd.viewers);i<i_end;++i)
    {
    const MyViewer *mv=(const MyViewer *)
                       PtrVector_get(&rd.viewers,i);
    if(Viewer_view(&mv->v)&&
       GfxState_visible(Viewer_state(&mv->v))) // FIXME: race condition
      {
#if defined ASYNC_BOIDS
      const SpatialBasis *location=&f->location;
#else
      const SpatialBasis *location=&f->location[toState2];
#endif
      if(Perspective_boxVisibility(Viewer_perspective(&mv->v),
                                   Viewer_viewPoint(&mv->v),&rd.fishBbox,
                                   location,(SpatialTransform *)0))
        {
        real64_t xc=BoundingBox_centerX(&rd.fishBbox);
        real64_t yc=BoundingBox_centerY(&rd.fishBbox);
        real64_t zc=BoundingBox_centerZ(&rd.fishBbox);
        SpatialBasis_localToGlobalPosition(location,&xc,&yc,&zc);
        ThreadViewerData *tvd=mv->tvd[toState2][cpuCtx];
        tvd->xSeen+=xc;
        tvd->ySeen+=yc;
        tvd->zSeen+=zc;
        PtrVector_push(&tvd->seen,f);
        }
      }
    }
  }
}

void
Rendering_activate(int cpuCtx,
                   void *data)
{
(void)data;
bool needsGfxSync=false;
int toState2=Scheduler_toState2(&simu.sched);
if(simu.needsRedraw[toState2])
  {
  for(intptr_t i=PtrVector_size(&rd.viewers);i--;)
    {
    MyViewer *mv=(MyViewer *)PtrVector_get(&rd.viewers,i);
    ThreadViewerData *tvd=mv->tvd[toState2][cpuCtx];
    PtrVector_clear(&tvd->seen);
    tvd->xSeen=0.0;
    tvd->ySeen=0.0;
    tvd->zSeen=0.0;
    if(!cpuCtx&&mv->close)
      {
      PtrVector_fastErase(&rd.viewers,i); // FIXME: race condition
      if(!PtrVector_size(&rd.viewers))
        {
        simu.mustQuit=true;
        }
      Viewer_destroy(&mv->v);
      for(int c=0,c_end=Platform_cpuCtxCount();c<c_end;++c)
        {
        PtrVector_destroy(&mv->tvd[0][c]->seen);
        Memory_free(mv->tvd[0][c]);
        PtrVector_destroy(&mv->tvd[1][c]->seen);
        Memory_free(mv->tvd[1][c]);
        }
      Memory_free(mv->tvd[0]);
      Memory_free(mv->tvd[1]);
      Memory_free(mv);
      }
    }
  }
if(!cpuCtx&&simu.needsRedraw[Scheduler_fromState2(&simu.sched)])
  {
  for(intptr_t i=0,i_end=PtrVector_size(&rd.viewers);i<i_end;++i)
    {
    MyViewer *mv=(MyViewer *)PtrVector_get(&rd.viewers,i);
    GlView_postRedraw(Viewer_view(&mv->v));
    }
  GlView_processEvents(0);
  needsGfxSync=true;
  }
Scheduler_activate(cpuCtx,&simu.sched);
if(needsGfxSync)
  {
  GlView_sync();
  }
}

void
Rendering_init(void)
{
GlView_init();
GlDraw_init();
int indexCount,vertexCount;
int *indices;
real32_t *vertices;
real32_t *normals;
//-- Colors --
for(int blue=0;blue<3;++blue)
  {
  for(int green=0;green<3;++green)
    {
    for(int red=0;red<3;++red)
      {
      real32_t *c=rd.colors[blue+3*(green+3*red)];
      c[0]=0.2f+0.3f*red;
      c[1]=0.2f+0.3f*green;
      c[2]=0.8f-0.3f*blue;
      c[3]=1.0f;
      }
    }
  }
//-- Rocks --
Surface_allocSphere(&indexCount,&indices,&vertexCount,&vertices,
                    &normals,(real32_t **)0,true,0.5*ROCK_SIZE,32);
BoundingBox_init(&rd.rockBbox);
BoundingBox_addVertices32(&rd.rockBbox,vertices,vertexCount);
GlShape_init(&rd.rockShape,GLSHAPE_HAS_NORMALS,GL_TRIANGLES,
             indexCount,indices,vertexCount,vertices,normals,
             (const real32_t *)0,(const real32_t *)0);
Memory_free(indices); Memory_free(vertices); Memory_free(normals);
//-- Ground --
real64_t groundSize=LIMIT_SIZE*(sqrt((real64_t)simu.schoolCount)+2.0);
Surface_allocBox(&indexCount,&indices,&vertexCount,&vertices,
                 &normals,(real32_t **)0,true,groundSize,groundSize,0.01);
GlShape_init(&rd.groundShape,GLSHAPE_HAS_NORMALS,GL_TRIANGLES,
             indexCount,indices,vertexCount,vertices,normals,
             (const real32_t *)0,(const real32_t *)0);
//-- Sharks --
Surface_makeBox(indices,vertices,normals,(real32_t *)0,
                true,SHARK_SIZE,0.25*SHARK_SIZE,0.5*SHARK_SIZE);
BoundingBox_init(&rd.sharkBbox);
BoundingBox_addVertices32(&rd.sharkBbox,vertices,vertexCount);
GlShape_init(&rd.sharkShape,GLSHAPE_HAS_NORMALS,GL_TRIANGLES,
             indexCount,indices,vertexCount,vertices,normals,
             (const real32_t *)0,(const real32_t *)0);
//-- Fish --
Surface_makeBox(indices,vertices,normals,(real32_t *)0,
                true,FISH_SIZE,0.25*FISH_SIZE,0.5*FISH_SIZE);
BoundingBox_init(&rd.fishBbox);
BoundingBox_addVertices32(&rd.fishBbox,vertices,vertexCount);
GlShape_init(&rd.fishShape,GLSHAPE_HAS_NORMALS,GL_TRIANGLES,
             indexCount,indices,vertexCount,vertices,normals,
             (const real32_t *)0,(const real32_t *)0);
Memory_free(indices); Memory_free(vertices); Memory_free(normals);
//-- Viewers ---
PtrVector_init(&rd.viewers);
rd.viewerCount=0;
addViewer(&viewerCB);
}

#endif // !AVOID_RENDERING

//----------------------------------------------------------------------------
