//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "boids.h"

#include "TransProg/r64Vector.h"

Simulation simu;

Cell *
createCells(const SpatialGrid *grid)
{
intptr_t cellCount=SpatialGrid_cellCount(grid);
Cell *cells=Memory_alloc(cellCount,Cell);
for(intptr_t i=0;i<cellCount;++i)
  {
  SpinSynchro_init(&cells[i].sync);
  PtrVector_init(&cells[i].content);
  }
return cells;
}

void
deleteCells(Cell *cells,
             SpatialGrid *grid)
{
intptr_t cellCount=SpatialGrid_cellCount(grid);
for(intptr_t i=0;i<cellCount;++i)
  {
  SpinSynchro_destroy(&cells[i].sync);
  PtrVector_destroy(&cells[i].content);
  }
Memory_free(cells);
}

void
parseCommandLine(void)
{
Scheduler_init(&simu.sched);
simu.memo=10;
simu.schoolCount=Platform_cpuCtxCount();
simu.rockCount=0;
simu.fishCount=0;
simu.sharkCount=0;
simu.dt=MAX_DT;
simu.neighbourAffinity=true;
#if !defined AVOID_RENDERING
  simu.needsRedraw[0]=false;
  simu.needsRedraw[1]=false;
#endif
simu.mustQuit=false;
simu.numaMigration=10;
bool rockCountGiven=false;
bool fishCountGiven=false;
bool sharkCountGiven=false;
int prefetch;
for(int i=1,i_end=Platform_argc();i<i_end;++i)
  {
  if(!strcmp(Platform_argv(i),"-memo")&&(i+1<i_end))
    {
    ++i;
    sscanf(Platform_argv(i),"%d",&simu.memo);
    }
  else if(!strcmp(Platform_argv(i),"-school")&&(i+1<i_end))
    {
    ++i;
    sscanf(Platform_argv(i),"%d",&simu.schoolCount);
    }
  else if(!strcmp(Platform_argv(i),"-rock")&&(i+1<i_end))
    {
    ++i;
    sscanf(Platform_argv(i),"%d",&simu.rockCount);
    rockCountGiven=true;
    }
  else if(!strcmp(Platform_argv(i),"-fish")&&(i+1<i_end))
    {
    ++i;
    sscanf(Platform_argv(i),"%d",&simu.fishCount);
    fishCountGiven=true;
    }
  else if(!strcmp(Platform_argv(i),"-shark")&&(i+1<i_end))
    {
    ++i;
    sscanf(Platform_argv(i),"%d",&simu.sharkCount);
    sharkCountGiven=true;
    }
  else if(!strcmp(Platform_argv(i),"-m")&&(i+1<i_end))
    {
    sscanf(Platform_argv(++i),"%d",&simu.numaMigration);
    }
  else if(!strcmp(Platform_argv(i),"-p")&&(i+1<i_end)&&
          (sscanf(Platform_argv(i+1),"%d",&prefetch)==1))
    {
    Scheduler_setPrefetch(&simu.sched,prefetch);
    ++i;
    }
  else if(!strcmp(Platform_argv(i),"-ws"))
    {
    Scheduler_setWorkStealingPeriod(&simu.sched,0);
    }
  else if(!strcmp(Platform_argv(i),"-na"))
    {
    simu.neighbourAffinity=false;
    }
  else
    {
    Platform_msg("ignoring argument '%s'\n",Platform_argv(i));
    }
  }
if(!rockCountGiven)
  {
  simu.rockCount=10*simu.schoolCount;
  }
if(!fishCountGiven)
  {
  simu.fishCount=1000*simu.schoolCount;
  }
if(!sharkCountGiven)
  {
  simu.sharkCount=simu.schoolCount;
  }
R64Vector_init(&simu.schools);
PtrVector_init(&simu.rocks);
SpinSynchro_init(&simu.sharksSync);
PtrVector_init(&simu.sharks);
SpinSynchro_init(&simu.fishSync);
PtrVector_init(&simu.fish);
real64_t unit=2.0*ROCK_SIZE;
intptr_t limit=(intptr_t)ceil(LIMIT_SIZE/unit);
SpatialGrid_init(&simu.rockGrid,unit,20*limit,20*limit,limit);
simu.rockCells=createCells(&simu.rockGrid);
unit=2.0*FISH_RANGE_MIN;
limit=(intptr_t)ceil(LIMIT_SIZE/unit);
SpatialGrid_init(&simu.fineFishGrid,unit,20*limit,20*limit,limit);
#if defined ASYNC_BOIDS
  simu.fineFishCells=createCells(&simu.fineFishGrid);
#else
  simu.fineFishCells[0]=createCells(&simu.fineFishGrid);
  simu.fineFishCells[1]=createCells(&simu.fineFishGrid);
#endif
unit=SHARK_RANGE;
limit=(intptr_t)ceil(LIMIT_SIZE/unit);
SpatialGrid_init(&simu.coarseFishGrid,unit,20*limit,20*limit,limit);
#if defined ASYNC_BOIDS
  simu.coarseFishCells=createCells(&simu.coarseFishGrid);
#else
  simu.coarseFishCells[0]=createCells(&simu.coarseFishGrid);
  simu.coarseFishCells[1]=createCells(&simu.coarseFishGrid);
#endif
unit=SHARK_SIZE;
limit=(intptr_t)ceil(LIMIT_SIZE/unit);
SpatialGrid_init(&simu.sharkGrid,unit,20*limit,20*limit,limit);
#if defined ASYNC_BOIDS
  simu.sharkCells=createCells(&simu.sharkGrid);
#else
  simu.sharkCells[0]=createCells(&simu.sharkGrid);
  simu.sharkCells[1]=createCells(&simu.sharkGrid);
#endif
}

intptr_t cellSum;
intptr_t cellCount;

int
main(int argc,
     char **argv)
{
Platform_init(argc,argv);
Platform_msg("cpu/mem: %d/%d\n",
             Platform_cpuCtxCount(),Platform_memCtxCount());
parseCommandLine();
#if !defined AVOID_RENDERING
  Rendering_init();
#endif
Platform_msg("creating %d schools\n",simu.schoolCount);
int cpuCtx=Platform_cpuCtx();
for(int i=0,i_end=simu.schoolCount;i<i_end;++i)
  {
  real64_t schoolX,schoolY,schoolZ;
  real64_t radius=LIMIT_SIZE;
  for(;;)
    {
    SpatialBasis b;
    SpatialBasis_init(&b);
    real64_t groundSize=LIMIT_SIZE*sqrt((real64_t)simu.schoolCount);
    SpatialBasis_translate(&b,groundSize*(Random_real64(cpuCtx)-0.5),
                              groundSize*(Random_real64(cpuCtx)-0.5),
                              0.25*LIMIT_SIZE);
    SpatialBasis_getPosition(&b,&schoolX,&schoolY,&schoolZ);
    intptr_t c,schoolSize=R64Vector_size(&simu.schools);
    for(c=0;c<schoolSize;c+=3)
      {
      real64_t d=cart2DToDistance64(
                 R64Vector_get(&simu.schools,c)-schoolX,
                 R64Vector_get(&simu.schools,c+1)-schoolY);
      if(d<radius)
        {
        break;
        }
      }
    if(c<schoolSize)
      {
      radius*=0.95;
      continue;
      }
    R64Vector_push(&simu.schools,schoolX);
    R64Vector_push(&simu.schools,schoolY);
    R64Vector_push(&simu.schools,schoolZ);
    SpatialBasis_destroy(&b);
    break;
    }
  }
Platform_msg("creating %d rocks\n",simu.rockCount);
for(int i=0,i_end=simu.rockCount;i<i_end;++i)
  {
  Rock_new(i%simu.schoolCount);
  }
Platform_msg("creating %d %s sharks\n",
             simu.sharkCount,
#if defined ASYNC_BOIDS
             "asynchronous");
#else
             "synchronous");
#endif
for(int i=0,i_end=simu.sharkCount;i<i_end;++i)
  {
  Shark_new(i%simu.schoolCount);
  }
Platform_msg("creating %d %s fish\n",
             simu.fishCount,
#if defined ASYNC_BOIDS
             "asynchronous");
#else
             "synchronous");
#endif
for(int i=0,i_end=simu.fishCount;i<i_end;++i)
  {
  Fish_new(i%simu.schoolCount);
  }
Platform_msg("scheduler prefetch: %"FPTR"d\n",
             Scheduler_prefetch(&simu.sched));
Platform_msg("work stealing: %d\n",Scheduler_workStealingPeriod(&simu.sched));
Platform_msg("neighbour affinity: %d\n",simu.neighbourAffinity);
if(Platform_numaCtxCount()<2)
  {
  simu.numaMigration=0;
  }
Platform_msg("numa migration: %d\n",simu.numaMigration);
int64_t now=Platform_time();
int64_t lastStep=now;
int64_t lastPerf=now;
int perfCount=0;
#if !defined AVOID_RENDERING
  int64_t lastDraw=now;
#endif
while(!simu.mustQuit)
  {
  ++perfCount;
  now=Platform_time();
  int64_t dt=now-lastPerf;
  if(dt>=2000000)
    {
    Platform_msg("%g\n",perfCount/(1e-6*dt));
    lastPerf+=dt;
    perfCount=0;
    }
  Scheduler_advanceStep(&simu.sched);
  dt=now-lastStep;
  lastStep+=dt;
  simu.dt=dt*1e-6;
  if(simu.dt>MAX_DT) simu.dt=MAX_DT;
#if defined AVOID_RENDERING
  Platform_parallel(&Scheduler_activate,&simu.sched);
#else
  dt=now-lastDraw;
  int toState2=Scheduler_toState2(&simu.sched);
  simu.needsRedraw[toState2]=(dt>=40000) ?  true : false;
  if(simu.needsRedraw[toState2])
    {
    lastDraw+=dt;
    }
  Platform_parallel(&Rendering_activate,(void *)0);
#endif
  }
Scheduler_destroy(&simu.sched);
R64Vector_destroy(&simu.schools);
PtrVector_destroy(&simu.rocks);
SpinSynchro_destroy(&simu.sharksSync);
PtrVector_destroy(&simu.sharks);
SpinSynchro_destroy(&simu.fishSync);
PtrVector_destroy(&simu.fish);
deleteCells(simu.rockCells,&simu.rockGrid);
#if defined ASYNC_BOIDS
  deleteCells(simu.fineFishCells,&simu.fineFishGrid);
  deleteCells(simu.coarseFishCells,&simu.coarseFishGrid);
  deleteCells(simu.sharkCells,&simu.sharkGrid);
#else
  deleteCells(simu.fineFishCells[0],&simu.fineFishGrid);
  deleteCells(simu.fineFishCells[1],&simu.fineFishGrid);
  deleteCells(simu.coarseFishCells[0],&simu.coarseFishGrid);
  deleteCells(simu.coarseFishCells[1],&simu.coarseFishGrid);
  deleteCells(simu.sharkCells[0],&simu.sharkGrid);
  deleteCells(simu.sharkCells[1],&simu.sharkGrid);
#endif
SpatialGrid_destroy(&simu.rockGrid);
SpatialGrid_destroy(&simu.fineFishGrid);
SpatialGrid_destroy(&simu.coarseFishGrid);
SpatialGrid_destroy(&simu.sharkGrid);
return 0;
}

DECLARE_WINDOWS_MAIN_FUNCTION

//----------------------------------------------------------------------------
