//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef MICROBIO_H
#define MICROBIO_H 1

#include "TransProg/scheduler.h"
#include "TransCuda/cuda.h"
#include "TransProg/spatialUtils.h"
#include "TransProg/coordIndex.h"
#include "TransProg/dynamicCompiler.h"

// #define AVOID_RENDERING
#define BACTERIUM_COUNT 100000
#define DETECT_NEIGHBOURS
#define SPATIAL_AFFINITY
#define SUBSTRATE_AFFINITY

// jz=-D*dC/dz
//   j: density vector (mol/(s*m*m))
//   D: diffusion (m*m/s)
//   C: concentration (mol/(m*m*m))

// Substrate:
//   glucose: 20g/l
//   glucose diffusion: 6.7e-10 m2/s
//   lactic acid diffusion: 2.35e-10 m2/s

// Carnobacterium piscicola:
//   diameter: 0.5--0.6 e-6
//   length: 1.1--1.4 e-6
//   division: split length, lose 1% (takes 42m in optimal conditions)
//   consumes glucose, produces lactic acid

// Listeria monocytogene:
//   diameter: 0.4--0.5 e-6
//   length: 0.5--2.0 e-6
//   division: split length (takes 33m in optimal conditions)

#define BACTERIUM_SIZE (0.2e-6f)
#define BACTERIUM_RENDER_SIZE (100.0f*BACTERIUM_SIZE)

#define SUBSTRATE_PRECISION_SHIFT 10
#define SUBSTRATE_PRECISION       (1<<SUBSTRATE_PRECISION_SHIFT)
#define SUBSTRATE_PRECISION_MASK  (SUBSTRATE_PRECISION-1)

#define SUBSTRATE_WIDTH_SHIFT     SUBSTRATE_PRECISION_SHIFT
#define SUBSTRATE_WIDTH           SUBSTRATE_PRECISION
#define SUBSTRATE_WIDTH_MASK      SUBSTRATE_PRECISION_MASK
                               
#define SUBSTRATE_HEIGHT_SHIFT    SUBSTRATE_PRECISION_SHIFT
#define SUBSTRATE_HEIGHT          SUBSTRATE_PRECISION
#define SUBSTRATE_HEIGHT_MASK     SUBSTRATE_PRECISION_MASK

#define SUBSTRATE_SIZE      (0.01f)
#define SUBSTRATE_DIFF_A    (1.0e-6f)
#define SUBSTRATE_DIFF_B    (0.5f*SUBSTRATE_DIFF_A)
#define SUBSTRATE_STEP      (SUBSTRATE_SIZE/SUBSTRATE_PRECISION)
#define DT (0.2f*SUBSTRATE_STEP*SUBSTRATE_STEP/SUBSTRATE_DIFF_A)

#define GRID_PRECISION       1024
#define GRID_WIDTH           GRID_PRECISION
#define GRID_HEIGHT          GRID_PRECISION
#define GRID_STEP            (SUBSTRATE_SIZE/GRID_PRECISION)

#define GPU_KERNEL_NAME "microbio"
#define GPU_SOURCE_FILE GPU_KERNEL_NAME".gpu"

typedef struct _Bacterium
{
Activity act;
SpinSynchro sync;
bool dead;
real64_t x,y,z;
real64_t speed,direction;
int id[2];
real32_t dA[2];
real32_t dB[2];
#if defined DETECT_NEIGHBOURS
  intptr_t key;
#endif
} Bacterium;

//----------------------------------------------------------------------------

#if defined DETECT_NEIGHBOURS

typedef struct _Cell
{
SpinSynchro sync;
PtrVector content;
} Cell;

INLINE
void
Cell_add(Cell *cell,
         void *element)
{
SpinSynchro_lockW(&cell->sync);
PtrVector_push(&cell->content,element);
SpinSynchro_unlockW(&cell->sync);
}

INLINE
void
Cell_remove(Cell *cell,
            void *element)
{
SpinSynchro_lockW(&cell->sync);
PtrVector_fastErase(&cell->content,
                    PtrVector_find(&cell->content,
                                   (PtrVector_CmpFnct)0,
                                   element));
SpinSynchro_unlockW(&cell->sync);
}

#endif

//----------------------------------------------------------------------------

typedef struct _Worker
{
bool needsSend;
int64_t duration[3];
real32_t usage;
const CudaDevice *dev;
CudaStream *strm;
CudaMarker *markers[4];
CudaProgram *prg;
int blockDim;
int gridDim[2];
int idOffset[2];
int computeDim[2];
#if !defined AVOID_RENDERING
  CudaBuffer *rendering[2];
#endif
CudaBuffer *A[2];
CudaBuffer *B[2];
#if defined BACTERIUM_COUNT
  CudaBuffer *dA[2];
  CudaBuffer *dB[2];
  CudaBuffer *gxA[2];
  CudaBuffer *gxB[2];
  CudaBuffer *gyA[2];
  CudaBuffer *gyB[2];
#endif
} Worker;

typedef struct _AppliData
{
Scheduler sched;
#if defined DETECT_NEIGHBOURS
  SpatialGrid grid;
  Cell *cells;
#endif
bool mustQuit;
bool needsReset;
int totalDim;
int workerCount;
Worker *workers;
Worker *cpuWorker;
PluginHandle plugin;
void
(*cpuProgram)(
#if !defined AVOID_RENDERING
              int * RESTRICT rendering,
#endif
              const real32_t * RESTRICT A0,
#if defined BACTERIUM_COUNT
              const real32_t * RESTRICT dA,
              real32_t * RESTRICT gxA,
              real32_t * RESTRICT gyA,
#endif
              real32_t * RESTRICT A1,
              const real32_t * RESTRICT B0,
#if defined BACTERIUM_COUNT
              const real32_t * RESTRICT dB,
              real32_t * RESTRICT gxB,
              real32_t * RESTRICT gyB,
#endif
              real32_t * RESTRICT B1,
              int id);
#if !defined AVOID_RENDERING
  bool needsRedraw[2];
  int64_t renderingDuration;
  CudaLockedMem *renderingMem[2];
  int *rendering[2];
#endif
CudaLockedMem *Amem[2];
real32_t *A[2];
CudaLockedMem *Bmem[2];
real32_t *B[2];
#if defined BACTERIUM_COUNT
  CudaLockedMem *dAmem[2];
  real32_t *dA[2];
  CudaLockedMem *dBmem[2];
  real32_t *dB[2];
  CudaLockedMem *gxAmem[2];
  real32_t *gxA[2];
  CudaLockedMem *gxBmem[2];
  real32_t *gxB[2];
  CudaLockedMem *gyAmem[2];
  real32_t *gyA[2];
  CudaLockedMem *gyBmem[2];
  real32_t *gyB[2];
  int *ctx;
#endif
} AppliData;

extern AppliData ad;

void
AppliData_activate(int cpuCtx,
                   void *data);

#if !defined AVOID_RENDERING

void
Rendering_init(void);

int *
Rendering_buffers(void);

void
Rendering_destroyBuffers(void);

void
Rendering_activate(int cpuCtx,
                   void *data);

void
Rendering_showBacterium(Bacterium *b);

#endif

#endif // MICROBIO_H

//----------------------------------------------------------------------------
