//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransCuda/cuda.h"

int
main(int argc,
     char **argv)
{
Platform_init(argc,argv);
Cuda_init();
int i,j,k,deviceCount;
const CudaDevice *devices=Cuda_devices(&deviceCount);
for(i=0;i<deviceCount;++i)
  {
  const CudaDevice *d=devices+i;
  const CudaDeviceProperties *prop=CudaDevice_properties(d);
#define SHOW_PROPERTY(pname) \
        Platform_msg("  "#pname": %d\n",prop->pname)
  Platform_msg("CUDA device %d:\n",CudaDevice_id(d));
  Platform_msg("  name: %s\n",CudaDevice_name(d));
  Platform_msg("  compute_capability: %d.%d\n",
               prop->compute_capability_major,
               prop->compute_capability_minor);
  Platform_msg("  total_memory: %"F64"d\n",
               prop->total_memory);
  Platform_msg("  free_memory: %"F64"d\n",
               CudaDevice_freeMemory(d));
  SHOW_PROPERTY(max_threads_per_block);
  Platform_msg("  max_block_dim: %d %d %d\n",
               prop->max_block_dim_x,
               prop->max_block_dim_y,
               prop->max_block_dim_z);
  Platform_msg("  max_grid_dim: %d %d %d\n",
               prop->max_grid_dim_x,
               prop->max_grid_dim_y,
               prop->max_grid_dim_z);
  SHOW_PROPERTY(max_shared_memory_per_block);
  SHOW_PROPERTY(total_constant_memory);
  SHOW_PROPERTY(warp_size);
  SHOW_PROPERTY(max_registers_per_block);
  SHOW_PROPERTY(clock_rate_kHz);
  SHOW_PROPERTY(multiprocessor_count);
  SHOW_PROPERTY(cores_per_multiprocessor);
  SHOW_PROPERTY(core_count);
  SHOW_PROPERTY(kernel_exec_timeout);
  SHOW_PROPERTY(integrated);
  SHOW_PROPERTY(can_map_host_memory);
  SHOW_PROPERTY(compute_mode);
  SHOW_PROPERTY(concurrent_kernels);
  SHOW_PROPERTY(ecc_enabled);
  SHOW_PROPERTY(pci_bus_id);
  SHOW_PROPERTY(pci_device_id);
  SHOW_PROPERTY(tcc_driver);
  SHOW_PROPERTY(memory_clock_rate_kHz);
  SHOW_PROPERTY(global_memory_bus_width);
  SHOW_PROPERTY(l2_cache_size);
  SHOW_PROPERTY(max_threads_per_multiprocessor);
  SHOW_PROPERTY(async_engine_count);
  SHOW_PROPERTY(unified_addressing);
  SHOW_PROPERTY(pci_domain_id);
  SHOW_PROPERTY(stream_priorities_supported);
  SHOW_PROPERTY(global_l1_cache_supported);
  SHOW_PROPERTY(local_l1_cache_supported);
  SHOW_PROPERTY(max_shared_memory_per_multiprocessor);
  SHOW_PROPERTY(max_registers_per_multiprocessor);
  SHOW_PROPERTY(managed_memory);
  SHOW_PROPERTY(multi_gpu_board);
  SHOW_PROPERTY(multi_gpu_board_group_id);
  intptr_t bufferSize=-1;
  if(Platform_argc()>1)
    {
    sscanf(Platform_argv(1),"%"FPTR"d",&bufferSize);
    }
  if(bufferSize<0)
    {
    int extra=0x100;
    bufferSize=((intptr_t)prop->max_grid_dim_x)*
               ((intptr_t)prop->max_block_dim_x)*sizeof(int);
    if(6*(bufferSize+extra)>CudaDevice_freeMemory(d))
      {
      bufferSize=(intptr_t)((CudaDevice_freeMemory(d)/6)-extra);
      }
    }
  else
    {
    bufferSize=bufferSize*1024*1024;
    }
  if(bufferSize)
    {
    Platform_msg("  program:\n");
    const char * sourceCode=
      "\n"
      "extern \"C\" __global__\n"
      "void\n"
      "vectorAdd(int * __restrict__ out,\n"
      "          const int * __restrict__ in1,\n"
      "          const int * __restrict__ in2,\n"
      "          unsigned int arraySize)\n"
      "{\n"
      "unsigned int id=blockDim.x*blockIdx.x+threadIdx.x;\n"
      "if(id<arraySize)\n"
      "  {\n"
      "  out[id]=in1[id]+in2[id];\n"
      "  }\n"
      "}\n"
      "\n";
#if 0
    Platform_msg("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
                 "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    Platform_msg("%s",sourceCode);
    Platform_msg("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
                 "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
#endif
    CudaProgram p;
    CudaProgram_init(&p,d,"vectorAdd","--use_fast_math",sourceCode,true);
    if(CudaProgram_buildLog(&p))
      {
      Platform_msg("    build log: %s\n",
                   CudaProgram_buildLog(&p));
      }
    if(!CudaProgram_buildFailure(&p))
      {
      Platform_msg("    max_threads_per_block: %d\n",
                   CudaProgram_maxThreadsPerBlock(&p));
      Platform_msg("    shared size bytes: %d\n",
                   CudaProgram_sharedSizeBytes(&p));
      Platform_msg("    const_size_bytes: %d\n",
                   CudaProgram_constSizeBytes(&p));
      Platform_msg("    local_size_bytes: %d\n",
                   CudaProgram_localSizeBytes(&p));
      Platform_msg("    num_regs: %d\n",
                   CudaProgram_numRegs(&p));
      Platform_msg("    ptx_version: %d\n",
                   CudaProgram_ptxVersion(&p));
      Platform_msg("    binary_version: %d\n",
                   CudaProgram_binaryVersion(&p));
      Platform_msg("  execution:\n");
      Platform_msg("    buffer_size: %"FPTR"d (%"FPTR"d MB)\n",
                   bufferSize,bufferSize/1024/1024);
      int arraySize=(int)(bufferSize/sizeof(int));
      int blockDim=CudaProgram_maxThreadsPerBlock(&p);
      Platform_msg("    block_dim: %d\n",blockDim);
      int gridDim=0;
      while(gridDim*blockDim<arraySize)
        {
        ++gridDim;
        }
      Platform_msg("    grid_dim: %d\n",gridDim);
      CudaStream s;
      CudaStream_init(&s,d);
      CudaMarker m0,m1,m2,m3;
      CudaMarker_init(&m0,d);
      CudaMarker_init(&m1,d);
      CudaMarker_init(&m2,d);
      CudaMarker_init(&m3,d);
      CudaBuffer outBuf,in1Buf,in2Buf;
      CudaBuffer_init(&outBuf,d,bufferSize);
      CudaBuffer_init(&in1Buf,d,bufferSize);
      CudaBuffer_init(&in2Buf,d,bufferSize);
      CudaLockedMem outMem,in1Mem,in2Mem;
      CudaLockedMem_init(&outMem,false,bufferSize);
      CudaLockedMem_init(&in1Mem,true,bufferSize);
      CudaLockedMem_init(&in2Mem,true,bufferSize);
      int *out=(int *)CudaLockedMem_data(&outMem);
      int *in1=(int *)CudaLockedMem_data(&in1Mem);
      int *in2=(int *)CudaLockedMem_data(&in2Mem);
      const void *args[]={ CudaBuffer_programArg(&outBuf),
                           CudaBuffer_programArg(&in1Buf),
                           CudaBuffer_programArg(&in2Buf),
                           &arraySize };
      int data1[17]={37,50,54,50,56,15,51,32,12,20,43,56,99,50,25,15,18};
      int data2[17]={35,51,54,58,55,17,20,48,73,12,44,55,15,58,75,18,14};
      for(j=0;j<5;++j)
        {
        for(k=0;k<arraySize;++k)
          {
          in1[k]=data1[(k+j)%17];
          in2[k]=data2[(k+j)%17];
          }
        CudaMarker_set(&m0,&s);
        CudaBuffer_sendAll(&in1Buf,in1,&s);
        CudaBuffer_sendAll(&in2Buf,in2,&s);
        CudaMarker_set(&m1,&s);
        for(int iter=0;iter<20;++iter)
          CudaProgram_launch(&p,gridDim,1,1,blockDim,1,1,0,args,&s);
        CudaMarker_set(&m2,&s);
        CudaBuffer_receiveAll(&outBuf,out,&s);
        CudaMarker_set(&m3,&s);
        CudaStream_hostSync(&s);
        Platform_msg("    ");
        for(k=0;k<20;++k) Platform_msg("%c",(char)out[k]);
        Platform_msg(" ... ");
        for(k=0;k<20;++k) Platform_msg("%c",(char)out[arraySize-20+k]);
        Platform_msg(" (%"F64"d+%"F64"d+%"F64"d us)\n",
                     CudaMarker_duration(&m1,&m0),
                     CudaMarker_duration(&m2,&m1),
                     CudaMarker_duration(&m3,&m2));
        }
      CudaLockedMem_destroy(&outMem);
      CudaLockedMem_destroy(&in1Mem);
      CudaLockedMem_destroy(&in2Mem);
      CudaBuffer_destroy(&outBuf);
      CudaBuffer_destroy(&in1Buf);
      CudaBuffer_destroy(&in2Buf);
      CudaMarker_destroy(&m0);
      CudaMarker_destroy(&m1);
      CudaMarker_destroy(&m2);
      CudaMarker_destroy(&m3);
      CudaStream_destroy(&s);
      }
    CudaProgram_destroy(&p);
    }
  }
return 0;
}

DECLARE_WINDOWS_MAIN_FUNCTION

//----------------------------------------------------------------------------
