//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "microbio.h"

#include "TransProg/random.h"
#include "TransProg/path.h"

AppliData ad;

#if defined BACTERIUM_COUNT

ActivityResult Bacterium_activate(Activity *a,Scheduler *s);

void
Bacterium_init(Bacterium *b,
               real64_t x,
               real64_t y,
               real64_t z,
               real64_t speed,
               real64_t direction,
               int cpuCtx)
{
SpinSynchro_init(&b->sync);
b->dead=false;
b->x=x;
b->y=y;
b->z=z;
b->speed=speed;
b->direction=direction;
int xs=SUBSTRATE_WIDTH_MASK&
       (int)((b->x+0.5*SUBSTRATE_SIZE)/SUBSTRATE_STEP);
int ys=SUBSTRATE_HEIGHT_MASK&
       (int)((b->y+0.5*SUBSTRATE_SIZE)/SUBSTRATE_STEP);
b->id[0]=b->id[1]=(ys<<SUBSTRATE_WIDTH_SHIFT)|xs;
b->dA[0]=b->dA[1]=0.0f;
b->dB[0]=b->dB[1]=0.0f;
Activity_init(&b->act,&Bacterium_activate,&ad.sched,cpuCtx);
#if defined DETECT_NEIGHBOURS
  b->key=SpatialGrid_coordToKey(&ad.grid,x,y,z);
  Cell_add(ad.cells+b->key,b);
#endif
}

Bacterium *
Bacterium_new(real64_t x,
              real64_t y,
              real64_t z,
              real64_t speed,
              real64_t direction,
              int cpuCtx)
{
Bacterium *b=Memory_alloc(1,Bacterium);
Bacterium_init(b,x,y,z,speed,direction,cpuCtx);
return b;
}

void
Bacterium_destroy(Bacterium *b)
{
Activity_destroy(&b->act,&ad.sched);
SpinSynchro_destroy(&b->sync);
#if defined DETECT_NEIGHBOURS
  Cell_remove(ad.cells+b->key,b);
#endif
}

void
Bacterium_delete(Bacterium *b)
{
Bacterium_destroy(b);
Memory_free(b);
}

INLINE
int
spatialAffinity(int id)
{
return (int)((id/(real64_t)(SUBSTRATE_WIDTH*SUBSTRATE_HEIGHT))*
             Platform_cpuCtxCount());
}

ActivityResult
Bacterium_activate(Activity *a,
                   Scheduler *s)
{
(void)s;
Bacterium *b=STRUCT_OF_MEMBER(a,Bacterium,act);
if(!SpinSynchro_tryLockW(&b->sync))
  {
  return Activity_retryCount(a) ? ACT_SKIP : ACT_RETRY;
  }
if(b->dead)
  {
  Bacterium_delete(b);
  return ACT_DEAD;
  }
#if !defined AVOID_RENDERING
  Rendering_showBacterium(b);
#endif
int xs=SUBSTRATE_WIDTH_MASK&(int)((b->x+0.5*SUBSTRATE_SIZE)/SUBSTRATE_STEP);
int ys=SUBSTRATE_HEIGHT_MASK&(int)((b->y+0.5*SUBSTRATE_SIZE)/SUBSTRATE_STEP);
int id=(ys<<SUBSTRATE_WIDTH_SHIFT)|xs;
#if defined SPATIAL_AFFINITY
  Activity_adjustAffinity(a,s,spatialAffinity(id),1);
#endif
int fromState2=Scheduler_fromState2(s);
int toState2=Scheduler_toState2(s);
const real32_t *A=ad.A[fromState2];
real32_t *dA=ad.dA[toState2];
const real32_t *gxA=ad.gxA[fromState2];
const real32_t *gyA=ad.gyA[fromState2];
const real32_t *B=ad.B[fromState2];
real32_t *dB=ad.dB[toState2];
const real32_t *gxB=ad.gxB[fromState2];
const real32_t *gyB=ad.gyB[fromState2];
real32_t valueA=A[id];
real32_t valueB=B[id];
real32_t deltaA=(valueB-valueA)*10000.0f*DT;
real32_t deltaB=(valueA-valueB)*10000.0f*DT;
if(id==b->id[toState2])
  {
  Atomic_postAddR32(dA+id,deltaA-b->dA[toState2]);
  Atomic_postAddR32(dB+id,deltaB-b->dB[toState2]);
  }
else
  {
  Atomic_postAddR32(dA+b->id[toState2],-b->dA[toState2]);
  Atomic_postAddR32(dA+id,deltaA);
  Atomic_postAddR32(dB+b->id[toState2],-b->dB[toState2]);
  Atomic_postAddR32(dB+id,deltaB);
  b->id[toState2]=id;
  }
b->dA[toState2]=deltaA;
b->dB[toState2]=deltaB;
#if defined SUBSTRATE_AFFINITY
  int downId=(((ys-1)&SUBSTRATE_HEIGHT_MASK)<<SUBSTRATE_WIDTH_SHIFT)|xs;
  int upId=(((ys+1)&SUBSTRATE_HEIGHT_MASK)<<SUBSTRATE_WIDTH_SHIFT)|xs;
  int leftId=(ys<<SUBSTRATE_WIDTH_SHIFT)|((xs-1)&SUBSTRATE_WIDTH_MASK);
  int rightId=(ys<<SUBSTRATE_WIDTH_SHIFT)|((xs+1)&SUBSTRATE_WIDTH_MASK);
  Activity_adjustAffinity(a,s,ad.ctx[id],1);
  Activity_adjustAffinity(a,s,ad.ctx[downId],1);
  Activity_adjustAffinity(a,s,ad.ctx[upId],1);
  Activity_adjustAffinity(a,s,ad.ctx[leftId],1);
  Activity_adjustAffinity(a,s,ad.ctx[rightId],1);
  ad.ctx[id]=Activity_cpuCtx(a);
#endif
const real64_t maxSpeed=0.01*BACTERIUM_SIZE/DT;
real64_t dx=gxA[id]-gxB[id];
real64_t dy=gyA[id]-gyB[id];
real64_t speed,direction;
cart2DToPolar64(dx,dy,&speed,&direction);
if(speed>0.01)
  {
  b->speed+=0.1*(speed*maxSpeed-b->speed);
  b->direction+=0.1*(betweenPi64(direction-b->direction));
  }
else
  {
  int cpuCtx=Activity_cpuCtx(a);
  if(!Random_int(cpuCtx,100))
    {
    b->direction+=M_PI*(0.5-Random_real64(cpuCtx))/8.0;
    b->speed=maxSpeed*Random_real64(cpuCtx);
    }
  }
real64_t sx,sy;
polarToCart2D64(b->speed,b->direction,&sx,&sy);
b->x+=DT*sx;
b->y+=DT*sy;
if(b->x>0.5*SUBSTRATE_SIZE) b->x-=SUBSTRATE_SIZE;
else if(b->x<-0.5*SUBSTRATE_SIZE) b->x+=SUBSTRATE_SIZE;
if(b->y>0.5*SUBSTRATE_SIZE) b->y-=SUBSTRATE_SIZE;
else if(b->y<-0.5*SUBSTRATE_SIZE) b->y+=SUBSTRATE_SIZE;
#if defined DETECT_NEIGHBOURS
  SpatialGridNeighbourhood sgn;
  for(intptr_t key=SpatialGridNeighbourhood_start(&sgn,&ad.grid,
                                                  b->x,b->y,b->z,
                                                  SUBSTRATE_STEP);
      key>=0;key=SpatialGridNeighbourhood_next(&sgn))
    {
    Cell *cell=ad.cells+key;
    PtrVector *cellContent=&cell->content;
    intptr_t contentSize=PtrVector_size(cellContent);
    if(!contentSize)
      {
      continue;
      }
    SpinSynchro_lockR(&cell->sync);
    for(intptr_t i=0;i<contentSize;++i)
      {
      Bacterium *b2=(Bacterium *)PtrVector_get(cellContent,i);
      Activity_adjustAffinity(a,s,Activity_cpuCtx(&b2->act),4);
      }
    SpinSynchro_unlockR(&cell->sync);
    }
  intptr_t key=SpatialGrid_coordToKey(&ad.grid,b->x,b->y,b->z);
  if(b->key!=key)
    {
    Cell_add(ad.cells+key,b);
    Cell_remove(ad.cells+b->key,b);
    b->key=key;
    }
#endif
SpinSynchro_unlockW(&b->sync);
return ACT_DONE;
}

void
createBacteria(int cpuCtx,
               void *data)
{
(void)data;
intptr_t begin=Platform_splitBegin(0,BACTERIUM_COUNT,
                                   Platform_cpuCtxCount(),cpuCtx);
intptr_t end=Platform_splitEnd(0,BACTERIUM_COUNT,
                               Platform_cpuCtxCount(),cpuCtx);
for(intptr_t i=begin;i<end;++i)
  {
  real64_t x=SUBSTRATE_SIZE*(Random_real64(cpuCtx)-0.5);
  real64_t y=SUBSTRATE_SIZE*(Random_real64(cpuCtx)-0.5);
  real64_t z=0.5*BACTERIUM_RENDER_SIZE;
  real64_t speed=(BACTERIUM_SIZE/DT)*Random_real64(cpuCtx);
  real64_t direction=M_2PI*Random_real64(cpuCtx);
  Bacterium_new(x,y,z,speed,direction,cpuCtx);
  }
}

#endif

//----------------------------------------------------------------------------

void
AppliData_reset(void)
{
int fromState2=Scheduler_fromState2(&ad.sched);
real32_t *A=ad.A[fromState2];
real32_t *B=ad.B[fromState2];
#if defined BACTERIUM_COUNT
  real32_t *gxA=ad.gxA[fromState2];
  real32_t *gxB=ad.gxB[fromState2];
  real32_t *gyA=ad.gyA[fromState2];
  real32_t *gyB=ad.gyB[fromState2];
#endif
for(int y=0;y<SUBSTRATE_HEIGHT;++y)
  {
  for(int x=0;x<SUBSTRATE_WIDTH;++x)
    {
    int id=y*SUBSTRATE_WIDTH+x;
    A[id]=0.0f;
    B[id]=0.0f;
#if defined BACTERIUM_COUNT
    gxA[id]=0.0f;
    gxB[id]=0.0f;
    gyA[id]=0.0f;
    gyB[id]=0.0f;
#endif
    }
  }
int d=SUBSTRATE_PRECISION/8;
for(int j=-d;j<=d;++j)
  {
  for(int i=-d;i<=d;++i)
    {
    for(int k=1;k<=3;++k)
      {
      int x=i+(SUBSTRATE_WIDTH*k/4);
      int y=j+(SUBSTRATE_HEIGHT*k/4);
      int id=y*SUBSTRATE_WIDTH+x;
      A[id]=1.0f;
      x=i+(SUBSTRATE_WIDTH*k/4);
      y=j+(SUBSTRATE_HEIGHT*(4-k)/4);
      id=y*SUBSTRATE_WIDTH+x;
      B[id]=1.0f;
      }
    }
  }
for(int i=0,i_end=ad.workerCount;i<i_end;++i)
  {
  ad.workers[i].needsSend=true;
  }
ad.needsReset=false;
}

void
AppliData_activate(int cpuCtx,
                   void *data)
{
(void)data;
int fromState2=Scheduler_fromState2(&ad.sched);
int toState2=Scheduler_toState2(&ad.sched);
int64_t t0=Platform_time();
if(!cpuCtx)
  {
  for(int i=0,i_end=ad.workerCount;i<i_end;++i)
    {
    Worker *w=ad.workers+i;
    if(w->dev&&w->computeDim[fromState2])
      {
      int argCount=0;
      const void *args[32];
#if !defined AVOID_RENDERING
      if(ad.needsRedraw[toState2])
        {
        args[argCount++]=CudaBuffer_programArg(w->rendering[toState2]);
        }
      else
        {
        args[argCount++]=CudaBuffer_nullProgramArg();
        }
#endif
      args[argCount++]=CudaBuffer_programArg(w->A[fromState2]);
#if defined BACTERIUM_COUNT
      args[argCount++]=CudaBuffer_programArg(w->dA[fromState2]);
      args[argCount++]=CudaBuffer_programArg(w->gxA[toState2]);
      args[argCount++]=CudaBuffer_programArg(w->gyA[toState2]);
#endif
      args[argCount++]=CudaBuffer_programArg(w->A[toState2]);
      args[argCount++]=CudaBuffer_programArg(w->B[fromState2]);
#if defined BACTERIUM_COUNT
      args[argCount++]=CudaBuffer_programArg(w->dB[fromState2]);
      args[argCount++]=CudaBuffer_programArg(w->gxB[toState2]);
      args[argCount++]=CudaBuffer_programArg(w->gyB[toState2]);
#endif
      args[argCount++]=CudaBuffer_programArg(w->B[toState2]);
      args[argCount++]=w->idOffset+fromState2;
      CudaMarker_set(w->markers[0],w->strm);
      int idOffset=w->idOffset[fromState2];
      int computeDim=w->computeDim[fromState2];
#if 1
      if(w->needsSend)
        {
        CudaBuffer_send(w->A[fromState2],
                        idOffset*sizeof(real32_t),
                        computeDim*sizeof(real32_t),
                        ad.A[fromState2]+idOffset,
                        w->strm);
        CudaBuffer_send(w->B[fromState2],
                        idOffset*sizeof(real32_t),
                        computeDim*sizeof(real32_t),
                        ad.B[fromState2]+idOffset,
                        w->strm);
        w->needsSend=false;
        }
      if(ad.workerCount>1)
        {
        int extraDim=SUBSTRATE_WIDTH;
        if(idOffset<extraDim)
          {
          extraDim-=idOffset;
          int newOffset=ad.totalDim-extraDim;
          CudaBuffer_send(w->A[fromState2],
                          newOffset*sizeof(real32_t),
                          extraDim*sizeof(real32_t),
                          ad.A[fromState2]+newOffset,
                          w->strm);
          CudaBuffer_send(w->B[fromState2],
                          newOffset*sizeof(real32_t),
                          extraDim*sizeof(real32_t),
                          ad.B[fromState2]+newOffset,
                          w->strm);
          extraDim=idOffset;
          }
        if(extraDim)
          {
          int newOffset=idOffset-extraDim;
          CudaBuffer_send(w->A[fromState2],
                          newOffset*sizeof(real32_t),
                          extraDim*sizeof(real32_t),
                          ad.A[fromState2]+newOffset,
                          w->strm);
          CudaBuffer_send(w->B[fromState2],
                          newOffset*sizeof(real32_t),
                          extraDim*sizeof(real32_t),
                          ad.B[fromState2]+newOffset,
                          w->strm);
          }
        extraDim=SUBSTRATE_WIDTH;
        if(idOffset+computeDim+extraDim>ad.totalDim)
          {
          extraDim-=ad.totalDim-(idOffset+computeDim);
          CudaBuffer_send(w->A[fromState2],
                          0,
                          extraDim*sizeof(real32_t),
                          ad.A[fromState2],
                          w->strm);
          CudaBuffer_send(w->B[fromState2],
                          0,
                          extraDim*sizeof(real32_t),
                          ad.B[fromState2],
                          w->strm);
          extraDim=ad.totalDim-(idOffset+computeDim);
          }
        if(extraDim)
          {
          int newOffset=idOffset+computeDim;
          CudaBuffer_send(w->A[fromState2],
                          newOffset*sizeof(real32_t),
                          extraDim*sizeof(real32_t),
                          ad.A[fromState2]+newOffset,
                          w->strm);
          CudaBuffer_send(w->B[fromState2],
                          newOffset*sizeof(real32_t),
                          extraDim*sizeof(real32_t),
                          ad.B[fromState2]+newOffset,
                          w->strm);
          }
        }
#if defined BACTERIUM_COUNT
      CudaBuffer_send(w->dA[fromState2],
                      idOffset*sizeof(real32_t),
                      computeDim*sizeof(real32_t),
                      ad.dA[fromState2]+idOffset,
                      w->strm);
      CudaBuffer_send(w->dB[fromState2],
                      idOffset*sizeof(real32_t),
                      computeDim*sizeof(real32_t),
                      ad.dB[fromState2]+idOffset,
                      w->strm);
#endif
#endif
      CudaMarker_set(w->markers[1],w->strm);
      CudaProgram_launch(w->prg,
                         w->gridDim[fromState2],1,1,
                         w->blockDim,1,1,
                         0,args,w->strm);
      CudaMarker_set(w->markers[2],w->strm);
#if 1
      CudaBuffer_receive(w->A[toState2],
                         idOffset*sizeof(real32_t),
                         computeDim*sizeof(real32_t),
                         ad.A[toState2]+idOffset,
                         w->strm);
#if defined BACTERIUM_COUNT
      CudaBuffer_receive(w->gxA[toState2],
                         idOffset*sizeof(real32_t),
                         computeDim*sizeof(real32_t),
                         ad.gxA[toState2]+idOffset,
                         w->strm);
      CudaBuffer_receive(w->gyA[toState2],
                         idOffset*sizeof(real32_t),
                         computeDim*sizeof(real32_t),
                         ad.gyA[toState2]+idOffset,
                         w->strm);
#endif
      CudaBuffer_receive(w->B[toState2],
                         idOffset*sizeof(real32_t),
                         computeDim*sizeof(real32_t),
                         ad.B[toState2]+idOffset,
                         w->strm);
#if defined BACTERIUM_COUNT
      CudaBuffer_receive(w->gxB[toState2],
                         idOffset*sizeof(real32_t),
                         computeDim*sizeof(real32_t),
                         ad.gxB[toState2]+idOffset,
                         w->strm);
      CudaBuffer_receive(w->gyB[toState2],
                         idOffset*sizeof(real32_t),
                         computeDim*sizeof(real32_t),
                         ad.gyB[toState2]+idOffset,
                         w->strm);
#endif
#if !defined AVOID_RENDERING
      if(ad.needsRedraw[toState2])
        {
        CudaBuffer_receive(w->rendering[toState2],
                           idOffset*sizeof(int),
                           computeDim*sizeof(int),
                           ad.rendering[toState2]+idOffset,
                           w->strm);
        }
#endif
#endif
      CudaMarker_set(w->markers[3],w->strm);
      }
    }
  }
int64_t t1=Platform_time();
if(ad.cpuWorker)
  {
  int threadCount=Platform_cpuCtxCount();
  int idOffset=ad.cpuWorker->idOffset[fromState2];
  int computeDim=ad.cpuWorker->computeDim[fromState2];
  int idBegin=(int)Platform_splitBegin(idOffset,idOffset+computeDim,
                                       threadCount,cpuCtx);
  int idEnd=(int)Platform_splitEnd(idOffset,idOffset+computeDim,
                                   threadCount,cpuCtx);
#if !defined AVOID_RENDERING
  int *rendering=ad.needsRedraw[toState2] ? ad.rendering[toState2] : (int *)0;
#endif
  const real32_t *A0=ad.A[fromState2];
#if defined BACTERIUM_COUNT
  const real32_t *dA=ad.dA[fromState2];
  real32_t *gxA=ad.gxA[toState2];
  real32_t *gyA=ad.gyA[toState2];
#endif
  real32_t *A1=ad.A[toState2];
  const real32_t *B0=ad.B[fromState2];
#if defined BACTERIUM_COUNT
  const real32_t *dB=ad.dB[fromState2];
  real32_t *gxB=ad.gxB[toState2];
  real32_t *gyB=ad.gyB[toState2];
#endif
  real32_t *B1=ad.B[toState2];
  for(int id=idBegin;id<idEnd;++id)
    {
    (*ad.cpuProgram)(
#if !defined AVOID_RENDERING
                     rendering,
#endif
                     A0,
#if defined BACTERIUM_COUNT
                     dA,gxA,gyA,
#endif
                     A1,
                     B0,
#if defined BACTERIUM_COUNT
                     dB,gxB,gyB,
#endif
                     B1,
                     id);
    }
  }
Scheduler_activate(cpuCtx,&ad.sched);
if(!cpuCtx)
  {
  int64_t t2=Platform_time();
  for(int i=0,i_end=ad.workerCount;i<i_end;++i)
    {
    Worker *w=ad.workers+i;
    if(w->dev)
      {
      CudaStream_hostSync(w->strm);
      w->duration[0]+=CudaMarker_duration(w->markers[1],w->markers[0]);
      w->duration[1]+=CudaMarker_duration(w->markers[2],w->markers[1]);
      w->duration[2]+=CudaMarker_duration(w->markers[3],w->markers[2]);
      }
    else
      {
      int64_t t3=Platform_time();
      w->duration[0]+=t1-t0;
      w->duration[1]+=t2-t1;
      w->duration[2]+=t3-t2;
      }
    // FIXME: adjust dynamically
    w->gridDim[toState2]=w->gridDim[fromState2];
    w->idOffset[toState2]=w->idOffset[fromState2];
    w->computeDim[toState2]=w->computeDim[fromState2];
    if((w->idOffset[toState2]!=w->idOffset[fromState2])||
       (w->computeDim[toState2]!=w->computeDim[fromState2]))
      {
      w->needsSend=true;
      }
    }
  }
}

void
AppliData_init(void)
{
Scheduler_init(&ad.sched);
int toState2=Scheduler_toState2(&ad.sched);
#if defined DETECT_NEIGHBOURS
  SpatialGrid_init(&ad.grid,GRID_STEP,GRID_WIDTH,GRID_HEIGHT,3);
  ad.cells=Memory_alloc(SpatialGrid_cellCount(&ad.grid),Cell);
  for(intptr_t c=0,c_end=SpatialGrid_cellCount(&ad.grid);c<c_end;++c)
    {
    SpinSynchro_init(&ad.cells[c].sync);
    PtrVector_init(&ad.cells[c].content);
    }
#endif
int i,s;
ad.mustQuit=false;
ad.needsReset=true;
ad.totalDim=SUBSTRATE_WIDTH*SUBSTRATE_HEIGHT;
ad.workerCount=0;
ad.workers=(Worker *)0;
ad.cpuWorker=(Worker *)0;
#if !defined AVOID_RENDERING
  ad.renderingDuration=0;
  ad.needsRedraw[0]=ad.needsRedraw[1]=false;
#endif
intptr_t sz=Path_fileSize(GPU_SOURCE_FILE);
if(sz<0) DROP_DEAD("no file '"GPU_SOURCE_FILE"'");
char *programSrc=Memory_alloc(sz+0x1000,char);
int offset=sprintf(programSrc,
  "#define SUBSTRATE_WIDTH_SHIFT %d\n"
  "#define SUBSTRATE_WIDTH %d\n"
  "#define SUBSTRATE_WIDTH_MASK %d\n"
  "#define SUBSTRATE_HEIGHT_SHIFT %d\n"
  "#define SUBSTRATE_HEIGHT %d\n"
  "#define SUBSTRATE_HEIGHT_MASK %d\n"
  "#define DT %ef\n"
  "#define SUBSTRATE_STEP %ef\n"
  "#define SUBSTRATE_DIFF_A %ef\n"
  "#define SUBSTRATE_DIFF_B %ef\n"
#if defined AVOID_RENDERING
  "#define AVOID_RENDERING\n"
#endif
#if defined BACTERIUM_COUNT
  "#define BACTERIUM_COUNT %d\n"
#endif
  "#line 1 \"%s\"\n",
  SUBSTRATE_WIDTH_SHIFT,SUBSTRATE_WIDTH,SUBSTRATE_WIDTH_MASK,
  SUBSTRATE_HEIGHT_SHIFT,SUBSTRATE_HEIGHT,SUBSTRATE_HEIGHT_MASK,
  DT,SUBSTRATE_STEP,
  SUBSTRATE_DIFF_A,SUBSTRATE_DIFF_B,
#if defined BACTERIUM_COUNT
  BACTERIUM_COUNT,
#endif
  GPU_SOURCE_FILE);
if(Path_readFile(GPU_SOURCE_FILE,programSrc+offset,sz,0)!=sz)
  DROP_DEAD("cannot read file '"GPU_SOURCE_FILE"'");
programSrc[offset+sz]='\0';
#if 0
  Platform_msg("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
               "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
  Platform_msg("%s\n",programSrc);
  Platform_msg("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
               "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
#endif
ad.workerCount=0;
Cuda_init();
int deviceCount;
const CudaDevice *devices=Cuda_devices(&deviceCount);
ad.workers=Memory_initAlloc(deviceCount+1,Worker);
real32_t usageSum=0.0f;
for(i=0;i<=deviceCount;++i)
  {
  Worker *w=ad.workers+ad.workerCount;
  w->usage=0.0f;
  if(Platform_argc()==1)
    {
    w->usage=1.0f/(deviceCount+1);
    }
  else if(Platform_argc()>i+1)
    {
    sscanf(Platform_argv(i+1)," %g ",&w->usage);
    }
  if(w->usage>0.0f)
    {
    usageSum+=w->usage;
    if(i<deviceCount)
      {
      w->dev=devices+i;
      }
    else
      {
      ad.cpuWorker=w;
      }
    ++ad.workerCount;
    }
  }
for(i=0;i<ad.workerCount;++i)
  {
  Worker *w=ad.workers+i;
  w->usage/=usageSum;
  if(w->dev)
    {
    const CudaDeviceProperties *prop=CudaDevice_properties(w->dev);
    Platform_msg("Worker %d uses GPU %d (%s)\n",
                 i,CudaDevice_id(w->dev),
                 CudaDevice_name(w->dev));
    Platform_msg("  multiprocessor_count: %d\n",
                 prop->multiprocessor_count);
    Platform_msg("  core_count: %d\n",
                 prop->core_count);
    Platform_msg("  warp_size: %d\n",
                 prop->warp_size);
    Platform_msg("  compute_capability: %d.%d\n",
                 prop->compute_capability_major,
                 prop->compute_capability_minor);
    Platform_msg("  max_threads_per_block: %d\n",
                 prop->max_threads_per_block);
    Platform_msg("  max_block_dim: %d %d %d\n",
                 prop->max_block_dim_x,
                 prop->max_block_dim_y,
                 prop->max_block_dim_z);
    Platform_msg("  max_grid_dim: %d %d %d\n",
                 prop->max_grid_dim_x,
                 prop->max_grid_dim_y,
                 prop->max_grid_dim_z);
    Platform_msg("  max_shared_memory_per_block: %d\n",
                 prop->max_shared_memory_per_block);
    Platform_msg("  GL_sharing: %d\n",
                 w->dev==Cuda_glDevice());
    w->strm=CudaStream_new(w->dev);
    w->markers[0]=CudaMarker_new(w->dev);
    w->markers[1]=CudaMarker_new(w->dev);
    w->markers[2]=CudaMarker_new(w->dev);
    w->markers[3]=CudaMarker_new(w->dev);
    char binaryFile[0x100];
    snprintf(binaryFile,0x100,"%s_bin%d%d",
             GPU_SOURCE_FILE,
             prop->compute_capability_major,
             prop->compute_capability_minor);
    sz=Path_fileSize(binaryFile);
    if(sz>0)
      {
      void *binaryCode=Memory_allocVoid(sz);
      if(Path_readFile(binaryFile,binaryCode,sz,0)==sz)
        {
        w->prg=CudaProgram_newFromBinary(w->dev,GPU_KERNEL_NAME,
                                         binaryCode,sz,true);
        }
      Memory_free(binaryCode);
      }
    if(!w->prg)
      {
      w->prg=CudaProgram_new(w->dev,GPU_KERNEL_NAME,
                             "--use_fast_math",programSrc,true);
#if 1
      intptr_t binaryCodeSize;
      const void *binaryCode=CudaProgram_binaryCode(w->prg,&binaryCodeSize);
      if(binaryCode)
        {
        Path_writeFile(binaryFile,binaryCode,binaryCodeSize,false);
        }
#endif
      }
    if(CudaProgram_buildLog(w->prg))
      {
      Platform_msg("build log: %s\n",CudaProgram_buildLog(w->prg));
      }
    if(CudaProgram_buildFailure(w->prg))
      {
      DROP_DEAD("gpu code compilation failure");
      }
    Platform_msg("  program:\n");
    Platform_msg("    max_threads_per_block: %d\n",
                 CudaProgram_maxThreadsPerBlock(w->prg));
    Platform_msg("    shared_size_bytes: %d\n",
                 CudaProgram_sharedSizeBytes(w->prg));
    Platform_msg("    const_size_bytes: %d\n",
                 CudaProgram_constSizeBytes(w->prg));
    Platform_msg("    local_size_bytes: %d\n",
                 CudaProgram_localSizeBytes(w->prg));
    Platform_msg("    num_regs: %d\n",
                 CudaProgram_numRegs(w->prg));
    Platform_msg("    ptx_version: %d\n",
                 CudaProgram_ptxVersion(w->prg));
    Platform_msg("    binary_version: %d\n",
                 CudaProgram_binaryVersion(w->prg));
    // FIXME: really best block dim ?
    w->blockDim=CudaProgram_maxThreadsPerBlock(w->prg);
    }
  else
    {
    Platform_msg("Worker %d uses %d threads\n",i,Platform_cpuCtxCount());
    DynamicCompiler dc;
    DynamicCompiler_init(&dc);
    ad.plugin=DynamicCompiler_build(&dc,programSrc);
    if(DynamicCompiler_buildLog(&dc))
      {
      Platform_msg("build log: %s\n",DynamicCompiler_buildLog(&dc));
      }
    if(ad.plugin)
      {
      Plugin_function(&ad.cpuProgram,ad.plugin,GPU_KERNEL_NAME);
      }
    if(!ad.cpuProgram)
      {
      char pluginMsg[0x100];
      Plugin_lastErrorMsg(pluginMsg,0x100);
      char msg[0x200];
      snprintf(msg,0x200,"dynamic code failure: %s\n",pluginMsg);
      DROP_DEAD(msg);
      }
    DynamicCompiler_destroy(&dc);
    w->blockDim=1;
    }
  int remainingDim=ad.totalDim;
  if(i)
    {
    const Worker *prev=ad.workers+i-1;
    w->idOffset[toState2]=prev->idOffset[toState2]+prev->computeDim[toState2];
    remainingDim=remainingDim>w->idOffset[toState2]
                 ? remainingDim-w->idOffset[toState2] : 0;
    }
  else
    {
    w->idOffset[toState2]=0;
    }
  if(i!=ad.workerCount-1)
    {
    int amount=(int)(ad.totalDim*w->usage);
    if(amount<=remainingDim)
      {
      remainingDim=amount;
      }
    }
  w->gridDim[toState2]=(remainingDim+w->blockDim-1)/w->blockDim;
  w->computeDim[toState2]=w->gridDim[toState2]*w->blockDim;
  if(w->idOffset[toState2]+w->computeDim[toState2]>ad.totalDim)
    {
    w->computeDim[toState2]=ad.totalDim-w->idOffset[toState2];
    }
  Platform_msg("  execution:\n");
  Platform_msg("    block_dim: %d\n",
               w->blockDim);
  Platform_msg("    grid_dim: %d\n",
               w->gridDim[toState2]);
  Platform_msg("    id_offset: %d\n",
               w->idOffset[toState2]);
  Platform_msg("    compute_dim: %d\n",
               w->computeDim[toState2]);
  if(w->dev)
    {
    const CudaDeviceProperties *prop=CudaDevice_properties(w->dev);
    if((w->blockDim>prop->max_block_dim_x)||
       (w->blockDim>CudaProgram_maxThreadsPerBlock(w->prg)))
      {
      DROP_DEAD("block dim is too large");
      }
    if(w->gridDim[toState2]>prop->max_grid_dim_x)
      {
      DROP_DEAD("grid dim is too large");
      }
    for(s=0;s<2;++s)
      {
#if !defined AVOID_RENDERING
      w->rendering[s]=CudaBuffer_new(w->dev,ad.totalDim*sizeof(int));
#endif
      w->A[s]=CudaBuffer_new(w->dev,ad.totalDim*sizeof(real32_t));
      w->B[s]=CudaBuffer_new(w->dev,ad.totalDim*sizeof(real32_t));
#if defined BACTERIUM_COUNT
      w->dA[s]=CudaBuffer_new(w->dev,ad.totalDim*sizeof(real32_t));
      w->dB[s]=CudaBuffer_new(w->dev,ad.totalDim*sizeof(real32_t));
      w->gxA[s]=CudaBuffer_new(w->dev,ad.totalDim*sizeof(real32_t));
      w->gxB[s]=CudaBuffer_new(w->dev,ad.totalDim*sizeof(real32_t));
      w->gyA[s]=CudaBuffer_new(w->dev,ad.totalDim*sizeof(real32_t));
      w->gyB[s]=CudaBuffer_new(w->dev,ad.totalDim*sizeof(real32_t));
#endif
      }
    }
  }
Memory_free(programSrc);
for(s=0;s<2;++s)
  {
  CudaLockedMem *mem;
#if !defined AVOID_RENDERING
  ad.renderingMem[s]=mem=CudaLockedMem_new(false,ad.totalDim*sizeof(int));
  ad.rendering[s]=(int *)CudaLockedMem_data(mem);
#endif
  ad.Amem[s]=mem=CudaLockedMem_new(false,ad.totalDim*sizeof(real32_t));
  ad.A[s]=(real32_t *)CudaLockedMem_data(mem);
  ad.Bmem[s]=mem=CudaLockedMem_new(false,ad.totalDim*sizeof(real32_t));
  ad.B[s]=(real32_t *)CudaLockedMem_data(mem);
#if defined BACTERIUM_COUNT
  ad.dAmem[s]=mem=CudaLockedMem_new(false,ad.totalDim*sizeof(real32_t));
  ad.dA[s]=(real32_t *)CudaLockedMem_data(mem);
  ad.dBmem[s]=mem=CudaLockedMem_new(false,ad.totalDim*sizeof(real32_t));
  ad.dB[s]=(real32_t *)CudaLockedMem_data(mem);
  ad.gxAmem[s]=mem=CudaLockedMem_new(false,ad.totalDim*sizeof(real32_t));
  ad.gxA[s]=(real32_t *)CudaLockedMem_data(mem);
  ad.gxBmem[s]=mem=CudaLockedMem_new(false,ad.totalDim*sizeof(real32_t));
  ad.gxB[s]=(real32_t *)CudaLockedMem_data(mem);
  ad.gyAmem[s]=mem=CudaLockedMem_new(false,ad.totalDim*sizeof(real32_t));
  ad.gyA[s]=(real32_t *)CudaLockedMem_data(mem);
  ad.gyBmem[s]=mem=CudaLockedMem_new(false,ad.totalDim*sizeof(real32_t));
  ad.gyB[s]=(real32_t *)CudaLockedMem_data(mem);
#endif
  }
#if defined BACTERIUM_COUNT
  ad.ctx=Memory_alloc(ad.totalDim,int);
  for(int id=0,id_end=ad.totalDim;id<id_end;++id)
    {
    ad.ctx[id]=spatialAffinity(id);
    ad.dA[0][id]=ad.dA[1][id]=ad.dB[0][id]=ad.dB[1][id]=0.0f;
    }
#endif
}

int
main(int argc,
     char **argv)
{
Platform_init(argc,argv);
#if !defined AVOID_RENDERING
  Rendering_init();
#endif
AppliData_init();
Platform_msg("DT=%gs\n",DT);
#if defined BACTERIUM_COUNT
  Platform_msg("%d bacteria\n",BACTERIUM_COUNT);
  Platform_parallel(&createBacteria,(void *)0);
#endif
int64_t now=Platform_time();
int64_t lastPerf=now;
int perfCount=0;
#if !defined AVOID_RENDERING
  int64_t lastDraw=now;
#endif
while(!ad.mustQuit)
  {
  ++perfCount;
  now=Platform_time();
  int64_t dt=now-lastPerf;
  if(dt>=2000000)
    {
    Platform_msg("%g Hz\n",perfCount/(1e-6*dt));
    for(int i=0,i_end=ad.workerCount;i<i_end;++i)
      {
      Worker *w=ad.workers+i;
      real64_t workRatio=w->computeDim[Scheduler_toState2(&ad.sched)]*100.0
                         /ad.totalDim;
      real64_t timeRatio1=w->duration[0]*100.0/dt;
      real64_t timeRatio2=w->duration[1]*100.0/dt;
      real64_t timeRatio3=w->duration[2]*100.0/dt;
      real64_t timeRatio=timeRatio1+timeRatio2+timeRatio3;
      if(w->dev)
        {
        Platform_msg("  GPU %d (%s): %.4g%% work, "
                     "%.4g%% (%.4g+%.4g+%.4g) cycle\n",
                     CudaDevice_id(w->dev),
                     CudaDevice_name(w->dev),
                     workRatio,timeRatio,timeRatio1,timeRatio2,timeRatio3);
        }
      else
        {
        Platform_msg("  %d CPU: %.4g%% work, "
                     "%.4g%% (%.4g+%.4g+%.4g) cycle\n",
                     Platform_cpuCtxCount(),
                     workRatio,timeRatio,timeRatio1,timeRatio2,timeRatio3);
        }
      w->duration[0]=w->duration[1]=w->duration[2]=0;
      }
#if !defined AVOID_RENDERING
    Platform_msg("  rendering: %.4g%% cycle\n",
                 ad.renderingDuration*100.0/dt);
    ad.renderingDuration=0;
#endif
    lastPerf+=dt;
    perfCount=0;
    }
  Scheduler_advanceStep(&ad.sched);
  if(ad.needsReset) AppliData_reset();
#if defined AVOID_RENDERING
  Platform_parallel(&AppliData_activate,(void *)0);
#else
  dt=now-lastDraw;
  int toState2=Scheduler_toState2(&ad.sched);
  ad.needsRedraw[toState2]=(dt>=40000) ?  true : false;
  if(ad.needsRedraw[toState2])
    {
    lastDraw+=dt;
    }
  Platform_parallel(&Rendering_activate,(void *)0);
#endif
  }
for(int i=0,i_end=ad.workerCount;i<i_end;++i)
  {
  Worker *w=ad.workers+i;
  if(w->dev)
    {
    for(int s=0;s<2;++s)
      {
#if !defined AVOID_RENDERING
      CudaBuffer_delete(w->rendering[s]);
#endif
      CudaBuffer_delete(w->A[s]);
      CudaBuffer_delete(w->B[s]);
#if defined BACTERIUM_COUNT
      CudaBuffer_delete(w->dA[s]);
      CudaBuffer_delete(w->dB[s]);
      CudaBuffer_delete(w->gxA[s]);
      CudaBuffer_delete(w->gxB[s]);
      CudaBuffer_delete(w->gyA[s]);
      CudaBuffer_delete(w->gyB[s]);
#endif
      }
    CudaProgram_delete(w->prg);
    CudaStream_delete(w->strm);
    }
  }
for(int s=0;s<2;++s)
  {
#if !defined AVOID_RENDERING
  if(ad.renderingMem[s]) CudaLockedMem_delete(ad.renderingMem[s]);
#endif
  CudaLockedMem_delete(ad.Amem[s]);
  CudaLockedMem_delete(ad.Bmem[s]);
#if defined BACTERIUM_COUNT
  CudaLockedMem_delete(ad.dAmem[s]);
  CudaLockedMem_delete(ad.dBmem[s]);
  CudaLockedMem_delete(ad.gxAmem[s]);
  CudaLockedMem_delete(ad.gxBmem[s]);
  CudaLockedMem_delete(ad.gyAmem[s]);
  CudaLockedMem_delete(ad.gyBmem[s]);
#endif
  }
if(ad.plugin)
  {
  Plugin_close(ad.plugin);
  }
#if defined BACTERIUM_COUNT
  Memory_free(ad.ctx);
#endif
#if !defined AVOID_RENDERING
  Rendering_destroyBuffers();
#endif
return 0;
}

DECLARE_WINDOWS_MAIN_FUNCTION

//----------------------------------------------------------------------------
