//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSCUDA_CUDA_H
#define TRANSCUDA_CUDA_H 1

#include "TransCuda/api.h"

#include "TransProg/text.h"

#define CUDA_DROP_DEAD(msg,err)                                    \
        do                                                         \
          {                                                        \
          char __errBuf[0x100];                                    \
          Cuda_errorMsg((err),__errBuf,0x100);                     \
          Platform_msg("\n%s:%d:%s()\n  %s --- %s\n",              \
                       __FILE__,__LINE__,__func__,(msg),__errBuf); \
          Platform_dropDead();                                     \
          } while(0)

typedef struct _CudaDeviceProperties
{
int64_t total_memory;
int max_threads_per_block;
int max_block_dim_x;
int max_block_dim_y;
int max_block_dim_z;
int max_grid_dim_x;
int max_grid_dim_y;
int max_grid_dim_z;
int max_shared_memory_per_block;
int total_constant_memory;
int warp_size;
int max_pitch;
int max_registers_per_block;
int clock_rate_kHz;
int texture_alignment;
int multiprocessor_count;
int kernel_exec_timeout;
int integrated;
int can_map_host_memory;
int compute_mode;
int maximum_texture1d_width;
int maximum_texture2d_width;
int maximum_texture2d_height;
int maximum_texture3d_width;
int maximum_texture3d_height;
int maximum_texture3d_depth;
int maximum_texture2d_layered_width;
int maximum_texture2d_layered_height;
int maximum_texture2d_layered_layers;
int surface_alignment;
int concurrent_kernels;
int ecc_enabled;
int pci_bus_id;
int pci_device_id;
int tcc_driver;
int memory_clock_rate_kHz;
int global_memory_bus_width;
int l2_cache_size;
int max_threads_per_multiprocessor;
int async_engine_count;
int unified_addressing;
int maximum_texture1d_layered_width;
int maximum_texture1d_layered_layers;
int maximum_texture2d_gather_width;
int maximum_texture2d_gather_height;
int maximum_texture3d_width_alternate;
int maximum_texture3d_height_alternate;
int maximum_texture3d_depth_alternate;
int pci_domain_id;
int texture_pitch_alignment;
int maximum_texturecubemap_width;
int maximum_texturecubemap_layered_width;
int maximum_texturecubemap_layered_layers;
int maximum_surface1d_width;
int maximum_surface2d_width;
int maximum_surface2d_height;
int maximum_surface3d_width;
int maximum_surface3d_height;
int maximum_surface3d_depth;
int maximum_surface1d_layered_width;
int maximum_surface1d_layered_layers;
int maximum_surface2d_layered_width;
int maximum_surface2d_layered_height;
int maximum_surface2d_layered_layers;
int maximum_surfacecubemap_width;
int maximum_surfacecubemap_layered_width;
int maximum_surfacecubemap_layered_layers;
int maximum_texture1d_linear_width;
int maximum_texture2d_linear_width;
int maximum_texture2d_linear_height;
int maximum_texture2d_linear_pitch;
int maximum_texture2d_mipmapped_width;
int maximum_texture2d_mipmapped_height;
int compute_capability_major;
int compute_capability_minor;
int maximum_texture1d_mipmapped_width;
int stream_priorities_supported;
int global_l1_cache_supported;
int local_l1_cache_supported;
int max_shared_memory_per_multiprocessor;
int max_registers_per_multiprocessor;
int managed_memory;
int multi_gpu_board;
int multi_gpu_board_group_id;
int cores_per_multiprocessor;
int core_count;
} CudaDeviceProperties;

typedef struct _CudaDevice
{
int id;
struct CUctx_st *context;
uint64_t peerMask;
char name[0x80];
CudaDeviceProperties properties;
} CudaDevice;

typedef struct _CudaStream
{
const CudaDevice *device;
struct CUstream_st *stream;
} CudaStream;

typedef struct _CudaMarker
{
const CudaDevice *device;
struct CUevent_st *event;
} CudaMarker;

typedef struct _CudaProgram
{
const CudaDevice *device;
char *name;
char *options;
char *sourceCode;
uint8_t *binaryCode;
intptr_t binaryCodeSize;
bool prefersCacheToShared;
Text buildLog;
struct CUmod_st *module;
struct CUfunc_st *kernel;
int max_threads_per_block;
int shared_size_bytes;
int const_size_bytes;
int local_size_bytes;
int num_regs;
int ptx_version;
int binary_version;
} CudaProgram;

typedef struct _CudaBuffer
{
const CudaDevice *device;
uintptr_t devptr;
size_t size;
struct CUgraphicsResource_st *gfxResource;
int gl;
} CudaBuffer;

typedef struct _CudaLockedMem
{
void *data;
intptr_t size;
bool writeOnly;
uintptr_t devptr;
} CudaLockedMem;

TRANSCUDA_API
void
Cuda_init(void);

TRANSCUDA_API
void
Cuda_errorMsg(int errorCode,
              char *out_msg,
              int msgSize);

TRANSCUDA_API
void
Cuda_profiling(bool onOff);

TRANSCUDA_API
const CudaDevice *
Cuda_devices(int *out_deviceCount);

TRANSCUDA_API
const CudaDevice * // device used for OpenGL rendering or 0
Cuda_glDevice(void);

INLINE
int
CudaDevice_id(const CudaDevice *device)
{
return device->id;
}

INLINE
const char *
CudaDevice_name(const CudaDevice *device)
{
return device->name;
}

INLINE
const CudaDeviceProperties *
CudaDevice_properties(const CudaDevice *device)
{
return &device->properties;
}

TRANSCUDA_API
int64_t
CudaDevice_freeMemory(const CudaDevice *device);

TRANSCUDA_API
void
CudaStream_init(CudaStream *stream,
                const CudaDevice *device);

INLINE
CudaStream *
CudaStream_new(const CudaDevice *device)
{
CudaStream *stream=Memory_alloc(1,CudaStream);
CudaStream_init(stream,device);
return stream;
}

TRANSCUDA_API
void
CudaStream_destroy(CudaStream *stream);

INLINE
void
CudaStream_delete(CudaStream *stream)
{
CudaStream_destroy(stream);
Memory_free(stream);
}

INLINE
const CudaDevice *
CudaStream_device(const CudaStream *stream)
{
return stream->device;
}

TRANSCUDA_API
void
CudaStream_hostSync(CudaStream *stream);

TRANSCUDA_API
void
CudaMarker_init(CudaMarker *marker,
                const CudaDevice *device);

INLINE
CudaMarker *
CudaMarker_new(const CudaDevice *device)
{
CudaMarker *marker=Memory_alloc(1,CudaMarker);
CudaMarker_init(marker,device);
return marker;
}

TRANSCUDA_API
void
CudaMarker_destroy(CudaMarker *marker);

INLINE
void
CudaMarker_delete(CudaMarker *marker)
{
CudaMarker_destroy(marker);
Memory_free(marker);
}

INLINE
const CudaDevice *
CudaMarker_device(const CudaMarker *marker)
{
return marker->device;
}

TRANSCUDA_API
void
CudaMarker_set(CudaMarker *marker,
               CudaStream *stream);

TRANSCUDA_API
void
CudaMarker_deviceSync(CudaMarker *marker,
                      CudaStream *stream);

TRANSCUDA_API
void
CudaMarker_hostSync(CudaMarker *marker);

TRANSCUDA_API
bool // previous work is done
CudaMarker_test(const CudaMarker *marker);

TRANSCUDA_API
int64_t // microseconds
CudaMarker_duration(const CudaMarker *marker,
                    const CudaMarker *previous);

TRANSCUDA_API
void
CudaProgram_init(CudaProgram *program,
                 const CudaDevice *device,
                 const char *name,
                 const char *options,
                 const char *sourceCode,
                 bool prefersCacheToShared);

INLINE
CudaProgram *
CudaProgram_new(const CudaDevice *device,
                const char *name,
                const char *options,
                const char *sourceCode,
                bool prefersCacheToShared)
{
CudaProgram *program=Memory_alloc(1,CudaProgram);
CudaProgram_init(program,device,name,options,sourceCode,prefersCacheToShared);
return program;
}

TRANSCUDA_API
void
CudaProgram_initFromBinary(CudaProgram *program,
                           const CudaDevice *device,
                           const char *name,
                           const void *binaryCode,
                           intptr_t binaryCodeSize,
                           bool prefersCacheToShared);

INLINE
CudaProgram *
CudaProgram_newFromBinary(const CudaDevice *device,
                          const char *name,
                          const void *binaryCode,
                          intptr_t binaryCodeSize,
                          bool prefersCacheToShared)
{
CudaProgram *program=Memory_alloc(1,CudaProgram);
CudaProgram_initFromBinary(program,device,name,binaryCode,binaryCodeSize,
                           prefersCacheToShared);
return program;
}

TRANSCUDA_API
void
CudaProgram_destroy(CudaProgram *program);

INLINE
void
CudaProgram_delete(CudaProgram *program)
{
CudaProgram_destroy(program);
Memory_free(program);
}

INLINE
const CudaDevice *
CudaProgram_device(const CudaProgram *program)
{
return program->device;
}

INLINE
const char *
CudaProgram_name(const CudaProgram *program)
{
return program->name;
}

INLINE
const char *
CudaProgram_options(const CudaProgram *program)
{
return program->options;
}

INLINE
const char *
CudaProgram_sourceCode(const CudaProgram *program)
{
return program->sourceCode;
}

INLINE
const void *
CudaProgram_binaryCode(const CudaProgram *program,
                       intptr_t *binaryCodeSize)
{
*binaryCodeSize=program->binaryCodeSize;
return (const void *)program->binaryCode;
}

INLINE
bool
CudaProgram_prefersCacheToShared(const CudaProgram *program)
{
return program->prefersCacheToShared;
}

INLINE
bool
CudaProgram_buildFailure(const CudaProgram *program)
{
return !program->kernel;
}

INLINE
const char * // build log or 0
CudaProgram_buildLog(const CudaProgram *program)
{
return Text_length(&program->buildLog) ? Text_get(&program->buildLog)
                                       : (const char *)0;
}

INLINE
int
CudaProgram_maxThreadsPerBlock(const CudaProgram *program)
{
return program->max_threads_per_block;
}

INLINE
int
CudaProgram_sharedSizeBytes(const CudaProgram *program)
{
return program->shared_size_bytes;
}

INLINE
int
CudaProgram_constSizeBytes(const CudaProgram *program)
{
return program->const_size_bytes;
}

INLINE
int
CudaProgram_localSizeBytes(const CudaProgram *program)
{
return program->local_size_bytes;
}

INLINE
int
CudaProgram_numRegs(const CudaProgram *program)
{
return program->num_regs;
}

INLINE
int
CudaProgram_ptxVersion(const CudaProgram *program)
{
return program->ptx_version;
}

INLINE
int
CudaProgram_binaryVersion(const CudaProgram *program)
{
return program->binary_version;
}

TRANSCUDA_API
void
CudaProgram_launch(const CudaProgram *program,
                   int gridDimX,
                   int gridDimY,
                   int gridDimZ,
                   int blockDimX,
                   int blockDimY,
                   int blockDimZ,
                   int sharedMemorySize,
                   const void * const *args,
                   CudaStream *stream);

TRANSCUDA_API
void
CudaBuffer_init(CudaBuffer *buffer,
                const CudaDevice *device,
                intptr_t size);

INLINE
CudaBuffer *
CudaBuffer_new(const CudaDevice *device,
               intptr_t size)
{
CudaBuffer *buffer=Memory_alloc(1,CudaBuffer);
CudaBuffer_init(buffer,device,size);
return buffer;
}

TRANSCUDA_API
void
CudaBuffer_initFromGl(CudaBuffer *buffer,
                      CudaStream *stream,
                      bool writeOnly,
                      int glBuffer);

INLINE
CudaBuffer *
CudaBuffer_newFromGl(CudaStream *stream,
                     bool writeOnly,
                     int glBuffer)
{
CudaBuffer *buffer=Memory_alloc(1,CudaBuffer);
CudaBuffer_initFromGl(buffer,stream,writeOnly,glBuffer);
return buffer;
}

TRANSCUDA_API
void
CudaBuffer_destroy(CudaBuffer *buffer);

INLINE
void
CudaBuffer_delete(CudaBuffer *buffer)
{
CudaBuffer_destroy(buffer);
Memory_free(buffer);
}

INLINE
const CudaDevice *
CudaBuffer_device(const CudaBuffer *buffer)
{
return buffer->device;
}

INLINE
const void * // buffer as program argument
CudaBuffer_programArg(const CudaBuffer *buffer)
{
return (const void *)&buffer->devptr;
}

INLINE
const void * // null buffer as program argument
CudaBuffer_nullProgramArg(void)
{
extern TRANSCUDA_API_VAR const void *_CudaBuffer_nullArg;
return (const void *)&_CudaBuffer_nullArg;
}

INLINE
intptr_t
CudaBuffer_size(const CudaBuffer *buffer)
{
return buffer->size;
}

INLINE
int
CudaBuffer_gl(const CudaBuffer *buffer)
{
return buffer->gl;
}

TRANSCUDA_API
void
CudaBuffer_acquireGl(CudaBuffer *buffer,
                     CudaStream *stream);

TRANSCUDA_API
void
CudaBuffer_releaseGl(CudaBuffer *buffer,
                     CudaStream *stream);

TRANSCUDA_API
void
CudaBuffer_send(CudaBuffer *buffer,
                intptr_t offset,
                intptr_t size,
                const void *data,
                CudaStream *stream);

INLINE
void
CudaBuffer_sendAll(CudaBuffer *buffer,
                   const void *data,
                   CudaStream *stream)
{
CudaBuffer_send(buffer,0,buffer->size,data,stream);
}

TRANSCUDA_API
void
CudaBuffer_receive(CudaBuffer *buffer,
                   intptr_t offset,
                   intptr_t size,
                   void *data,
                   CudaStream *stream);

INLINE
void
CudaBuffer_receiveAll(CudaBuffer *buffer,
                      void *data,
                      CudaStream *stream)
{
CudaBuffer_receive(buffer,0,buffer->size,data,stream);
}

INLINE
bool // copy from srcBuffer to destBuffer will not involve host
CudaBuffer_directCopyAvailable(const CudaBuffer *destBuffer,
                               const CudaBuffer *srcBuffer)
{
return (destBuffer->device==srcBuffer->device)||
       (destBuffer->device->peerMask&(1ULL<<srcBuffer->device->id));
}

TRANSCUDA_API
void
CudaBuffer_copy(CudaBuffer *destBuffer,
                const CudaBuffer *srcBuffer,
                intptr_t destOffset,
                intptr_t srcOffset,
                intptr_t size,
                CudaStream *stream);

TRANSCUDA_API
void
CudaLockedMem_init(CudaLockedMem *mem,
                   bool writeOnly,
                   intptr_t size);

INLINE
CudaLockedMem *
CudaLockedMem_new(bool writeOnly,
                  intptr_t size)
{
CudaLockedMem *mem=Memory_alloc(1,CudaLockedMem);
CudaLockedMem_init(mem,writeOnly,size);
return mem;
}

TRANSCUDA_API
void
CudaLockedMem_destroy(CudaLockedMem *mem);

INLINE
void
CudaLockedMem_delete(CudaLockedMem *mem)
{
CudaLockedMem_destroy(mem);
Memory_free(mem);
}

INLINE
void *
CudaLockedMem_data(const CudaLockedMem *mem)
{
return mem->data;
}

INLINE
intptr_t
CudaLockedMem_size(const CudaLockedMem *mem)
{
return mem->size;
}

INLINE
bool
CudaLockedMem_writeOnly(const CudaLockedMem *mem)
{
return mem->writeOnly;
}

INLINE
const void * // memory as zero-copy buffer program argument
CudaLockedMem_programArg(const CudaLockedMem *mem)
{
return (const void *)&mem->devptr;
}

#endif // TRANSCUDA_CUDA_H

//----------------------------------------------------------------------------
