//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSGL_VRROOM_H
#define TRANSGL_VRROOM_H 1

#include "TransProg/spatialUtils.h"
#include "TransGl/glView.h"
#include "TransGl/glDraw.h"

typedef struct _VrRoom VrRoom;

typedef
void
(*VrRoomCallback)(VrRoom *vr);

struct _VrRoom
{
VrRoomCallback callback;
void *data;
int screenCount;
struct _VrScreen *screens;
int xOffset;
int yOffset;
GlView *view;
int fboWidth;
int fboHeight;
int fboTexture;
int fboDepth;
int fbo;
int shader;
int v_Vertex;
int u_Warning;
int u_TexMatrix;
int u_TexUnit;
GlVertexArray vertexArray;
int vertexBuffer;
real32_t externalRgba[4];
int currentScreen;
bool currentRightEye;
bool hasRightEye;
};

struct _VrScreen
{
int x,y,w,h;
real64_t xbl,ybl,xbr,ybr,xtr,ytr,xtl,ytl;
real64_t tx,ty,tz,yaw,pitch,roll;
real64_t width,height;
real64_t usedSurface;
real32_t texMatrix[16];
SpatialBasis leftViewPoint;
SpatialBasis rightViewPoint;
Frustum leftFrustum;
Frustum rightFrustum;
real32_t leftViewMtx[16];
real32_t rightViewMtx[16];
real32_t leftProjMtx[16];
real32_t rightProjMtx[16];
bool stereoPermutation;
bool warning;
};

TRANSGL_API
void
VrRoom_init(VrRoom *room,
            const char *fileName,
            VrRoomCallback callback,
            void *data);

INLINE
VrRoom *
VrRoom_new(const char *fileName,
           VrRoomCallback callback,
           void *data)
{
VrRoom *room=Memory_alloc(1,VrRoom);
VrRoom_init(room,fileName,callback,data);
return room;
}

TRANSGL_API
void
VrRoom_destroy(VrRoom *room);

INLINE
void
VrRoom_delete(VrRoom *room)
{
VrRoom_destroy(room);
Memory_free(room);
}

INLINE
void *
VrRoom_data(const VrRoom *room)
{
return room->data;
}

INLINE
GlView *
VrRoom_view(const VrRoom *room)
{
return room->view;
}

INLINE
const GfxState *
VrRoom_state(const VrRoom *room)
{
return GlView_state(room->view);
}

INLINE
real32_t * // external RGBA color
VrRoom_externalRgba(VrRoom *room)
{
return room->externalRgba;
}

INLINE
int
VrRoom_screenCount(const VrRoom *room)
{
return room->screenCount;
}

INLINE
bool
VrRoom_hasRightEye(const VrRoom *room)
{
return room->hasRightEye;
}

INLINE
bool
VrRoom_stereoPermutation(const VrRoom *room,
                         int screen)
{
const struct _VrScreen *s=room->screens+screen;
return s->stereoPermutation;
}

INLINE
void
VrRoom_setStereoPermutation(VrRoom *room,
                            int screen,
                            bool stereoPermutation)
{
struct _VrScreen *s=room->screens+screen;
s->stereoPermutation=stereoPermutation;
}

INLINE
const Frustum *
VrRoom_frustum(const VrRoom *room,
               int screen,
               bool rightEye)
{
const struct _VrScreen *s=room->screens+screen;
return rightEye ? &s->rightFrustum : &s->leftFrustum;
}

INLINE
const SpatialBasis *
VrRoom_viewPoint(const VrRoom *room,
                 int screen,
                 bool rightEye)
{
const struct _VrScreen *s=room->screens+screen;
return rightEye ? &s->rightViewPoint : &s->leftViewPoint;
}

INLINE
int // currently redrawn screen
VrRoom_currentScreen(const VrRoom *room)
{
return room->currentScreen;
}

INLINE
bool // right eye is currently redrawn
VrRoom_currentRightEye(const VrRoom *room)
{
return room->currentRightEye;
}

INLINE
const Frustum * // frustum for currently redrawn screen
VrRoom_currentFrustum(const VrRoom *room)
{
return VrRoom_frustum(room,room->currentScreen,room->currentRightEye);
}

INLINE
const SpatialBasis * // view-point for currently redrawn screen
VrRoom_currentViewPoint(const VrRoom *room)
{
return VrRoom_viewPoint(room,room->currentScreen,room->currentRightEye);
}

INLINE
const real32_t * // projection matrix for currently redrawn screen
VrRoom_projectionMatrix(const VrRoom *room)
{
const struct _VrScreen *s=room->screens+room->currentScreen;
return room->currentRightEye ? s->rightProjMtx : s->leftProjMtx;
}

INLINE
const real32_t * // view matrix for currently redrawn screen
VrRoom_viewMatrix(const VrRoom *room)
{
const struct _VrScreen *s=room->screens+room->currentScreen;
return room->currentRightEye ? s->rightViewMtx : s->leftViewMtx;
}

TRANSGL_API
void
VrRoom_applyPhysicalParams(VrRoom *room,
                           const SpatialBasis *origin,
                           const SpatialBasis *viewPoint,
                           real64_t eyeOffsetX,
                           real64_t eyeOffsetY,
                           real64_t eyeOffsetZ,
                           real64_t scale,
                           real64_t nearDistance,
                           real64_t farDistance);

TRANSGL_API
void
VrRoom_updateCorners(VrRoom *room);

#endif // TRANSGL_VRROOM_H

//----------------------------------------------------------------------------
