//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSPROG_GFXUTILS_H
#define TRANSPROG_GFXUTILS_H 1

#include "TransProg/spatialUtils.h"

typedef enum { GFX_NO_EVT=0,
               GFX_BUTTON_EVT,
               GFX_KEY_EVT,
               GFX_MOTION_EVT,
               GFX_CONFIG_EVT,
               GFX_CLOSE_EVT,
               GFX_DRAW_EVT } GfxEvent;

typedef enum { GFX_NO_BTN=0,
               GFX_LEFT_BTN,
               GFX_MIDDLE_BTN,
               GFX_RIGHT_BTN,
               GFX_UP_BTN,
               GFX_DOWN_BTN } GfxButton;

typedef enum { GFX_NO_MDF      =0,
               GFX_CAPSLOCK_MDF=1<<0,
               GFX_SHIFT_MDF   =1<<1,
               GFX_CONTROL_MDF =1<<2,
               GFX_NUMLOCK_MDF =1<<3,
               GFX_LEFT_MDF    =1<<4,
               GFX_MIDDLE_MDF  =1<<5,
               GFX_RIGHT_MDF   =1<<6 } GfxModifierBit;
typedef int GfxModifier;

typedef enum { GFX_NO_STEREO=0,
               GFX_ACTIVE_STEREO,
               GFX_SPLIT_STEREO } GfxStereoMode;

typedef enum { GFX_NO_CSR=0,
               GFX_ARROW_CSR,
               GFX_TEXT_CSR,
               GFX_HAND_CSR,
               GFX_CROSS_CSR,
               GFX_WAIT_CSR,
               GFX_NORTH_CSR,
               GFX_SOUTH_CSR,
               GFX_EAST_CSR,
               GFX_WEST_CSR } GfxCursor;

typedef struct _GfxState
{
int width,height;
GfxModifier modifiers;
int mouseX,mouseY;
GfxEvent event;
GfxButton button;
char key[0x10];
bool pressed;
bool visible;
} GfxState;

INLINE
void
GfxState_init(GfxState *state,
              int width,
              int height)
{
state->width=width;
state->height=height;
state->modifiers=GFX_NO_MDF;
state->mouseX=0;
state->mouseY=0;
state->event=GFX_NO_EVT;
state->button=GFX_NO_BTN;
state->key[0]='\0';
state->pressed=false;
state->visible=false;
}

INLINE
GfxState *
GfxState_new(int width,
             int height)
{
GfxState *state=Memory_alloc(1,GfxState);
GfxState_init(state,width,height);
return state;
}

INLINE
void
GfxState_destroy(GfxState *state)
{
(void)state; // nothing to be done
}

INLINE
void
GfxState_delete(GfxState *state)
{
GfxState_destroy(state);
Memory_free(state);
}

INLINE
int // surface width
GfxState_width(const GfxState *state)
{
return state->width;
}

INLINE
int // surface height
GfxState_height(const GfxState *state)
{
return state->height;
}

INLINE
int // last mouse X coordinate
GfxState_mouseX(const GfxState *state)
{
return state->mouseX;
}

INLINE
int // last mouse Y coordinate
GfxState_mouseY(const GfxState *state)
{
return state->mouseY;
}

INLINE
GfxModifier // last modifiers seen
GfxState_modifiers(const GfxState *state)
{
return state->modifiers;
}

INLINE
GfxEvent // last event type occured
GfxState_event(const GfxState *state)
{
return state->event;
}

INLINE
GfxButton // last button pressed/released
GfxState_button(const GfxState *state)
{
return state->button;
}

INLINE
const char * // last key pressed/released
GfxState_key(const GfxState *state)
{
return state->key;
}

INLINE
bool // last key/button was pressed
GfxState_pressed(const GfxState *state)
{
return state->pressed;
}

INLINE
bool // surface is currently displayed
GfxState_visible(const GfxState *state)
{
return state->visible;
}

TRANSPROG_API
void
GfxState_displayEvent(const GfxState *state);

TRANSPROG_API
void 
Gfx_computeFrustumMatrix(real32_t *out_frustumMatrix,
                         const Frustum *frustum);

TRANSPROG_API
void
Gfx_computeProjectionMatrix(real32_t *out_projectionMatrix,
                            const Perspective *perspective);

TRANSPROG_API
void
Gfx_computeStereoProjectionMatrices(real32_t *out_leftProjectionMatrix,
                                    real32_t *out_rightProjectionMatrix,
                                    const Perspective *perspective);

TRANSPROG_API
void
Gfx_computeViewMatrix(real32_t *out_viewMatrix,
                      const SpatialBasis *viewPoint);

TRANSPROG_API
void
Gfx_computeStereoViewMatrices(real32_t *out_leftViewMatrix,
                              real32_t *out_rightViewMatrix,
                              const SpatialBasis *viewPoint,
                              const Perspective *perspective);

TRANSPROG_API
void
Gfx_computeModelViewMatrix(real32_t *out_modelViewMatrix,
                           const real32_t *viewMatrix,
                           const real64_t *modelMatrix);

TRANSPROG_API
bool // success
Gfx_saveTgaImage(const char *fileName,
                 int width,
                 int height,
                 int components,
                 const uint8_t *rgbStorage);

TRANSPROG_API
uint8_t * // allocated image storage (bottom to top) or 0
Gfx_loadTgaImage(const char *fileName,
                 int *out_width,
                 int *out_height,
                 int *out_components);

#endif // TRANSPROG_GFXUTILS_H

//----------------------------------------------------------------------------
