//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSPROG_MATH_H
#define TRANSPROG_MATH_H 1

#include "TransProg/platform.h"

//-------- Angles ------------------------------------------------------------

INLINE
void
sincos32(real32_t x,
         real32_t *out_sin,
         real32_t *out_cos)
{
#if defined HAS_SINCOS
  sincosf(x,out_sin,out_cos);
#else
  *out_sin=sinf(x);
  *out_cos=cosf(x);
#endif
}

INLINE
void
sincos64(real64_t x,
         real64_t *out_sin,
         real64_t *out_cos)
{
#if defined HAS_SINCOS
  sincos(x,out_sin,out_cos);
#else
  *out_sin=sin(x);
  *out_cos=cos(x);
#endif
}

INLINE
real32_t // angle in [-M_PI;M_PI]
betweenPi32(real32_t angle)
{
#if 0
  angle=fmodf(angle,M_2PIf);
  return angle>M_PIf ? angle-M_2PIf : angle<-M_PIf ? angle+M_2PIf : angle;
#else
  return angle-M_2PIf*floorf((angle+M_PIf)*(1.0f/M_2PIf));
#endif
}

INLINE
real32_t
degreeToRadian32(real32_t angle)
{
return angle*M_PI_180f;
}

INLINE
real32_t
radianToDegree32(real32_t angle)
{
return angle*M_180_PIf;
}

INLINE
real64_t // angle in [-M_PI;M_PI]
betweenPi64(real64_t angle)
{
#if 0
  angle=fmod(angle,M_2PI);
  return angle>M_PI ? angle-M_2PI :  angle<-M_PI ? angle+M_2PI : angle;
#else
  return angle-M_2PI*floor((angle+M_PI)*(1.0/M_2PI));
#endif
}

INLINE
real64_t
degreeToRadian64(real64_t angle)
{
return angle*M_PI_180;
}

INLINE
real64_t
radianToDegree64(real64_t angle)
{
return angle*M_180_PI;
}

//-------- Cartesian 2D <---> Polar ------------------------------------------

INLINE
real32_t
cart2DToAngle32(real32_t x,
                real32_t y)
{
return x||y ? atan2f(y,x) : 0.0f;
}

INLINE
real32_t
cart2DToDistance32(real32_t x,
                   real32_t y)
{
return sqrtf(x*x+y*y);
}

INLINE
void
cart2DToPolar32(real32_t x,
                real32_t y,
                real32_t *out_distance,
                real32_t *out_angle)
{
real32_t d=sqrtf(x*x+y*y);
*out_distance=d;
*out_angle=d ? atan2f(y,x) : 0.0f;
}

INLINE
void
polarToCart2D32(real32_t distance,
                real32_t angle,
                real32_t *out_x,
                real32_t *out_y)
{
real32_t s,c;
sincos32(angle,&s,&c);
*out_x=distance*c;
*out_y=distance*s;
}

INLINE
real64_t
cart2DToAngle64(real64_t x,
                real64_t y)
{
return x||y ? atan2(y,x) : 0.0;
}

INLINE
real64_t
cart2DToDistance64(real64_t x,
                   real64_t y)
{
return sqrt(x*x+y*y);
}

INLINE
void
cart2DToPolar64(real64_t x,
                real64_t y,
                real64_t *out_distance,
                real64_t *out_angle)
{
real64_t d=sqrt(x*x+y*y);
*out_distance=d;
*out_angle=d ? atan2(y,x) : 0.0;
}

INLINE
void
polarToCart2D64(real64_t distance,
                real64_t angle,
                real64_t *out_x,
                real64_t *out_y)
{
real64_t s,c;
sincos64(angle,&s,&c);
*out_x=distance*c;
*out_y=distance*s;
}

//-------- Cartesian 3D <---> Spherical --------------------------------------

INLINE
void
cart3DToAngles32(real32_t x,
                 real32_t y,
                 real32_t z,
                 real32_t *out_yaw,
                 real32_t *out_pitch)
{
real32_t dxy=sqrtf(x*x+y*y);
*out_yaw=dxy ? atan2f(y,x) : 0.0f;
*out_pitch=dxy||z ? -atan2f(z,dxy) : 0.0f;
}

INLINE
real32_t
cart3DToDistance32(real32_t x,
                   real32_t y,
                   real32_t z)
{
return sqrtf(x*x+y*y+z*z);
}


INLINE
void
cart3DToSpherical32(real32_t x,
                    real32_t y,
                    real32_t z,
                    real32_t *out_distance,
                    real32_t *out_yaw,
                    real32_t *out_pitch)
{
real32_t dxy=sqrtf(x*x+y*y);
real32_t d=sqrtf(dxy*dxy+z*z);
*out_distance=d;
*out_yaw=dxy ? atan2f(y,x) : 0.0f;
*out_pitch=d ? -atan2f(z,dxy) : 0.0f;
}

INLINE
void
sphericalToCart3D32(real32_t distance,
                    real32_t yaw,
                    real32_t pitch,
                    real32_t *out_x,
                    real32_t *out_y,
                    real32_t *out_z)
{
real32_t sp,cp,sy,cy;
sincos32(pitch,&sp,&cp);
sincos32(yaw,&sy,&cy);
real32_t dxy=distance*cp;
*out_x=dxy*cy;
*out_y=dxy*sy;
*out_z=-distance*sp;
}

INLINE
void
cart3DToAngles64(real64_t x,
                 real64_t y,
                 real64_t z,
                 real64_t *out_yaw,
                 real64_t *out_pitch)
{
real64_t dxy=sqrt(x*x+y*y);
*out_yaw=dxy ? atan2(y,x) : 0.0;
*out_pitch=dxy||z ? -atan2(z,dxy) : 0.0;
}

INLINE
real64_t
cart3DToDistance64(real64_t x,
                   real64_t y,
                   real64_t z)
{
return sqrt(x*x+y*y+z*z);
}

INLINE
void
cart3DToSpherical64(real64_t x,
                    real64_t y,
                    real64_t z,
                    real64_t *out_distance,
                    real64_t *out_yaw,
                    real64_t *out_pitch)
{
real64_t dxy=sqrt(x*x+y*y);
real64_t d=sqrt(dxy*dxy+z*z);
*out_distance=d;
*out_yaw=dxy ? atan2(y,x) : 0.0;
*out_pitch=d ? -atan2(z,dxy) : 0.0;
}

INLINE
void
sphericalToCart3D64(real64_t distance,
                    real64_t yaw,
                    real64_t pitch,
                    real64_t *out_x,
                    real64_t *out_y,
                    real64_t *out_z)
{
real64_t sp,cp,sy,cy;
sincos64(pitch,&sp,&cp);
sincos64(yaw,&sy,&cy);
real64_t dxy=distance*cp;
*out_x=dxy*cy;
*out_y=dxy*sy;
*out_z=-distance*sp;
}

//-------- Half precision floating point -------------------------------------

INLINE
uint16_t
real32ToHalf(real32_t value)
{
extern TRANSPROG_API_VAR uint16_t _Math_halfB[];
extern TRANSPROG_API_VAR uint8_t _Math_halfS[];
union
  {
  real32_t r;
  uint32_t i;
  } u;
u.r=value;
return (uint16_t)(_Math_halfB[(u.i>>23)&0x1FF]+
                  ((u.i&0x007FFFFF)>>_Math_halfS[(u.i>>23)&0x1FF]));
}

INLINE
real32_t
halfToReal32(uint16_t value)
{
extern TRANSPROG_API_VAR uint32_t _Math_halfM[];
extern TRANSPROG_API_VAR uint16_t _Math_halfO[];
extern TRANSPROG_API_VAR uint32_t _Math_halfE[];
union
  {
  real32_t r;
  uint32_t i;
  } u;
u.i=_Math_halfM[_Math_halfO[value>>10]+(value&0x3FF)]+
    _Math_halfE[value>>10];
return u.r;
}

INLINE
uint16_t
real64ToHalf(real64_t value)
{
return real32ToHalf((real32_t)value);
}

INLINE
real64_t
halfToReal64(uint16_t value)
{
return halfToReal32(value);
}

//-------- Square, Square-root  ----------------------------------------------

INLINE
real32_t // x*x with the sign of x
signedSquare32(real32_t x)
{
return x>=0.0f ? x*x : -x*x;
}

INLINE
real64_t // x*x with the sign of x
signedSquare64(real64_t x)
{
return x>=0.0 ? x*x : -x*x;
}

INLINE
real32_t // sqrtf(fabsf(x)) with the sign of x
signedSqrt32(real32_t x)
{
return x>=0.0f ? sqrtf(x) : -sqrtf(-x);
}

INLINE
real64_t // sqrt(fabs(x)) with the sign of x
signedSqrt64(real64_t x)
{
return x>=0.0 ? sqrt(x) : -sqrt(-x);
}

// FIXME: this approximation may not be faster than the original
INLINE
real32_t // (not always) fast approximation of 1.0f/sqrtf(x) (0.18% error)
approxInvSqrt32(real32_t x)
{
const real32_t x_2=x*0.5f;
union { real32_t y; int32_t i; } u;
u.y=x;
u.i=0x5f3759df-(u.i>>1); // initial approximation, 3.5% error
u.y*=1.5f-(x_2*u.y*u.y); // first Newton iteration, 0.18% error
// u.y*=1.5f-(x_2*u.y*u.y); // second Newton iteration, 0.0005% error
return u.y;
}

INLINE
real64_t // fast approximation of 1.0/sqrt(x) (0.18% error)
approxInvSqrt64(real64_t x)
{
const real64_t x_2=x*0.5;
union { real64_t y; int64_t i; } u;
u.y=x;
u.i=0x5fe6ec85e7de30da-(u.i>>1); // initial approximation, 3.5% error
u.y*=1.5-(x_2*u.y*u.y); // first Newton iteration, 0.18% error
// u.y*=1.5-(x_2*u.y*u.y); // second Newton iteration, 0.0005% error
return u.y;
}

//-------- Misc --------------------------------------------------------------

INLINE
real32_t
clamp32(real32_t x,
        real32_t xMin,
        real32_t xMax)
{
return fmaxf(xMin,fminf(xMax,x));
}

INLINE
real32_t
clamp64(real64_t x,
        real64_t xMin,
        real64_t xMax)
{
return fmax(xMin,fmin(xMax,x));
}

INLINE
real32_t // fmodf(x,mod) with no symmetry over the origin
cyclicModulo32(real32_t x,
               real32_t mod)
{
return x-mod*floorf(x/mod);
}

INLINE
real64_t // fmod(x,mod) with no symmetry over the origin
cyclicModulo64(real64_t x,
               real64_t mod)
{
return x-mod*floor(x/mod);
}

INLINE
void
interpolationIndex32(int elemCount,
                     real32_t cursor,
                     int *out_previousIndex,
                     real32_t *out_nextPreviousRatio)
{
cursor*=--elemCount;
int index=(int)floorf(cursor);
if(index>=elemCount)
  {
  *out_previousIndex=elemCount-1;
  *out_nextPreviousRatio=1.0f;
  }
else if(index<0)
  {
  *out_previousIndex=0;
  *out_nextPreviousRatio=0.0f;
  }
else
  {
  *out_previousIndex=index;
  *out_nextPreviousRatio=cursor-index;
  }
}

INLINE
void
interpolationIndex64(int elemCount,
                     real64_t cursor,
                     int *out_previousIndex,
                     real64_t *out_nextPreviousRatio)
{
cursor*=--elemCount;
int index=(int)floor(cursor);
if(index>=elemCount)
  {
  *out_previousIndex=elemCount-1;
  *out_nextPreviousRatio=1.0;
  }
else if(index<0)
  {
  *out_previousIndex=0;
  *out_nextPreviousRatio=0.0;
  }
else
  {
  *out_previousIndex=index;
  *out_nextPreviousRatio=cursor-index;
  }
}

TRANSPROG_API
bool // success
gaussianElimination32(int varCount,
                      real32_t *inout_equationSystem, // [varCount][varCount+1]
                      real32_t *out_vars);

TRANSPROG_API
bool // success
gaussianElimination64(int varCount,
                      real64_t *inout_equationSystem, // [varCount][varCount+1]
                      real64_t *out_vars);

#endif // TRANSPROG_MATH_H

//----------------------------------------------------------------------------
