//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSPROG_NATNETCLIENT_H
#define TRANSPROG_NATNETCLIENT_H 1

#include "TransProg/socket.h"
#include "TransProg/ptrVector.h"
#include "TransProg/spatialUtils.h"

struct _NatnetRigidBody
{
int id;
int parentId;
real64_t xOffset, yOffset, zOffset;
real64_t tx,ty,tz,qx,qy,qz,qw;
char name[1]; // extended at allocation
};

struct _NatnetSkeleton
{
int id;
PtrVector rigidBodies;
char name[1]; // extended at allocation
};

typedef struct _NatnetClient
{
Ipv4Address address;
SocketHandle sock;
int version;
uint8_t *buffer;
uint8_t *nextBuffer;
SpatialBasis calibrationOffset;
PtrVector rigidBodies;
PtrVector skeletons;
} NatnetClient;

TRANSPROG_API
void
NatnetClient_init(NatnetClient *client,
                  const char *server);

INLINE
NatnetClient *
NatnetClient_new(const char *server)
{
NatnetClient *client=Memory_alloc(1,NatnetClient);
NatnetClient_init(client,server);
return client;
}

TRANSPROG_API
void
NatnetClient_destroy(NatnetClient *client);

INLINE
void
NatnetClient_delete(NatnetClient *client)
{
NatnetClient_destroy(client);
Memory_free(client);
}

INLINE
SpatialBasis *
NatnetClient_calibrationOffset(NatnetClient *client)
{
return &client->calibrationOffset;
}

INLINE
int
NatnetClient_rigidBodyCount(const NatnetClient *client)
{
return (int)PtrVector_size(&client->rigidBodies);
}

INLINE
const char *
NatnetClient_rigidBodyName(const NatnetClient *client,
                           int rigidBody)
{
const struct _NatnetRigidBody *rb=
  (const struct _NatnetRigidBody *)
  PtrVector_get(&client->rigidBodies,rigidBody);
return rb->name;
}

INLINE
int
NatnetClient_skeletonCount(const NatnetClient *client)
{
return (int)PtrVector_size(&client->skeletons);
}

INLINE
const char *
NatnetClient_skeletonName(const NatnetClient *client,
                          int skeleton)
{
const struct _NatnetSkeleton *sk=
  (const struct _NatnetSkeleton *)
  PtrVector_get(&client->skeletons,skeleton);
return sk->name;
}

INLINE
int
NatnetClient_skeletonRigidBodyCount(const NatnetClient *client,
                                    int skeleton)
{
const struct _NatnetSkeleton *sk=
  (const struct _NatnetSkeleton *)
  PtrVector_get(&client->skeletons,skeleton);
return (int)PtrVector_size(&sk->rigidBodies);
}

INLINE
const char *
NatnetClient_skeletonRigidBodyName(const NatnetClient *client,
                                   int skeleton,
                                   int rigidBody)
{
const struct _NatnetSkeleton *sk=
  (const struct _NatnetSkeleton *)
  PtrVector_get(&client->skeletons,skeleton);
const struct _NatnetRigidBody *rb=
  (const struct _NatnetRigidBody *)
  PtrVector_get(&sk->rigidBodies,rigidBody);
return rb->name;
}

TRANSPROG_API
int // rigid-body index or -1
NatnetClient_findRigidBody(const NatnetClient *client,
                           const char *name);

TRANSPROG_API
int // skeleton index or -1
NatnetClient_findSkeleton(const NatnetClient *client,
                          const char *name);

TRANSPROG_API
int // skeleton rigid-body index or -1
NatnetClient_findSkeletonRigidBody(const NatnetClient *client,
                                   int skeleton,
                                   const char *name);

TRANSPROG_API
void
NatnetClient_rigidBodyLocation(const NatnetClient *client,
                               int rigidBody,
                               SpatialBasis *out_location);

TRANSPROG_API
void
NatnetClient_skeletonRigidBodyLocation(const NatnetClient *client,
                                       int skeleton,
                                       int rigidBody,
                                       SpatialBasis *out_location);

TRANSPROG_API
void
NatnetClient_update(NatnetClient *client);

#endif // TRANSPROG_NATNETCLIENT_H

//----------------------------------------------------------------------------
