//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSPROG_PLATFORM_H
#define TRANSPROG_PLATFORM_H 1

#include "TransProg/api.h"

#if defined _WIN32
# define DECLARE_WINDOWS_MAIN_FUNCTION                                 \
         TRANSPROG_API int _windowsMainFunction(int (*)(int,char **)); \
         int                                                           \
         WINAPI                                                        \
         WinMain(HINSTANCE hInstance,                                  \
                 HINSTANCE hPrevInstance,                              \
                 LPSTR lpCmdLine,                                      \
                 int nCmdShow)                                         \
         {                                                             \
         (void)hInstance;                                              \
         (void)hPrevInstance;                                          \
         (void)lpCmdLine;                                              \
         (void)nCmdShow;                                               \
         return _windowsMainFunction(&main);                           \
         }
  struct _PlatformW32Fncts
  {
  HMODULE mainModule;
  HMODULE kernel32;
  SIZE_T
  (WINAPI *GetLargePageMinimum)(VOID);
  BOOL
  (WINAPI *GetLogicalProcessorInformation)
          (SYSTEM_LOGICAL_PROCESSOR_INFORMATION *,
           DWORD *);
  BOOL
  (WINAPI *GetLogicalProcessorInformationEx)
          (LOGICAL_PROCESSOR_RELATIONSHIP,
           SYSTEM_LOGICAL_PROCESSOR_INFORMATION_EX *,
           DWORD *);
  BOOL
  (WINAPI *SetThreadGroupAffinity)
          (HANDLE,
           const GROUP_AFFINITY *,
           PGROUP_AFFINITY);
  VOID
  (WINAPI *InitializeConditionVariable)
          (CONDITION_VARIABLE *);
  BOOL
  (WINAPI *SleepConditionVariableCS)
          (CONDITION_VARIABLE *,
           CRITICAL_SECTION *,
           DWORD);
  VOID
  (WINAPI *WakeAllConditionVariable)
          (CONDITION_VARIABLE *);
  };
#else
# define DECLARE_WINDOWS_MAIN_FUNCTION // empty
#endif

//
// Note on the usage of integer types:
//
// A signed int should be used as general purpose counter.
// It as generaly the faster integer for many arithmetic operations.
// As it is signed, many optimizations, such as auto-vectorization, will not
// consider any potential overflow and will be enabled.
//
// To count an amount of data in memory which could overflow the with of int,
// intptr_t has to be used instead. It maps to int32_t or int64_t depending on
// the size of the address-space.
//
// For an amount of data even bigger (a time in microseconds, a file size, an
// amount of physical memory greater than the 32-bit address-space...) int64_t
// will be wide enough.
//
// In any case, we avoid using unsigned integers since they are sometimes
// slower than their signed counterpart and can prevent some optimizations
// from being enabled.
//
// The only exception is when many integer values have to be stored in arrays
// or structures. We then consider using narrower integers to save space in
// order to lower cache pressure.
//

#define DROP_DEAD(msg)                                    \
        do                                                \
          {                                               \
          Platform_msg("\n%s:%d:%s()\n  %s\n",            \
                       __FILE__,__LINE__,__func__,(msg)); \
          Platform_dropDead();                            \
          } while(0)

#define SYS_DROP_DEAD2(msg,err)                                    \
        do                                                         \
          {                                                        \
          char __errBuf[0x100];                                    \
          Platform_errorMsg((err),__errBuf,0x100);                 \
          Platform_msg("\n%s:%d:%s()\n  %s --- %s\n",              \
                       __FILE__,__LINE__,__func__,(msg),__errBuf); \
          Platform_dropDead();                                     \
          } while(0)

#define SYS_DROP_DEAD(msg) \
        SYS_DROP_DEAD2(msg,Platform_lastError())

#if defined NDEBUG
# define ASSERT_CURRENT_CPUCTX(cpuCtx) \
         do { (void)(cpuCtx); } while(0)
# define ON_ERROR(condition,msg,action) \
         do { (void)(condition); } while(0)
#else
# define ASSERT_CURRENT_CPUCTX(cpuCtx)          \
         do                                     \
           {                                    \
           if((cpuCtx)!=Platform_cpuCtx())      \
             {                                  \
             DROP_DEAD("cpu context mismatch"); \
             }                                  \
           } while(0)
# define ON_ERROR(condition,msg,action)                      \
         do                                                  \
           {                                                 \
           if((condition))                                   \
             {                                               \
             Platform_msg("\n%s:%d:%s()\n  %s\n",            \
                          __FILE__,__LINE__,__func__,(msg)); \
             action;                                         \
             }                                               \
           } while(0)
#endif

#if defined _MSC_VER
# define CALL_CPUID(level,arg,a,b,c,d)      \
         do                                 \
           {                                \
           int __info[4];                   \
           __cpuidex(__info,(level),(arg)); \
           (a)=__info[0];                   \
           (b)=__info[1];                   \
           (c)=__info[2];                   \
           (d)=__info[3];                   \
           } while(0)
#else
# if defined __i386__
#   define CALL_CPUID(level,arg,a,b,c,d)                              \
           __asm__ __volatile__("\n"                                  \
           "pushl %%ebx\n\t"   /* save ebx (in case of PIC) */        \
           "cpuid\n\t"                                                \
           "movl %%ebx,%1\n\t" /* store ebx content to esi */         \
           "popl %%ebx\n\t"    /* restore ebx */                      \
           : "=a"(a),"=S"(b),"=c"(c),"=d"(d) : "a"(level),"c"(arg) )
# elif defined __x86_64__
#   define CALL_CPUID(level,arg,a,b,c,d)                              \
           __asm__ __volatile__("\n"                                  \
           "pushq %%rbx\n\t"   /* save rbx (in case of PIC) */        \
           "cpuid\n\t"                                                \
           "movl %%ebx,%1\n\t" /* store ebx content to esi */         \
           "popq %%rbx\n\t"    /* restore rbx */                      \
           : "=a"(a),"=S"(b),"=c"(c),"=d"(d) : "a"(level),"c"(arg) )
# endif
#endif

#define ASSUMED_MAX_CACHE_LEVEL 3   // adjust as needed
#define ASSUMED_CACHE_ALIGNMENT 64  // adjust as needed
#define ASSUMED_MAX_NUMA_COUNT  4   // adjust as needed
#define ASSUMED_MAX_MEM_COUNT   512 // adjust as needed
#define ASSUMED_MAX_CPU_COUNT   256 // adjust as needed

#if defined _MSC_VER
# define CACHE_ALIGNED __declspec(align(ASSUMED_CACHE_ALIGNMENT))
#else
# define CACHE_ALIGNED __attribute__((__aligned__(ASSUMED_CACHE_ALIGNMENT)))
#endif

#if (ASSUMED_MAX_CPU_COUNT-1)<=UINT8_MAX
  typedef uint8_t NarrowCpuCtx;
#elif (ASSUMED_MAX_CPU_COUNT-1)<=UINT16_MAX
  typedef uint16_t NarrowCpuCtx;
#endif

struct _CpumemInfoNode
{
CACHE_ALIGNED
int parent;
int firstChild;
int childCount;
int firstCpu;
int cpuCount;
int numa;
int cpuId;
int cacheSize[ASSUMED_MAX_CACHE_LEVEL];
int cacheLineSize[ASSUMED_MAX_CACHE_LEVEL];
};

struct _CpuMemInfo
{
const char *cpuMask;
bool flattenCpu;
bool attachCpu;
int64_t memSize; // can exceed address-space on 32-bit systems
intptr_t pageSize;
intptr_t hugePageSize;
int maxCacheLineSize;
int maxCacheLevel;
int maxCpuId;
int maxNumaId;
int memCount;
int cpuCount;
int numaCount;
int numaId[1+ASSUMED_MAX_NUMA_COUNT];
struct _CpumemInfoNode nodes[ASSUMED_MAX_MEM_COUNT];
int maxCpuDistance;
struct
  {
  CACHE_ALIGNED
  uint8_t distances[ASSUMED_MAX_CPU_COUNT]; // minimize space
  } cpuDistances[ASSUMED_MAX_CPU_COUNT];
struct
  {
  CACHE_ALIGNED
  uint8_t proximities[ASSUMED_MAX_CPU_COUNT]; // minimize space
  } cpuProximities[ASSUMED_MAX_CPU_COUNT];
struct
  {
  CACHE_ALIGNED
  NarrowCpuCtx trip[ASSUMED_MAX_CPU_COUNT]; // minimize space
  } cpuRoundTrips[ASSUMED_MAX_CPU_COUNT];
};

TRANSPROG_API
void
Platform_init(int argc,
              char **argv);

INLINE
int // last error code
Platform_lastError(void)
{
#if defined _WIN32
  return GetLastError();
#else
  return errno;
#endif
}

TRANSPROG_API
void
Platform_errorMsg(int errorCode,
                  char *msg,
                  int msgSize);

INLINE
void
Platform_lastErrorMsg(char *msg,
                      int msgSize)
{
Platform_errorMsg(Platform_lastError(),msg,msgSize);
}

TRANSPROG_API
PRINTF_FORMAT(1,2)
int
Platform_msg(const char *format,
             ...);

TRANSPROG_API
void
Platform_dropDead(void);

INLINE
int // number of command line arguments
Platform_argc(void)
{
extern TRANSPROG_API_VAR int _Platform_argc;
return _Platform_argc;
}

INLINE
const char * // index'th command line argument
Platform_argv(int index)
{
extern TRANSPROG_API_VAR char **_Platform_argv;
return _Platform_argv[index];
}

INLINE
const char * // current executable path
Platform_executable(void)
{
extern TRANSPROG_API_VAR char *_Platform_executable;
return _Platform_executable;
}

TRANSPROG_API
bool // success
Platform_environ(const char *varName,
                 char *out_value,
                 int size);

TRANSPROG_API
int // keyword index
Platform_environAsKeyword(const char *varName,
                          const char * const *keywords,
                          int keywordCount,
                          int defaultIndex);

TRANSPROG_API
bool // boolean value extracted from environment variable
Platform_environAsBool(const char *varName,
                       bool defaultValue);

TRANSPROG_API
int // integer value extracted from environment variable
Platform_environAsInt(const char *varName,
                      int defaultValue);

TRANSPROG_API
int64_t // microseconds since 1970/01/01 UTC
Platform_time(void);

TRANSPROG_API
void
Platform_sleep(int64_t microseconds);

INLINE
void
Platform_pause(void)
{
#if defined _MSC_VER
# if defined _M_IX86 || defined __INTEL_COMPILER
    __asm { pause }
# else
    __nop();
# endif
#elif defined __POWERPC__
   __asm__ __volatile__ ("ori r0,r0,0"); // fake nop
#elif defined __i386__ || defined __x86_64__
  __asm__ __volatile__ ("pause");
#elif 0 // may not be available everywhere
  __builtin_ia32_pause();
#else
  // just do nothing...
#endif
}

INLINE
void
Platform_prefetch(const void *addr)
{
#if defined _MSC_VER
  _mm_prefetch((char *)addr,_MM_HINT_T0);
#else
  __builtin_prefetch(addr);
#endif
}

INLINE
int // current cpu context
Platform_cpuCtx(void)
{
void *cpuData;
#if defined THREAD_STORAGE
  extern TRANSPROG_API_VAR THREAD_STORAGE void *_Platform_cpuData;
  cpuData=_Platform_cpuData;
#elif defined _WIN32
  extern TRANSPROG_API_VAR DWORD _Platform_cpuData;
  cpuData=TlsGetValue(_Platform_cpuData);
#else
  extern TRANSPROG_API_VAR pthread_key_t _Platform_cpuData;
  cpuData=pthread_getspecific(_Platform_cpuData);
#endif
return (int)(intptr_t)cpuData;
}

INLINE
int // max system cpu id or -1 if unknown
Platform_maxCpuId(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.maxCpuId;
}

INLINE
int // number of cpu contexts
Platform_cpuCtxCount(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.cpuCount;
}

INLINE
int // system cpu id of memory context or -1 if not a cpu
Platform_cpuId(int cpuCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.nodes[cpuCtx].cpuId;
}

INLINE
int // max cache distance from one cpu to another
Platform_maxCpuDistance(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.maxCpuDistance;
}

INLINE
int // cache distance from one cpu to another
Platform_cpuDistance(int cpuCtx,
                     int otherCpuCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.cpuDistances[cpuCtx].distances[otherCpuCtx];
}

INLINE
int // cache proximity from one cpu to another
Platform_cpuProximity(int cpuCtx,
                      int otherCpuCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.cpuProximities[cpuCtx].proximities[otherCpuCtx];
}

INLINE
const NarrowCpuCtx * // cache-aware round-trip starting from a cpu
Platform_cpuRoundTrip(int cpuCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.cpuRoundTrips[cpuCtx].trip;
}

INLINE
int // max system numa id or -1 if unknown
Platform_maxNumaId(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.maxNumaId;
}

INLINE
int // number of numa contexts
Platform_numaCtxCount(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.numaCount;
}

INLINE
int // system id of numa context or -1 if unknown
Platform_numaId(int numaCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.numaId[numaCtx];
}

INLINE
int // max cache level or 0 if unknown
Platform_maxCacheLevel(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.maxCacheLevel;
}

INLINE
int // max cache line size
Platform_maxCacheLineSize(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.maxCacheLineSize;
}

INLINE
int // root memory context
Platform_rootMemCtx(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.memCount-1;
}

INLINE
int // number of memory contexts
Platform_memCtxCount(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.memCount;
}

INLINE
int // numa context of memory context
Platform_memNumaCtx(int memCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.nodes[memCtx].numa;
}

INLINE
int // parent of memory context or -1 if none
Platform_memParent(int memCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.nodes[memCtx].parent;
}

INLINE
int // number of children in memory context
Platform_memChildCount(int memCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.nodes[memCtx].childCount;
}

INLINE
int // first child of memory context
Platform_memFirstChild(int memCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.nodes[memCtx].firstChild;
}

INLINE
int // number of cpu contexts in memory context
Platform_memCpuCount(int memCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.nodes[memCtx].cpuCount;
}

INLINE
int // first cpu context of memory context
Platform_memFirstCpu(int memCtx)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.nodes[memCtx].firstCpu;
}

INLINE
int // cache size in bytes of memory context or 0 if unavailable
Platform_memCacheSize(int memCtx,
                      int level)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return (level>0)&&(level<=_Platform_cpuMemInfo.maxCacheLevel)
       ? _Platform_cpuMemInfo.nodes[memCtx].cacheSize[level-1]
       : 0;
}

INLINE
int // cache line size in bytes of memory context or 0 if unavailable
Platform_memCacheLineSize(int memCtx,
                          int level)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return (level>0)&&(level<=_Platform_cpuMemInfo.maxCacheLevel)
       ? _Platform_cpuMemInfo.nodes[memCtx].cacheLineSize[level-1]
       : 0;
}

INLINE
int64_t // system memory size in bytes or 0 if unknown
Platform_memorySize(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.memSize;
}

INLINE
intptr_t // standard system memory page size in bytes
Platform_standardPageSize(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.pageSize;
}

INLINE
intptr_t // huge system memory page size in bytes or 0 if unavailable
Platform_hugePageSize(void)
{
extern TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;
return _Platform_cpuMemInfo.hugePageSize;
}

INLINE
intptr_t // system memory page size in bytes
Platform_pageSize(bool preferHuge)
{
intptr_t size=preferHuge ? Platform_hugePageSize() : 0;
return size ? size : Platform_standardPageSize();
}

INLINE
intptr_t // size rounded to next alignment
Platform_alignSize(intptr_t size,
                   intptr_t alignment)
{
size+=alignment-1;
return size-(size%alignment);
}

INLINE
void * // address rounded to next alignment
Platform_alignAddress(void *address,
                      intptr_t alignment)
{
intptr_t addr=(intptr_t)address;
addr+=alignment-1;
return (void *)(addr-(addr%alignment));
}

INLINE
intptr_t // begin index of part splitIndex of [begin;end[ cut in splitCount
Platform_splitBegin(intptr_t begin,
                    intptr_t end,
                    int splitCount,
                    int splitIndex)
{
int64_t range=end-begin;
intptr_t splitBegin=begin+(intptr_t)((range*splitIndex)/splitCount);
return splitBegin<end ? splitBegin : end;
}

INLINE
intptr_t // end index of part splitIndex of [begin;end[ cut in splitCount
Platform_splitEnd(intptr_t begin,
                  intptr_t end,
                  int splitCount,
                  int splitIndex)
{
return Platform_splitBegin(begin,end,splitCount,splitIndex+1);
}

TRANSPROG_API
void
Platform_parallel(void (*fnct)(int,void *),
                  void *data);

#endif // TRANSPROG_PLATFORM_H

//----------------------------------------------------------------------------
