//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSPROG_SKELETON_H
#define TRANSPROG_SKELETON_H 1

#include "TransProg/spatialUtils.h"

struct _SkelNode
{
real64_t values[6];
void (*applyHint)(const real64_t *,SpatialBasis *);
int hint;
int channels[6];
int parent;
int childCount;
int children;
};

typedef struct _Skeleton
{
int nodeCount;
struct _SkelNode *nodes;
char **names;
int channelCount;
real64_t frameTime;
int frameCount;
real64_t *frames;
} Skeleton;

TRANSPROG_API
void
Skeleton_init(Skeleton *s);

TRANSPROG_API
Skeleton *
Skeleton_new(void);

TRANSPROG_API
void
Skeleton_destroy(Skeleton *s);

TRANSPROG_API
void
Skeleton_delete(Skeleton *s);

INLINE
int
Skeleton_nodeCount(const Skeleton *s)
{
return s->nodeCount;
}

TRANSPROG_API
int // node or -1
Skeleton_findNode(const Skeleton *s,
                  const char *name);

INLINE
const char *
Skeleton_nodeName(const Skeleton *s,
                  int node)
{
return s->names[node];
}

INLINE
int
Skeleton_transformCount(const Skeleton *s,
                        int node)
{
return (s->nodes[node].hint>>8)&0x0007;
}

INLINE
int // 0: translate  1: rot_X  2: rot_Y  3: rot_Z
Skeleton_transformKind(const Skeleton *s,
                       int node,
                       int index)
{
return (s->nodes[node].hint>>(index<<1))&0x0003;
}

INLINE
real64_t
Skeleton_transformValue(const Skeleton *s,
                        int node,
                        int index)
{
return s->nodes[node].values[index];
}

INLINE
int // channel index or -1
Skeleton_transformChannel(const Skeleton *s,
                          int node,
                          int index)
{
return s->nodes[node].channels[index];
}

INLINE
int // parent or -1
Skeleton_parent(const Skeleton *s,
                int node)
{
return s->nodes[node].parent;
}

INLINE
int
Skeleton_childCount(const Skeleton *s,
                    int node)
{
return s->nodes[node].childCount;
}

INLINE
int
Skeleton_child(const Skeleton *s,
               int node,
               int index)
{
return s->nodes[node].children+index;
}

INLINE
int
Skeleton_channelCount(const Skeleton *s)
{
return s->channelCount;
}

INLINE
int
Skeleton_frameCount(const Skeleton *s)
{
return s->frameCount;
}

INLINE
real64_t
Skeleton_frameTime(const Skeleton *s)
{
return s->frameTime;
}

INLINE
real64_t
Skeleton_frameChannel(const Skeleton *s,
                      int frame,
                      int channel)
{
return s->frames[frame*s->channelCount+channel];
}

INLINE
real64_t
Skeleton_timeChannel(const Skeleton *s,
                     real64_t time,
                     int channel)
{
int prevIndex;
real64_t nextPrevRatio;
real64_t period=(s->frameCount-1)*s->frameTime;
interpolationIndex64(s->frameCount,fmod(time/period,1.0),
                     &prevIndex,&nextPrevRatio);
return Skeleton_frameChannel(s,prevIndex,channel)*(1.0-nextPrevRatio)+
       Skeleton_frameChannel(s,prevIndex+1,channel)*nextPrevRatio;
}

TRANSPROG_API
void
Skeleton_applyFrameToNode(const Skeleton *s,
                          int node,
                          int frame,
                          SpatialBasis *inout_basis);

TRANSPROG_API
void
Skeleton_applyFrameToNodes(const Skeleton *s,
                           int startNode,
                           int frame,
                           SpatialBasis *inout_bases);

TRANSPROG_API
void
Skeleton_applyTimeToNode(const Skeleton *s,
                         int node,
                         real64_t time,
                         SpatialBasis *inout_basis);

TRANSPROG_API
void
Skeleton_applyTimeToNodes(const Skeleton *s,
                          int startNode,
                          real64_t time,
                          SpatialBasis *inout_bases);

TRANSPROG_API
bool // success
Skeleton_load(Skeleton *s,
              FILE *input);

TRANSPROG_API
bool // success
Skeleton_loadFile(Skeleton *s,
                  const char *fileName);

#endif // TRANSPROG_SKELETON_H

//----------------------------------------------------------------------------
