//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef TRANSPROG_TEXT_H
#define TRANSPROG_TEXT_H 1

#include "TransProg/ui8Vector.h"

typedef struct _Text
{
Ui8Vector v;
} Text;

INLINE
void
Text_init(Text *t)
{
Ui8Vector_init(&t->v);
Ui8Vector_push(&t->v,'\0');
}

INLINE
Text *
Text_new(void)
{
Text *t=Memory_alloc(1,Text);
Text_init(t);
return t;
}

INLINE
void
Text_destroy(Text *t)
{
Ui8Vector_destroy(&t->v);
}

INLINE
void
Text_delete(Text *t)
{
Text_destroy(t);
Memory_free(t);
}

INLINE
intptr_t
Text_capacity(const Text *t)
{
return Ui8Vector_capacity(&t->v)-1;
}

INLINE
void
Text_shrink(Text *t)
{
Ui8Vector_shrink(&t->v);
}

INLINE
void
Text_ensureMinCapacity(Text *t,
                       intptr_t capacity)
{
Ui8Vector_ensureMinCapacity(&t->v,capacity+1);
}

INLINE
intptr_t
Text_length(const Text *t)
{
return Ui8Vector_size(&t->v)-1;
}

INLINE
void
Text_clear(Text *t)
{
Ui8Vector_clear(&t->v);
Ui8Vector_push(&t->v,'\0');
}

INLINE
void
Text_setLength(Text *t,
               intptr_t newLength)
{
Ui8Vector_setSize(&t->v,newLength+1);
Ui8Vector_set(&t->v,newLength,'\0');
}

INLINE
char *
Text_stealData(Text *t)
{
char *data=(char *)Ui8Vector_stealData(&t->v);
Ui8Vector_push(&t->v,'\0');
return data;
}

INLINE
char *
Text_get(const Text *t)
{
return (char *)Ui8Vector_data(&t->v);
}

INLINE
void
Text_insertChar(Text *t,
                intptr_t index,
                char c)
{
Ui8Vector_insert(&t->v,index,c);
}

INLINE
void
Text_insertSubStr(Text *t,
                  intptr_t index,
                  const char *str,
                  intptr_t length)
{
Ui8Vector_insertMany(&t->v,index,(uint8_t *)str,length);
}

INLINE
void
Text_insertStr(Text *t,
               intptr_t index,
               const char *str)
{
Text_insertSubStr(t,index,str,strlen(str));
}

INLINE
void
Text_insertText(Text *t,
                intptr_t index,
                const Text *text)
{
Text_insertSubStr(t,index,Text_get(text),Text_length(text));
}

TRANSPROG_API
PRINTF_FORMAT(3,4)
intptr_t // inserted chars
Text_insert(Text *t,
            intptr_t index,
            const char *format,
            ...);

INLINE
void
Text_appendChar(Text *t,
                char c)
{
Text_insertChar(t,Text_length(t),c);
}

INLINE
void
Text_appendSubStr(Text *t,
                  const char *str,
                  intptr_t length)
{
Text_insertSubStr(t,Text_length(t),str,length);
}

INLINE
void
Text_appendStr(Text *t,
               const char *str)
{
Text_insertStr(t,Text_length(t),str);
}

INLINE
void
Text_appendText(Text *t,
                const Text *text)
{
Text_insertText(t,Text_length(t),text);
}

TRANSPROG_API
PRINTF_FORMAT(2,3)
intptr_t // inserted chars
Text_append(Text *t,
            const char *format,
            ...);

INLINE
void
Text_erase(Text *t,
           intptr_t index,
           intptr_t charCount)
{
Ui8Vector_eraseMany(&t->v,index,charCount);
}

INLINE
intptr_t // found index or -1
Text_findChar(const Text *t,
              char c)
{
const char *begin=Text_get(t);
const char *pos=strchr(begin,c);
return pos ? (intptr_t)(pos-begin) : -1;
}

INLINE
intptr_t // found index or -1
Text_findLastChar(const Text *t,
                  char c)
{
const char *begin=Text_get(t);
for(const char *pos=begin+Text_length(t)-1;pos>=begin;--pos)
  {
  if(*pos==c)
    {
    return (intptr_t)(pos-begin);
    }
  }
return -1;
}

INLINE
intptr_t // found index or -1
Text_findSubStr(const Text *t,
                const char *str,
                intptr_t length)
{
const char *begin=Text_get(t);
const char *end=begin+Text_length(t)-length;
for(const char *pos=begin;pos<=end;++pos)
  {
  if(!strncmp(pos,str,(size_t)length))
    {
    return (intptr_t)(pos-begin);
    }
  }
return -1;
}

INLINE
intptr_t // found index or -1
Text_findLastSubStr(const Text *t,
                    const char *str,
                    intptr_t length)
{
const char *begin=Text_get(t);
const char *end=begin+Text_length(t)-length;
for(const char *pos=end;pos>=begin;--pos)
  {
  if(!strncmp(pos,str,length))
    {
    return (intptr_t)(pos-begin);
    }
  }
return -1;
}

INLINE
intptr_t // found index or -1
Text_findStr(const Text *t,
             const char *str)
{
const char *begin=Text_get(t);
const char *pos=strstr(begin,str);
return pos ? (intptr_t)(pos-begin) : -1;
}

INLINE
intptr_t // found index or -1
Text_findLastStr(const Text *t,
                 const char *str)
{
return Text_findLastSubStr(t,str,strlen(str));
}

INLINE
intptr_t // found index or -1
Text_findText(const Text *t,
              const Text *text)
{
return Text_findSubStr(t,Text_get(text),Text_length(text));
}

INLINE
intptr_t // found index or -1
Text_findLastText(const Text *t,
                  const Text *text)
{
return Text_findLastSubStr(t,Text_get(text),Text_length(text));
}

#endif // TRANSPROG_TEXT_H

//----------------------------------------------------------------------------
