#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (C) 2008-2016 Fabrice HARROUET (ENIB)
#
# This file is part of TransProg.
#   http://www.enib.fr/~harrouet/transprog.html
# 
# TransProg is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# TransProg is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
#-----------------------------------------------------------------------------

"""TransProgConfig

TransProg specific configuration tool to generate a makefile
"""

import sys
import os
import platform
import subprocess
import re
import glob
import copy
import traceback

try:
  exec23=eval('exec') # python3
except:
  exec('def exec23(s,g=None,l=None): exec s in g,l') # python2

def bytes_to_str(b):
  if type(b)!=str: return b.decode(sys.stdout.encoding)
  return b

#-----------------------------------------------------------------------------

class Tool(object):
  """Tool: main class to describe a TransProg project"""

  class Error(Exception):
    def __init__(self,*args,**kw):
      Exception.__init__(self,*args,**kw)

  def __init__(self,options=[]):
    """Tool creation, must be called once, this is a singleton"""
    if type(self)._instance:
      raise type(self).Error('Only one instance of %s is allowed'
                             % (type(self).__name__))
    type(self)._instance=self
    object.__init__(self)
    self._stdIncRE=re.compile(r'^[ \t]*#[ \t]*include[ \t]*<([^>]+)>')
    self._locIncRE=re.compile(r'^[ \t]*#[ \t]*include[ \t]*"([^"]+)"')
    self._optRE=re.compile(r'^[ \t]*([a-zA-Z_][0-9a-zA-Z_]*)[ \t]*=(.*)$')
    self._programs=[]
    self._plugins=[]
    self._libraries=[]
    self._generators=[]
    self._cleanList=[]
    self._cleaning=False
    self._system=self.sysName()
    self._python=os.path.normpath(sys.executable)
    self.settings={}
    self._options={}
    if not isinstance(options,list):
      options=[options]
    if type(self) is Tool:
      self.manageOptions(options)
      self.settings=self.createDefaultSettings()
      self.manageOptions(options)

  def sysName(self):
    if not hasattr(self,'_system'):
      self._system=platform.system()
      macName=platform.machine()
      if self._system=='Linux':
        if re.match(r'^i.86$',macName):
          self._system+='_x86_32'
        else:
          self._system+='_'+macName
      elif self._system.endswith('BSD'):
        if macName=='i386':
          self._system+='_x86_32'
        elif macName=='amd64':
          self._system+='_x86_64'
        else:
          self._system+='_'+macName
      elif self._system=='Darwin':
        if macName=='i386':
          if platform.architecture()[0]=='64bit':
            self._system+='_x86_64'
          else:
            self._system+='_x86_32'
        else:
          self._system+='_'+macName
      else:
        self._system+='_'+macName
    return self._system

  def python(self):
    return self._python

  def env(self,varName):
    return os.environ.get(varName)

  def manageOptions(self,options):
    m=sys.modules[type(self).__module__]
    f=os.path.splitext(os.path.realpath(m.__file__))[0]+'.options'
    try:
      self.parseOptionFile(f)
    except:
      pass
    for i in options:
      self.parseOption(i)
    self._optLine=[]
    self._cmdLine=[]
    for i in sys.argv:
      if self.parseOption(i):
        self._optLine.append(i)
      else:
        self._cmdLine.append(i)
    if self.testBoolOption('use_mic'):
      self._options['native']=False
      self._options['use_intel']=True

  def parseOptionFile(self,optFile):
    """evaluate the option file"""
    if isinstance(optFile,str):
      optFile=open(optFile,'r')
    self._options['self']=self
    try:
      exec23(optFile.read(),self._options)
    except:
      traceback.print_exc()
    self._options.pop('__builtins__',None)
    self._options.pop('self',None)

  def parseOption(self,optString):
    """parse and evaluate the option, returns success"""
    m=self._optRE.match(optString)
    if not m:
      return False
    name=m.group(1)
    value=m.group(2).strip()
    self._options['self']=self
    try:
      self._options[name]=eval(value,self._options)
    except:
      sys.stdout.write('Cannot evaluate option `%s\':\n' % (optString))
      sys.stdout.write('    %s\n' % (sys.exc_info()[1]))
      self._options[name]=None
    self._options.pop('__builtins__',None)
    self._options.pop('self',None)
    return True

  def getOption(self,opt):
    """retreive option's value or None"""
    return self._options.get(opt)

  def testBoolOption(self,opt):
    """test if an option, considered as boolean, is set"""
    v=self._options.get(opt)
    if not v:
      return False
    if isinstance(v,bool):
      return v
    if isinstance(v,str):
      try:
        v=int(v)
      except:
        try:
          v=long(v)
        except:
          try:
            v=float(v)
          except:
            pass
    if isinstance(v,int) or isinstance(v,long) or isinstance(v,float):
      return v!=0
    if isinstance(v,str):
      v=v.lower()
      return (v=='true') or (v=='yes') or (v=='y') or (v=='on')
    return False

  def createDefaultSettings(self):
    """create new settings with default values"""
    settings={}
    settings['MAKE']=[]
    settings['CC']=[]
    settings['DBGFLAGS']=[]
    settings['OPTFLAGS']=[]
    settings['CFLAGS']=[]
    settings['DEFINES']={}
    settings['INCDIRS']=[os.path.curdir]
    settings['LDFLAGS']=[]
    settings['LIBDIRS']=[os.path.curdir]
    settings['LIBS']=[]
    settings['HAS_RPATH']=True
    if not self.testBoolOption('no_transprog'):
      topDir=os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
      m=re.match(r'.*TransProg_([0-9]{4})m([0-9]{2})d([0-9]{2}).*',
                 os.path.basename(topDir))
      if m:
        transProgVersion=m.group(1)+m.group(2)+m.group(3)
      else:
        transProgVersion='0'
      settings['TRANSPROG_DIR']=topDir
      settings['DEFINES']['TRANSPROG_DIR']='"%s"' % topDir
      settings['TRANSPROG_VERSION']=transProgVersion
      settings['DEFINES']['TRANSPROG_VERSION']=transProgVersion
      settings['INCDIRS']+=[os.path.join(topDir,'include')]
      settings['LIBDIRS']+=[os.path.join(topDir,'lib')]
      settings['LIBS']+=['TransProg']
    if self._system.startswith('Windows'):
      settings['LIBS']+=['WS2_32','USER32','ADVAPI32']
    else:
      settings['LIBS']+=['m','pthread']
      if self._system.find('BSD')==-1:
        settings['LIBS']+=['dl']
    if self.testBoolOption('use_msc'):
      settings['MAKE']=['nmake.exe','/nologo','/k']
      if self.testBoolOption('use_cxx'): 
        settings['CFLAGS']+=['/TP','/EHsc']
      else:
        settings['CFLAGS']+=['/TC']
      if self.testBoolOption('use_intel'):
        settings['CC']=['icl.exe','/nologo']
        settings['OPTFLAGS']+=['/O3',
                               '/Qopt-report-phase:all',
                               '/Qopt-report:5','/Qopt-report-per-object']
        if self.testBoolOption('native'):
          settings['OPTFLAGS']+=['-QxHost']
        if self.testBoolOption('use_mic'):
          settings['CFLAGS']+=['/Qmic']
          settings['LDFLAGS']+=['/Qmic']
        settings['CFLAGS']+=['/Wall','/Wcheck']
      else:
        settings['CC']=['cl.exe','/nologo']
        settings['OPTFLAGS']+=['/O2']
        settings['CFLAGS']+=['/W3']
      settings['CFLAGS']+=['/MD']
      settings['DBGFLAGS']+=['/Z7']
      settings['LDFLAGS']+=['/link','/manifest']
      settings['HAS_RPATH']=False
    else:
      if self._system.startswith('Windows'):
        settings['MAKE']=['mingw32-make.exe','-k']
      else:
        settings['MAKE']=['make','-k']
      if self.testBoolOption('use_cxx'): 
        if self.testBoolOption('use_intel'):
          settings['CC']=['icpc']
        else:
          settings['CC']=['g++']
        settings['CFLAGS']+=['-xc++']
      else:
        if self.testBoolOption('use_intel'):
          settings['CC']=['icc']
        else:
          settings['CC']=['gcc']
        if self.testBoolOption('use_mic'):
          settings['CFLAGS']+=['-mmic']
          settings['LDFLAGS']+=['-mmic']
        settings['CFLAGS']+=['-xc','-std=c99','-U__STRICT_ANSI__']
      settings['CFLAGS']+=['-Wall']
      if self.testBoolOption('use_intel'):
        settings['CFLAGS']+=['-Wcheck']
      else:
        settings['CFLAGS']+=['-Wextra','-Wno-long-long','-pedantic']
      if not self.testBoolOption('use_cxx'):
        if not self.testBoolOption('use_intel'):
          settings['CFLAGS']+=['-Wc++-compat']
      settings['OPTFLAGS']+=['-O3','-DNDEBUG']
      if self.testBoolOption('use_intel'): 
        settings['OPTFLAGS']+=['-qopt-report-phase=all',
                               '-qopt-report=5','-qopt-report-per-object']
      else:
        settings['OPTFLAGS']+=['-ffast-math',
                               '-fomit-frame-pointer']
        if not self._system.startswith('Darwin'):
          settings['OPTFLAGS']+=['-s']
      if self.testBoolOption('native'):
        if self.testBoolOption('use_intel'): 
          settings['OPTFLAGS']+=['-xHost']
        else:
          settings['OPTFLAGS']+=['-march=native']
      settings['DBGFLAGS']+=['-g','-UNDEBUG']
      settings['DEFINES']['_REENTRANT']=None
      if self._system.startswith('Windows'):
        settings['HAS_RPATH']=False
      else:
        settings['CFLAGS']+=['-fPIC']
        if self._system.startswith('Darwin'):
          settings['HAS_RPATH']=False
    return settings

  def addClSettings(self,settings=None):
    """add platform specific settings for OpenCL"""
    if settings is None:
      settings=self.settings
    if self._system.startswith('Darwin'):
      settings['LDFLAGS']+=['-framework','OpenCL']
    else:
      settings['LIBS']+=['OpenCL']

  def addCudaSettings(self,settings=None):
    """add platform specific settings for Cuda"""
    if settings is None:
      settings=self.settings
    if self._system.startswith('Darwin'):
      settings['LDFLAGS']+=['-framework','CUDA']
      settings['LIBS']+=['nvrtc']
    elif self._system.startswith('Windows'):
      settings['LIBS']+=['cuda','nvrtc']
    else:
      for d in ['/opt','/usr/local']:
        settings['LIBDIRS']+=[d+'/cuda/lib64']
        settings['INCDIRS']+=[d+'/cuda/include']
      settings['LIBDIRS']+=['/usr/lib64/nvidia']
      settings['LIBS']+=['cuda','nvrtc']

  def copySettings(self,settings=None):
    """returns a deep copy of settings"""
    if settings is None:
      settings=self.settings
    return copy.deepcopy(settings)

  def importFlags(self,flags,settings=None):
    """parses cc-like flags and fills settings accordingly, returns success"""
    #----
    def unquote(str):
      if (str.startswith('"') and str.endswith('"')) or \
         (str.startswith("'") and str.endswith("'")):
        return str[1:-1]
      else:
        return str.replace('\\ ',' ')
    #----
    def moreArg(flags,i,arg):
      args=[arg]
      count=len(flags)
      while i<count:
        f=flags[i]
        if not f.startswith('-'):
          args.append(f)
          i+=1
        else:
          break
      return (i,unquote(' '.join(args).strip()))
    #----
    if flags.startswith('!'):
      try:
        p=subprocess.Popen(flags[1:],shell=True,stdout=subprocess.PIPE)
        flags=p.communicate()[0]
        if p.returncode:
          return False
      except:
        # traceback.print_exc()
        return False
    if settings is None:
      settings=self.settings
    flags=flags.replace('\n',' ').split(' ')
    count=len(flags)
    i=0
    while i<count:
      f=flags[i]
      i+=1
      if f.startswith('-I'):
        (i,arg)=moreArg(flags,i,f[2:])
        settings['INCDIRS'].append(arg)
      elif f.startswith('-L'):
        (i,arg)=moreArg(flags,i,f[2:])
        settings['LIBDIRS'].append(arg)
      elif f.startswith('-l'):
        (i,arg)=moreArg(flags,i,f[2:])
        settings['LIBS'].append(arg)
      elif f.startswith('-D'):
        (i,arg)=moreArg(flags,i,f[2:])
        macro=arg.split('=',1)
        if macro and macro[0]:
          key=macro[0]
          value=None
          if len(macro)>1 and macro[1]:
            value=macro[1]
          else:
            value=None
          settings['DEFINES'][key]=value
      elif f.startswith('-Wl,'):
        (i,arg)=moreArg(flags,i,f)
        settings['LDFLAGS'].append(arg)
      else:
        args=[f]
        (i,arg)=moreArg(flags,i,f)
        settings['CFLAGS'].append(arg)
    return True

  def collectCppDeps(self,filename,settings=None):
    """finds filename dependencies according to #include"""
    #----
    def collect(tool,filename,deps,settings):
      if filename not in deps:
        try:
          f=open(filename,'r')
        except:
          t=tool._findTarget(filename)
          if not t:
            return False
          deps.append(filename)
          for s in t.sources:
            collect(tool,s,deps,settings)
          return True
        deps.append(filename)
        for l in f:
          l=bytes_to_str(l)
          m=tool._stdIncRE.match(l)
          if m: # incdirs --> (stddirs) --> localdir
            inc=m.group(1)
            done=False
            for d in settings['INCDIRS']:
              header=os.path.normpath(os.path.join(d,inc))
              if collect(tool,header,deps,settings):
                done=True
                break
            if not done:
              header=os.path.normpath(os.path.join(os.path.dirname(filename),
                                                   inc))
              collect(tool,header,deps,settings)
          else:
            m=tool._locIncRE.match(l)
            if m: # localdir --> incdirs --> (stddirs)
              inc=m.group(1)
              header=os.path.normpath(os.path.join(os.path.dirname(filename),
                                                   inc))
              if not collect(tool,header,deps,settings):
                for d in settings['INCDIRS']:
                  header=os.path.normpath(os.path.join(d,inc))
                  if collect(tool,header,deps,settings):
                    break
      return True
    #----
    if settings is None:
      settings=self.settings
    deps=[]
    filename=os.path.normpath(filename)
    collect(self,filename,deps,settings)
    if not deps:
      deps=[filename]
    return deps

  def findExecutable(self,cmd):
    cmd=os.path.normpath(cmd)
    if cmd.find(os.path.pathsep)!=-1: # path to executable given
      dirs=['']
    else:
      dirs=[i or os.path.curdir
            for i in os.environ['PATH'].split(os.path.pathsep)]
    for i in dirs:
      i=os.path.join(i,cmd)
      if self._system.startswith('Windows'):
        execExt=['.com','.exe','.bat']
        if os.path.splitext(i)[1].lower() in execExt:
          if os.path.isfile(i) and os.access(i,os.X_OK):
            return i
        else:
          for ext in execExt:
            prog=i+ext
            if os.path.isfile(prog) and os.access(prog,os.X_OK):
              return prog
      else:
        if os.path.isfile(i) and os.access(i,os.X_OK):
          return i
    return None

  def compilerCmd(self,target,source,settings=None):
    if settings is None:
      settings=self.settings
    cmd=[]
    if self.testBoolOption('color'):
      cmd+=[self._python]
      cmd+=[os.path.join(os.path.dirname(os.path.realpath(__file__)),
                         'color_cc.py')]
    cmd+=settings['CC']
    if self.testBoolOption('use_msc'):
      cmd+=['/Fo%s' % (target),'/c']
    else:
      cmd+=['-o',target,'-c']
    cmd+=self.showSettings(['--cflags'],settings)[0]
    cmd+=[source]
    return cmd

  def linkerCmd(self,kind,target,objects,settings=None):
    if settings is None:
      settings=self.settings
    if not isinstance(objects,list):
      objects=[objects]
    cmd=[]
    if self.testBoolOption('color'):
      cmd+=[self._python]
      cmd+=[os.path.join(os.path.dirname(os.path.realpath(__file__)),
                         'color_cc.py')]
    cmd+=settings['CC']
    cmd+=objects
    cmd+=self.showSettings(['--libs'],settings)[0]
    if self.testBoolOption('use_msc'):
      cmd+=['/out:%s' % (target)]
      if (kind=='plug') or (kind=='lib'):
        lib=os.path.splitext(target)[0]+'.lib'
        cmd+=['/dll','/implib:%s' % (lib)]
      elif kind=='prog':
        if self.testBoolOption('hide_windows_console'):
          cmd+=['/subsystem:windows']
        else:
          cmd+=['/subsystem:console']
    else:
      cmd+=['-o',target]
      if self._system.startswith('Darwin'):
        if kind=='prog':
          pass
        elif kind=='plug':
          cmd+=['-bundle']
        elif kind=='lib':
          # library will be found near the executable
          cmd+=['-dynamiclib','-install_name',
                os.path.join('@executable_path',os.path.basename(target))]
      else:
        if (kind=='plug') or (kind=='lib'):
          cmd+=['-shared']
        elif (kind=='prog') and self._system.startswith('Windows'):
          if self.testBoolOption('hide_windows_console'):
            cmd+=['-Wl,--subsystem,windows']
          else:
            cmd+=['-Wl,--subsystem,console']
    if settings['HAS_RPATH']:
      for i in settings['LIBDIRS']:
        cmd.append('-Wl,-rpath,%s' % (i))
    return cmd

  def versionInfo(self,settings=None):
    """give version informations"""
    if settings is None:
      settings=self.settings
    result='TransProg'
    v=settings.get('TRANSPROG_VERSION')
    if v is not None:
      result+=' version %s' % v
    result+=' (under %s)\n' % self.sysName()
    d=settings.get('TRANSPROG_DIR')
    if d is None:
      d=os.path.dirname(os.path.realpath(__file__))
    result+='in `%s\'\n' % d
    return result

  def showSettings(self,args,settings=None):
    """expand settings according to args, returns (out_list,error_list)"""
    if settings is None:
      settings=self.settings
    if args is None:
      args=self._cmdLine[1:]
    if not isinstance(args,list):
      args=[args]
    out=[]
    err=[]
    cc=False
    cflags=''
    libs=''
    help=False
    for i in args:
      if i=='--cc':
        cc=True
      elif (i=='--cflags') or (i=='--cflags-only-I') or \
           (i=='--cflags-only-D') or (i=='--cflags-only-other'):
        cflags=i
      elif (i=='--libs') or (i=='--libs-only-L') or \
           (i=='--libs-only-l') or (i=='--libs-only-other'):
        libs=i
      elif i=='--help':
        help=True
      else:
        err.append('%s: unknown option' % (i))
        help=True
    if help:
      cc=' '.join(settings['CC'])
      help=self.versionInfo(settings)
      help+='  --cc                   Compiler (%s)\n' % cc
      help+='  --cflags               All compiler flags\n'
      help+='  --cflags-only-I        Only -I compiler flags\n'
      help+='  --cflags-only-D        Only -D compiler flags\n'
      help+='  --cflags-only-other    Neither -I nor -D compiler flags\n'
      help+='  --libs                 All linker flags\n'
      help+='  --libs-only-L          Only -L linker flags\n'
      help+='  --libs-only-l          Only -l linker flags\n'
      help+='  --libs-only-other      Neither -L nor -l linker flags\n'
      help+='  --help                 This help message'
      if err:
        err.append(help)
      else:
        out.append(help)
    else:
      for s in ['INCDIRS','LIBDIRS']:
        unique=[]
        for i in settings[s]:
          i=os.path.normpath(i)
          if i not in unique and os.path.isdir(i):
            unique.append(i)
        settings[s]=unique
      unique=[]
      for i in settings['LIBS']:
        if i not in unique:
          unique.append(i)
      settings['LIBS']=unique
      if cc:
        out+=settings['CC']
      if (cflags=='--cflags') or (cflags=='--cflags-only-other'):
        for i in settings['CFLAGS']:
          out.append(i)
        if self.testBoolOption('optimize'):
          for i in settings['OPTFLAGS']:
            out.append(i)
        if self.testBoolOption('debug'):
          for i in settings['DBGFLAGS']:
            out.append(i)
      if (cflags=='--cflags') or (cflags=='--cflags-only-D'):
        if self.testBoolOption('use_msc'):
          opt='/D'
        else:
          opt='-D'
        for i in settings['DEFINES']:
          v=settings['DEFINES'][i]
          if v is None:
            out.append('%s%s' % (opt,i))
          else:
            out.append('%s%s=%s' % (opt,i,v))
      if (cflags=='--cflags') or (cflags=='--cflags-only-I'):
        if self.testBoolOption('use_msc'):
          opt='/I'
        else:
          opt='-I'
        for i in settings['INCDIRS']:
          out.append('%s%s' % (opt,os.path.normpath(i)))
      if (libs=='--libs') or (libs=='--libs-only-other'):
        for i in settings['LDFLAGS']:
          out.append(i)
      if (libs=='--libs') or (libs=='--libs-only-L'):
        if self.testBoolOption('use_msc'):
          opt='/libpath:'
        else:
          opt='-L'
        for i in settings['LIBDIRS']:
          out.append('%s%s' % (opt,os.path.normpath(i)))
      if (libs=='--libs') or (libs=='--libs-only-l'):
        if self.testBoolOption('use_msc'):
          preOpt=''
          postOpt='.lib'
        else:
          preOpt='-l'
          postOpt=''
        for i in settings['LIBS']:
          out.append('%s%s%s' % (preOpt,i,postOpt))
    return (out,err)

  def Program(self,target,sources,cb=None,cbArg=None,settings=None):
    if settings is None:
      settings=self.settings
    settings=copy.deepcopy(settings)
    if not isinstance(sources,list):
      sources=[sources]
    target=os.path.normpath(target)
    prog=os.path.basename(target)
    if self._system.startswith('Windows'):
      prog+='.exe'
    elif self._system.startswith('Darwin'):
      prog=os.path.join(prog+'.app','Contents','MacOS',prog);
    prog=os.path.join(os.path.dirname(target),prog)
    if self._findTarget(prog):
      raise type(self).Error('Multiple occurences for the target `%s\''
                             % (prog))
    sources=[os.path.normpath(i) for i in sources]
    self._programs.append(_BinTarget(self,'prog',prog,sources,
                                     cb,cbArg,settings))
    return prog

  def Plugin(self,target,sources,cb=None,cbArg=None,settings=None):
    if settings is None:
      settings=self.settings
    settings=copy.deepcopy(settings)
    if not isinstance(sources,list):
      sources=[sources]
    target=os.path.normpath(target)
    plug=os.path.basename(target)
    if self._system.startswith('Windows'):
      plug+='.dll'
    elif self._system.startswith('Darwin'):
      plug+='.bundle'
    else:
      plug+='.so'
    plug=os.path.join(os.path.dirname(target),plug)
    if self._findTarget(plug):
      raise type(self).Error('Multiple occurences for the target `%s\''
                             % (plug))
    sources=[os.path.normpath(i) for i in sources]
    self._plugins.append(_BinTarget(self,'plug',plug,sources,
                                    cb,cbArg,settings))
    return plug

  def Library(self,target,sources,cb=None,cbArg=None,settings=None):
    if settings is None:
      settings=self.settings
    settings=copy.deepcopy(settings)
    if not isinstance(sources,list):
      sources=[sources]
    target=os.path.normpath(target)
    lib=os.path.basename(target)
    if self._system.startswith('Windows'):
      lib+='.dll'
    elif self._system.startswith('Darwin'):
      lib='lib%s.dylib' % (lib)
    else:
      lib='lib%s.so' % (lib)
    lib=os.path.join(os.path.dirname(target),lib)
    if self._findTarget(lib):
      raise type(self).Error('Multiple occurences for the target `%s\''
                             % (lib))
    sources=[os.path.normpath(i) for i in sources]
    self._libraries.append(_BinTarget(self,'lib',lib,sources,
                                      cb,cbArg,settings))
    return lib

  def Generator(self,target,sources,cb,cbArg,settings=None):
    if settings is None:
      settings=self.settings
    settings=copy.deepcopy(settings)
    if not isinstance(sources,list):
      sources=[sources]
    target=os.path.normpath(target)
    if self._findTarget(target):
      raise type(self).Error('Multiple occurences for the target `%s\''
                             % (target))
    sources=[os.path.normpath(i) for i in sources]
    self._generators.append(_Target(self,'gen',target,sources,
                                    cb,cbArg,settings))
    return target

  def Template(self,target,sources,tmplArgs=[],
                    cb=None,cbArg=None,settings=None):
    if settings is None:
      settings=self.settings
    settings=copy.deepcopy(settings)
    if not isinstance(sources,list):
      sources=[sources]
    target=os.path.normpath(target)
    if not isinstance(tmplArgs,list):
      tmplArgs=[tmplArgs]
    if self._findTarget(target):
      raise type(self).Error('Multiple occurences for the target `%s\''
                             % (target))
    sources=[os.path.normpath(i) for i in sources]
    self._generators.append(_TemplateTarget(self,'gen',target,sources,
                                            tmplArgs,cb,cbArg,settings))
    return target

  def Clean(self,files):
    if not isinstance(files,list):
      files=[files]
    self._cleanList+=files

  def programs(self):
    return [i.name for i in self._programs]

  def plugins(self):
    return [i.name for i in self._plugins]

  def libraries(self):
    return [i.name for i in self._libraries]

  def generators(self):
    return [i.name for i in self._generators]

  def targets(self):
    return self.generators()+self.libraries()+self.plugins()+self.programs()

  def run(self,all=None):
    """generate makefile with default targets or perform command line"""
    quiet=self.testBoolOption('quiet')
    if len(self._cmdLine)>1:
      status=0
      cmd=self._cmdLine[1]
      if cmd=='--make':
        cmdLine=self.settings['MAKE']+self._cmdLine[2:]
        cmdLine[0]=os.path.normpath(cmdLine[0])
        try:
          proc=subprocess.Popen(cmdLine)
          proc.wait()
          status=proc.returncode
        except:
          status=1
      elif cmd=='--clean':
        if len(self._cmdLine)!=2:
          sys.stderr.write('no argument expected for the `%s\' command\n'
                           % (cmd))
          status=1
        else:
          self._cleaning=True
          if not quiet:
            sys.stdout.write('Cleaning...\n')
          for i in self._generators+self._libraries+ \
                   self._plugins+self._programs:
            i.clean()
          self.remove(self._cleanList)
          self.remove('makefile')
          self._cleaning=False
      elif cmd=='--callback':
        if len(self._cmdLine)!=3:
          sys.stderr.write('one argument is expected for the `%s\' command\n'
                           % (cmd))
          status=1
        else:
          arg=self._cmdLine[2]
          target=self._findTarget(arg)
          if not target:
            sys.stderr.write('no generator for the target `%s\'\n'
                             % (arg))
            status=1
          if not status:
            status=target.callback()
      else:
        sys.stderr.write('unknown command `%s\'\n' % (cmd))
        status=1
      sys.exit(status)
    else:
      #----
      if self._system.startswith('Windows'):
        def cmdProtect(s):
          s2=s.replace('"','\\"')
          if (s2==s) and (s.find(' ')<0):
            return s
          else:
            return '"%s"' % (s2)
        def mkProtect(s):
          if s.find(' ')<0:
            return s
          else:
            return '"%s"' % (s)
        def mkEcho(msg=''):
          if self.testBoolOption('use_msc'):
            msg=msg.replace('%','^%')
          else:
            msg=msg.replace('%','%%')
          if msg:
            return '\t@echo %s\n' % (msg)
          else:
            return '\t@echo.\n'
      else: # posix system
        def cmdProtect(s):
          esc=['|','&',';','<','>','(',')','$','`','\\','"','\'',' ',
               '\t','\r','\n','*','?','~',]
          prot=False
          for i in s:
            if i in esc:
              prot=True
              break
          if prot:
            return '\'%s\'' % (s.replace('\'','\\\''))
          else:
            return s
        def mkProtect(s):
          return s.replace(' ','\\ ')
        def mkEcho(msg=''):
          return '\t@echo \'%s\'\n' % (msg.replace('\'','\\\''))
      #----
      if not quiet:
        sys.stdout.write('Generating makefile...\n')
      mk=open('makefile','w')
      mk.write('\n')
      mk.write('# This file is generated by %s\n' % (__name__))
      mk.write('# Do not edit or your changes will be lost!')
      mk.write('\n')
      targets=self._generators+self._libraries+self._plugins+self._programs
      if all is None:
        all=[i.name for i in targets]
      elif not isinstance(all,list):
        all=[all]
      hint=''
      if self._programs:
        path=[]
        for i in self._programs:
          for j in i.settings['LIBDIRS']:
            j=os.path.normpath(j)
            if j not in path:
              path.append(j)
        if path:
          path=os.path.pathsep.join(path)
          if self._system.startswith('Windows'):
            var='PATH'
            setCmd='set %s=%s'
            resetCmd='set %s=%%%s%%%s%s'
          else: # posix system
            if self._system.startswith('Darwin'):
              var='DYLD_LIBRARY_PATH'
            else:
              var='LD_LIBRARY_PATH'
            setCmd='export %s="%s"'
            resetCmd='export %s="$$%s%s%s"'
          if var in os.environ and os.environ[var]!='':
            hint=resetCmd % (var,var,os.pathsep,path)
          else:
            hint=setCmd % (var,path)
      mk.write('\n__ALL_DEFAULT_TARGETS__ : ')
      for i in all:
        mk.write('\\\n    %s ' % (mkProtect(os.path.normpath(i))))
      if hint:
        mk.write('\\\n    __DYNAMIC_LIB_HINT__')
        mk.write('\n')
        mk.write('\n__DYNAMIC_LIB_HINT__ :\n')
        mk.write(mkEcho())
        mk.write(mkEcho('If any dynamic library is not found at run-time,'))
        mk.write(mkEcho('you may try the following command:'))
        mk.write(mkEcho())
        mk.write(mkEcho('  %s' % (hint)))
        mk.write(mkEcho())
      mk.write('\n')
      doneList=[]
      for i in targets:
        if i in doneList:
          mk.write('\n# `%s\' already generated above\n\n' % (i))
        else:
          if not quiet:
            sys.stdout.write('    %s\n' % (i.name))
          i.generateMake(mk,cmdProtect,mkProtect,mkEcho)
          doneList.append(i)
          for j in i.sub:
            if j in doneList:
              mk.write('\n# `%s\' already generated above\n\n' % (j))
            else:
              if not quiet:
                sys.stdout.write('    %s\n' % (j))
              i.generateMake(mk,cmdProtect,mkProtect,mkEcho,j)
              doneList.append(j)
      mk.write('\nclean :\n')
      mk.write('\t@%s %s %s --clean\n'
               % (cmdProtect(self._python),
                  cmdProtect(self._cmdLine[0]),
                  ' '.join([cmdProtect(o) for o in self._optLine])))
      mk.write('\n.c.o :\n')
      mk.write(mkEcho('NO .c.o DEFAULT RULE FOR $<'))
      mk.write('\n.cpp.o :\n')
      mk.write(mkEcho('NO .cpp.o DEFAULT RULE FOR $<'))
      mk.write('\n')
      mk.close()
      if not quiet:
        sys.stdout.write('Python is `%s\' (%s)\n' %
                         (self._python,platform.python_version()))
        m=sys.modules[type(self).__module__]
        sys.stdout.write('%s is `%s\'\n' %
                         (m.__name__,os.path.realpath(m.__file__)))
        sys.stdout.write('System is `%s\'\n' % (self._system))
        sys.stdout.write('According to PATH:\n')
        make=self.settings['MAKE'][0]
        exe=self.findExecutable(make)
        if exe:
          sys.stdout.write('    the make program is `%s\'\n'
                           % (exe))
        else:
          sys.stdout.write('    the make program (%s) cannot be found\n'
                           % (make))
        cc=self.settings['CC'][0]
        exe=self.findExecutable(cc)
        if exe:
          sys.stdout.write('    the compiler is `%s\'\n'
                           % (exe))
        else:
          sys.stdout.write('    the compiler (%s) cannot be found\n'
                           % (cc))

  def remove(self,files):
    display=self._cleaning and not self.testBoolOption('quiet')
    if not isinstance(files,list):
      files=[files]
    for i in files:
      for j in glob.glob(i):
        try:
          if os.path.isdir(j):
            for k in os.listdir(j):
              self.remove(os.path.join(j,k))
            os.rmdir(j)
          else:
            os.unlink(j)
          if display:
            sys.stdout.write('    %s\n' % (j))
        except:
          pass

  def _findTarget(self,target,candidates=None):
    if candidates is None:
      candidates=self._generators+self._libraries+self._plugins+self._programs
    target=os.path.abspath(os.path.normpath(target))
    for i in candidates:
      if os.path.abspath(os.path.normpath(i.name))==target:
        return i
    return None

  def _main(self):
    (out,err)=self.showSettings(None)
    if err:
      for i in err:
        sys.stderr.write('%s\n' % (i))
      sys.exit(1)
    if out:
      sys.stdout.write('%s\n' % (' '.join(out)))

  _instance=None

#-----------------------------------------------------------------------------

class _Target(object):

  def __init__(self,tool,kind,name,sources,cb,cbArg,settings):
    object.__init__(self)
    self.tool=tool
    self.kind=kind
    self.name=name
    self.sources=[i for i in sources if sources.count(i)==1]
    self.sub=[]
    self.cb=cb
    self.cbArg=cbArg
    self.settings=settings

  def clean(self):
    if self.cb:
      self.cb(True,self.name,self.sources,self.cbArg,self.settings)
    cleanList=[]
    optimizerReport='.optrpt'
    for i in self.sub:
      cleanList.append(i)
      cleanList.append(os.path.splitext(i)[0]+optimizerReport)
    self.tool.remove(cleanList+[self.name])
    for i in self.sources:
      try:
        os.utime(i,None)
      except:
        pass

  def callback(self):
    if self.cb:
      return self.cb(False,self.name,self.sources,self.cbArg,self.settings)
    else:
      return 0 # success

  def writeMake(self,mk,cmdProtect,mkProtect,mkEcho,target,deps,cmds):
    mk.write('\n%s : ' % (mkProtect(target)))
    for i in deps:
      mk.write('\\\n    %s ' % (mkProtect(i)))
    mk.write('\n')
    mk.write(mkEcho())
    mk.write(mkEcho('==== %s ====' % (target)))
    for cmd in cmds:
      mk.write('\t%s\n' % (' '.join([cmdProtect(i) for i in cmd])))
    if (target==self.name) and self.cb:
      mk.write('\t@%s %s %s --callback %s\n'
               % (cmdProtect(self.tool._python),
                  cmdProtect(self.tool._cmdLine[0]),
                  ' '.join([cmdProtect(o) for o in self.tool._optLine]),
                  cmdProtect(target)))
    mk.write('\n')

  def generateMake(self,mk,cmdProtect,mkProtect,mkEcho,subTarget=None):
    if subTarget:
      target=subTarget
    else:
      target=self.name
    if subTarget:
      return # nothing to be done at this level of abstraction
    self.writeMake(mk,cmdProtect,mkProtect,mkEcho,target,self.sources,[])

#-----------------------------------------------------------------------------

class _BinTarget(_Target):

  def __init__(self,tool,kind,name,sources,cb,cbArg,settings):
    _Target.__init__(self,tool,kind,name,sources,cb,cbArg,settings)
    if self.tool.testBoolOption('use_msc'):
      objExt='.obj'
    else:
      objExt='.o'
    for s in self.sources:
      (name,ext)=os.path.splitext(s)
      ext=ext.lower()
      if ext in ['.c','.cpp','.cc','.cxx','.m']:
        self.sub.append(name+objExt)
      else:
        self.sub.append(s)

  def clean(self):
    _Target.clean(self)
    if self.tool.testBoolOption('use_msc'):
      self.tool.remove(self.name+'.manifest')
      if (self.kind=='plug') or (self.kind=='lib'):
        prefix=os.path.splitext(self.name)[0]
        self.tool.remove([prefix+'.lib',prefix+'.exp'])
    elif self.tool._system.startswith('Darwin'):
      if self.kind=='prog':
        prog=os.path.basename(self.name)
        d=os.path.dirname(self.name)
        if (os.path.basename(d)=='MacOS') and os.path.isdir(d):
          d=os.path.dirname(d)
          if (os.path.basename(d)=='Contents') and os.path.isdir(d):
            d=os.path.dirname(d)
            if (os.path.basename(d)==prog+'.app') and os.path.isdir(d):
              self.tool.remove(d)

  def generateMake(self,mk,cmdProtect,mkProtect,mkEcho,subTarget=None):
    if subTarget:
      target=subTarget
    else:
      target=self.name
    if subTarget:
      src=self.sources[self.sub.index(target)]
      if src!=target:
        deps=self.tool.collectCppDeps(src,self.settings)
        cmds=[self.tool.compilerCmd(target,src,self.settings)]
        self.writeMake(mk,cmdProtect,mkProtect,mkEcho,target,deps,cmds)
    else:
      deps=self.sub
      cmds=[]
      if self.tool._system.startswith('Darwin') and (self.kind=='prog'):
        cmds+=[['mkdir','-p',os.path.dirname(target)]]
      cmds+=[self.tool.linkerCmd(self.kind,target,self.sub,self.settings)]
      if self.tool.testBoolOption('use_msc'):
        if self.kind=='prog':
          resId=';1'
        else:
          resId=';2'
        cmds+=[['mt.exe','/nologo','/manifest',self.name+'.manifest',
                '/outputresource:'+self.name+resId]]
      self.writeMake(mk,cmdProtect,mkProtect,mkEcho,target,deps,cmds)

#-----------------------------------------------------------------------------

class _TemplateTarget(_Target):

  def __init__(self,tool,kind,name,sources,tmplArgs,cb,cbArg,settings):
    _Target.__init__(self,tool,kind,name,sources,cb,cbArg,settings)
    self.tmplArgs=tmplArgs

  def generateMake(self,mk,cmdProtect,mkProtect,mkEcho,subTarget=None):
    if subTarget:
      target=subTarget
    else:
      target=self.name
    if subTarget:
      return # nothing to be done
    deps=[]
    for i in self.sources: # may be redundant with target dependancies
      deps+=self.tool.collectCppDeps(i,self.settings)
    cmds=[[self.tool._python,self.sources[0],self.name]+
          self.sources[1:]+self.tmplArgs]
    self.writeMake(mk,cmdProtect,mkProtect,mkEcho,target,deps,cmds)

#-----------------------------------------------------------------------------

if __name__=='__main__':
  Tool()._main()

#-----------------------------------------------------------------------------
