//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
// 
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransGl/glView.h"

#include "TransProg/ptrVector.h"

#include <Cocoa/Cocoa.h>
#include <OpenGL/CGLDevice.h>

// FIXME: we should not have to use Carbon in a Cocoa application but
//        there is no Cocoa equivalent for GetCurrentEventButtonState()
#include <Carbon/Carbon.h>

// FIXME: no time to spend understanding retain/release/autorelease stuff
//        (it may leak ...)

#include "glView.c"

#if defined MAC_OS_X_VERSION_10_6
@interface CocoaView : NSView<NSWindowDelegate>
#else
@interface CocoaView : NSView
#endif
{
@public
GlView *glView;
bool destroyFlag;
bool forceInitialConfig;
}
@end

@interface CocoaViewManager : NSObject
{
@public
int initCpuCtx;
NSAutoreleasePool *pool;
CocoaView *commonView;
PtrVector views;
bool processingEvents;
int eventCount;
GfxStereoMode stereoMode;
NSCursor *cursors[10];
NSMenu *mainMenu;
bool manageMenu;
#if defined MAC_OS_X_VERSION_10_6
  CGLPixelFormatObj pixelFormat;
#else
  NSOpenGLPixelFormat *pixelFormat;
#endif
NSOpenGLContext *oglContext;
}
- (void)buildMenu;
@end

static CocoaViewManager *mgr;

//----------------------------------------------------------------------------

GfxStereoMode
GlView_stereoMode(void)
{
ASSERT_CURRENT_CPUCTX(mgr->initCpuCtx);
return mgr->stereoMode;
}

int // screen width
GlView_screenWidth(void)
{
ASSERT_CURRENT_CPUCTX(mgr->initCpuCtx);
return [[NSScreen mainScreen] visibleFrame].size.width;
}

int // screen height
GlView_screenHeight(void)
{
ASSERT_CURRENT_CPUCTX(mgr->initCpuCtx);
return [[NSScreen mainScreen] visibleFrame].size.height;
}

static
void
CocoaView_updateTitle(CocoaView *v)
{
ASSERT_CURRENT_CPUCTX(mgr->initCpuCtx);
[[v window] setTitle:[[NSString alloc] initWithUTF8String:v->glView->title]];
[mgr buildMenu];
}

static
void
CocoaView_updateCursor(CocoaView *v)
{
ASSERT_CURRENT_CPUCTX(mgr->initCpuCtx);
NSWindow *win=[v window];
[win invalidateCursorRectsForView:v];
#if 1
[win resetCursorRects];
#else
[win performSelector:@selector(resetCursorRects)
     withObject:nil
     afterDelay:0.0];
#endif
}

static
void
CocoaView_processEvent(CocoaView *v)
{
ASSERT_CURRENT_CPUCTX(mgr->initCpuCtx);
if(!mgr->processingEvents)
  {
  // DROP_DEAD("event processed outside of event loop");
  return; // FIXME: should enqueue event
  }
if(v->destroyFlag)
  {
  return;
  }
GlView *view=v->glView;
NSWindow *win=[v window];
NSRect cr=[win contentRectForFrameRect:[win frame]];
view->state.width=cr.size.width;
view->state.height=cr.size.height;
NSPoint p=[NSEvent mouseLocation];
view->state.mouseX=p.x-cr.origin.x;
view->state.mouseY=cr.origin.y+cr.size.height-p.y;
view->state.modifiers=GFX_NO_MDF;
int modifiers=[[NSApp currentEvent] modifierFlags];
if(modifiers&NSControlKeyMask)    view->state.modifiers|=GFX_CONTROL_MDF;
if(modifiers&NSShiftKeyMask)      view->state.modifiers|=GFX_SHIFT_MDF;
if(modifiers&NSAlphaShiftKeyMask) view->state.modifiers|=GFX_CAPSLOCK_MDF;
if(modifiers&NSNumericPadKeyMask) view->state.modifiers|=GFX_NUMLOCK_MDF;
int btnState=GetCurrentEventButtonState(); // FIXME: Carbon function!
if(btnState&(1<<0)) view->state.modifiers|=GFX_LEFT_MDF;
if(btnState&(1<<1)) view->state.modifiers|=GFX_RIGHT_MDF;
if(btnState&(1<<2)) view->state.modifiers|=GFX_MIDDLE_MDF;
view->state.visible=[win isVisible] ? true : false;
if(!(*view->callback)(view))
  {
  v->destroyFlag=true;
  }
++(mgr->eventCount);
}

int // events occured
GlView_processEvents(int64_t usecTimeout)
{
ASSERT_CURRENT_CPUCTX(mgr->initCpuCtx);
mgr->eventCount=0;
mgr->processingEvents=true;
for(;;)
  {
  NSDate *limit;
  if(usecTimeout)
    {
    limit=(usecTimeout>0)
          ? [NSDate dateWithTimeIntervalSinceNow: usecTimeout*1e-6]
          : [NSDate distantFuture];
    usecTimeout=0; // don't wait after first event
    }
  else
    {
    limit=[NSDate distantPast];
    }
  // FIXME: overflow warning when mask is NSAnyEventMask
  NSEvent *evt=[NSApp nextEventMatchingMask:NSUIntegerMax
                      untilDate:limit
                      inMode:NSDefaultRunLoopMode
                      dequeue:YES];
  if(!evt)
    {
    break;
    }
  [NSApp sendEvent:evt];
  }
mgr->processingEvents=false;
int destroyCount=0;
for(int i=PtrVector_size(&mgr->views);i--;)
  {
  CocoaView *v=(CocoaView *)PtrVector_get(&mgr->views,i);
  if(v->destroyFlag)
    {
    ++destroyCount;
    PtrVector_erase(&mgr->views,
                    PtrVector_find(&mgr->views,(PtrVector_CmpFnct)0,v));
    GlView *view=v->glView;
#if 0
    // FIXME: [NSOpenGLContext setView:] says 'invalid drawable' because
    //        the corresponding window has not been on the screen yet!
    [mgr->oglContext setView:mgr->commonView];
#endif
    [mgr->oglContext makeCurrentContext];
    NSWindow *win=[v window];
    [win close];
    [win release];
    [v releaseGState];
    [v release];
    Memory_free(view);
    }
  }
if(destroyCount)
  {
  [mgr buildMenu];
  }
return mgr->eventCount;
}

void
GlView_sync(void)
{
ASSERT_CURRENT_CPUCTX(mgr->initCpuCtx);
glFinish();
CGContextRef ctx=(CGContextRef)[[NSGraphicsContext currentContext]
                                graphicsPort];
CGContextFlush(ctx);
// FIXME: is this a XSync/GdiFlush equivalent?
}

static
void
CocoaView_open(GlView *view,
               int initX,
               int initY,
               bool borderless)
{
ASSERT_CURRENT_CPUCTX(mgr->initCpuCtx);
unsigned int style=borderless ? NSBorderlessWindowMask
                              : (NSTitledWindowMask|
                                 NSClosableWindowMask|
                                 NSMiniaturizableWindowMask|
                                 NSResizableWindowMask|
                                 NSTexturedBackgroundWindowMask);
NSRect screen=[[NSScreen mainScreen] visibleFrame];
NSRect fr=NSMakeRect(0,0,16,16);
if(view)
  {
  fr=NSMakeRect(screen.origin.x+initX,
                screen.origin.y+screen.size.height-view->state.height-initY,
                view->state.width,
                view->state.height);
  }
NSRect cr=[NSWindow contentRectForFrameRect:fr styleMask:style];
NSWindow *win=[[NSWindow alloc] initWithContentRect:cr
                                styleMask:style
                                backing:NSBackingStoreBuffered
                                defer:NO];
[win setFrameOrigin:fr.origin]; // ensure desired position
CocoaView *v=[[CocoaView alloc] initWithFrame:[win frame]];
v->glView=view;
v->destroyFlag=false;
v->forceInitialConfig=true;
[win setReleasedWhenClosed:NO];
[win setContentView:v];
[win setAcceptsMouseMovedEvents:YES];
[win makeFirstResponder:v];
[win setDelegate:v];
[win useOptimizedDrawing:YES];
[v allocateGState]; // try to speed-up lockFocus ...
if(view)
  {
  PtrVector_push(&mgr->views,v);
  view->specific=v;
  CocoaView_updateTitle(v);
  CocoaView_updateCursor(v);
  [win makeKeyAndOrderFront:nil];
  [win deminiaturize:nil];
  NSRect cr=[win contentRectForFrameRect:[win frame]];
  view->state.width=cr.size.width;
  view->state.height=cr.size.height;
  }
else
  {
  mgr->commonView=v;
  }
}

GlView * // created view
GlView_open(int initX,
            int initY,
            int initWidth,
            int initHeight,
            bool borderless,
            GlViewCallback callback,
            void *data)
{
ASSERT_CURRENT_CPUCTX(mgr->initCpuCtx);
GlView *view=Memory_alloc(1,GlView);
strcpy(view->title,"GlView");
view->cursor=GFX_ARROW_CSR;
GfxState_init(&view->state,initWidth,initHeight);
view->keyRepeat=true;
view->state.visible=false;
CocoaView_open(view,initX,initY,borderless);
view->callback=callback;
view->data=data;
return view;
}

void
GlView_setTitle(GlView *view,
                const char *title)
{
ASSERT_CURRENT_CPUCTX(mgr->initCpuCtx);
int maxLen=sizeof(view->title)-1;
strncpy(view->title,title,maxLen);
view->title[maxLen]='\0';
CocoaView_updateTitle((CocoaView *)view->specific);
}

void
GlView_setCursor(GlView *view,
                 GfxCursor cursor)
{
ASSERT_CURRENT_CPUCTX(mgr->initCpuCtx);
view->cursor=cursor;
CocoaView_updateCursor((CocoaView *)view->specific);
}

void
GlView_postRedraw(GlView *view)
{
ASSERT_CURRENT_CPUCTX(mgr->initCpuCtx);
CocoaView *v=(CocoaView *)view->specific;
[v setNeedsDisplay:YES];
}

void
GlView_makeCurrentGlTarget(GlView *view)
{
ASSERT_CURRENT_CPUCTX(mgr->initCpuCtx);
CocoaView *v=(CocoaView *)view->specific;
[mgr->oglContext setView:v];
[mgr->oglContext makeCurrentContext];
}

const intptr_t *
GlView_glContext(void)
{
static intptr_t params[2];
CGLContextObj cglCtx=[mgr->oglContext CGLContextObj];
CGLShareGroupObj cglShareGrp=CGLGetShareGroup(cglCtx);
params[0]=(intptr_t)cglCtx;
params[1]=(intptr_t)cglShareGrp;
return params;
}

//----------------------------------------------------------------------------

@implementation CocoaView

- (BOOL)isOpaque
{
return YES;
}

- (void)lockFocus
{
[super lockFocus];
[mgr->oglContext setView:self];
[mgr->oglContext makeCurrentContext];
}

- (void)drawRect:(NSRect)frameRect
{
(void)frameRect;
if(glView)
  {
  if(glView->callback) // avoid during opening
    {
    if(forceInitialConfig)
      {
      forceInitialConfig=false;
      glView->state.event=GFX_CONFIG_EVT;
      CocoaView_processEvent(self);
      }
    [mgr->oglContext update];
    glView->state.event=GFX_DRAW_EVT;
    CocoaView_processEvent(self);
    [mgr->oglContext flushBuffer]; // buffer swap with implicit glFlush()
    }
  else
    {
    GlView_postRedraw(glView); // FIXME: this does not postpone
    }
  }
}

- (void)mouseButtonEvent:(NSEvent *)evt
{
GfxButton btn=GFX_NO_BTN;
bool pressed=false;
switch([evt type])
  {
  case NSLeftMouseDown:
  case NSRightMouseDown:
  case NSOtherMouseDown:
    {
    pressed=true;
    break;
    }
  case NSLeftMouseUp:
  case NSRightMouseUp:
  case NSOtherMouseUp:
    {
    pressed=false;
    break;
    }
  case NSScrollWheel:
    {
    real64_t dy=[evt deltaY];
    if(!dy) return;
    pressed=true;
    btn=dy>0.0 ? GFX_UP_BTN : GFX_DOWN_BTN;
    break;
    }
  default: return;
  }
if(!btn)
  {
  switch([evt buttonNumber])
    {
    case 0: btn=GFX_LEFT_BTN;   break;
    case 1: btn=GFX_RIGHT_BTN;  break;
    case 2: btn=GFX_MIDDLE_BTN; break;
    default: return;
    }
  }
glView->state.event=GFX_BUTTON_EVT;
glView->state.button=btn;
glView->state.pressed=pressed;
CocoaView_processEvent(self);
}

- (void)mouseDown:(NSEvent *)evt      { [self mouseButtonEvent:evt]; }
- (void)rightMouseDown:(NSEvent *)evt { [self mouseButtonEvent:evt]; }
- (void)otherMouseDown:(NSEvent *)evt { [self mouseButtonEvent:evt]; }
- (void)mouseUp:(NSEvent *)evt        { [self mouseButtonEvent:evt]; }
- (void)rightMouseUp:(NSEvent *)evt   { [self mouseButtonEvent:evt]; }
- (void)otherMouseUp:(NSEvent *)evt   { [self mouseButtonEvent:evt]; }
- (void)scrollWheel:(NSEvent *)evt    { [self mouseButtonEvent:evt]; }

- (void)mouseMotionEvent:(NSEvent *)evt
{
(void)evt;
glView->state.event=GFX_MOTION_EVT;
CocoaView_processEvent(self);
}

- (void)mouseMoved:(NSEvent *)evt        { [self mouseMotionEvent:evt]; }
- (void)mouseDragged:(NSEvent *)evt      { [self mouseMotionEvent:evt]; }
- (void)rightMouseDragged:(NSEvent *)evt { [self mouseMotionEvent:evt]; }
- (void)otherMouseDragged:(NSEvent *)evt { [self mouseMotionEvent:evt]; }

- (void)keyboardEvent:(NSEvent *)evt
{
char buffer[0x20];
const char *key=(const char *)0;
bool pressed=false;
switch([evt type])
  {
  case NSKeyDown:
  case NSKeyUp:
    {
    if(!glView->keyRepeat&&[evt isARepeat]) return;
    NSString *str=[evt characters];
    if([str length]==1)
      {
      switch([str characterAtIndex: 0])
        {
        case NSUpArrowFunctionKey:    key="Up";     break;
        case NSDownArrowFunctionKey:  key="Down";   break;
        case NSLeftArrowFunctionKey:  key="Left";   break;
        case NSRightArrowFunctionKey: key="Right";  break;
        case NSPageUpFunctionKey:     key="Prior";  break;
        case NSPageDownFunctionKey:   key="Next";   break;
        case NSHomeFunctionKey:       key="Home";   break;
        case NSEndFunctionKey:        key="End";    break;
        case NSBeginFunctionKey:      key="Begin";  break;
        case NSInsertFunctionKey:     key="Insert"; break;
        case NSDeleteFunctionKey:     key="Delete"; break;
        case NSF1FunctionKey:         key="F1";     break;
        case NSF2FunctionKey:         key="F2";     break;
        case NSF3FunctionKey:         key="F3";     break;
        case NSF4FunctionKey:         key="F4";     break;
        case NSF5FunctionKey:         key="F5";     break;
        case NSF6FunctionKey:         key="F6";     break;
        case NSF7FunctionKey:         key="F7";     break;
        case NSF8FunctionKey:         key="F8";     break;
        case NSF9FunctionKey:         key="F9";     break;
        case NSF10FunctionKey:        key="F10";    break;
        case NSF11FunctionKey:        key="F11";    break;
        case NSF12FunctionKey:        key="F12";    break;
        }       
      }
    if(!key)
      {
      str=[evt charactersIgnoringModifiers];
      strcpy(buffer,[str UTF8String]);
      if([str length]==1)
        {
        switch(buffer[0])
          {
          case 0x7f: buffer[0]='\b'; break; // adjust BackSpace
          case '\r': buffer[0]='\n'; break; // adjust Enter
          case 0x03: buffer[0]='\n'; break; // adjust Numeric Enter
          } 
        }
      key=buffer;
      }
    pressed=[evt type]==NSKeyDown ? true : false;
    break;
    }
  }
if(!key)
  {
  key=""; // generate an empty key event in order to update modifiers
  }
glView->state.event=GFX_KEY_EVT;
int maxLen=sizeof(glView->state.key)-1;
strncpy(glView->state.key,key,maxLen);
glView->state.key[maxLen]='\0';
glView->state.pressed=pressed;
CocoaView_processEvent(self);
}

- (void)keyDown:(NSEvent *)evt      { [self keyboardEvent:evt]; }
- (void)keyUp:(NSEvent *)evt        { [self keyboardEvent:evt]; }
- (void)flagsChanged:(NSEvent *)evt { [self keyboardEvent:evt]; }

- (void)windowDidResize:(NSNotification *)notif
{
(void)notif;
glView->state.event=GFX_CONFIG_EVT;
CocoaView_processEvent(self);
}

- (void)windowDidMiniaturize:(NSNotification *)notif
{
(void)notif;
glView->state.event=GFX_CONFIG_EVT;
CocoaView_processEvent(self);
}

- (void)windowDidDeminiaturize:(NSNotification *)notif
{
(void)notif;
glView->state.event=GFX_CONFIG_EVT;
CocoaView_processEvent(self);
}

- (void)windowDidBecomeKey:(NSNotification *)notif
{
(void)notif;
for(int i=0,i_end=PtrVector_size(&mgr->views);i<i_end;++i)
  {
  CocoaView *v=(CocoaView *)PtrVector_get(&mgr->views,i);
  if(v!=self)
    {
    // avoid reporting motion events in many windows at a time
    [[v window] resignKeyWindow];
    }
  }
}

- (BOOL)windowShouldClose:(id)win
{
(void)win;
glView->state.event=GFX_CLOSE_EVT;
CocoaView_processEvent(self);
return NO; // will be closed afterward if requested by the callback
}

- (void)resetCursorRects
{
[self addCursorRect:[self visibleRect]
      cursor:mgr->cursors[glView->cursor]];
}

@end

@implementation CocoaViewManager

- (void)menuAction:(id)sender
{ 
NSMenu *appMenu=[[mainMenu itemAtIndex:0] submenu];
int count=[appMenu numberOfItems];
int index=[appMenu indexOfItem:sender];
if(index==count-1)
  {
  [NSApp terminate:nil];
  }
else if(index<PtrVector_size(&views))
  {
  CocoaView *v=(CocoaView *)PtrVector_get(&views,index);
  [[v window] makeKeyAndOrderFront:nil];
  }
}

- (void)buildMenu
{
if(!manageMenu) return;
if(mainMenu) [mainMenu release];
mainMenu=[[NSMenu alloc] initWithTitle:@"MainMenuName"];
NSMenuItem *appMenuItem=[mainMenu addItemWithTitle:@"AppMenuName"
                                  action:nil keyEquivalent:@""];
NSMenu *appMenu=[[NSMenu alloc]
                 initWithTitle:NSLocalizedString(@"Action",nil)];
// Ugly hack: Using the private/undocumented [NSApp setAppleMenu: appMenu]
// method allows to set the content of the main application submenu
// (on the left, right after the Apple logo).
// Without this hack, the main application submenu is left empty
// (because we don't use any NIB file) and our action submenu appears
// right after.
[NSApp performSelector:@selector(setAppleMenu:) withObject:appMenu];
int viewCount=PtrVector_size(&views);
if(viewCount) 
  {   
  for(int i=0;i<viewCount;++i)
    { 
    CocoaView *v=(CocoaView *)PtrVector_get(&views,i);
    [[appMenu addItemWithTitle:[[v window] title]
              action:@selector(menuAction:)
              keyEquivalent:@""] setTarget:self];
    }
  [appMenu addItem:[NSMenuItem separatorItem]];
  }
[[appMenu addItemWithTitle:NSLocalizedString(@"Quit",nil)
          action:@selector(menuAction:)
          keyEquivalent:@"q"] setTarget:self];
[mainMenu setSubmenu:appMenu forItem:appMenuItem];
[NSApp setMainMenu:mainMenu];
}

@end

//----------------------------------------------------------------------------

#if defined MAC_OS_X_VERSION_10_6
  static
  CGLPixelFormatObj
  CocoaView_choosePixelFormat(bool doubleBuffer,
                              bool stereo,
                              bool accum,
                              bool stencil)
  {
  CGLPixelFormatAttribute attribs[0x20];
  int attribCount=0;
  attribs[attribCount++]=kCGLPFAColorSize;
  attribs[attribCount++]=1;
  attribs[attribCount++]=kCGLPFAAlphaSize;
  attribs[attribCount++]=1;
  attribs[attribCount++]=kCGLPFADepthSize;
  attribs[attribCount++]=1;
  if(doubleBuffer)
    {
    attribs[attribCount++]=kCGLPFADoubleBuffer;
    }
  if(stereo)
    {
    attribs[attribCount++]=kCGLPFAStereo;
    }
  if(accum)
    {
    attribs[attribCount++]=kCGLPFAAccumSize;
    attribs[attribCount++]=1;
    }
  if(stencil)
    {
    attribs[attribCount++]=kCGLPFAStencilSize;
    attribs[attribCount++]=1;
    }
  // FIXME: handle OpenGL3 contexts
  // attribs[attribCount++]=kCGLPFAOpenGLProfile;
  // attribs[attribCount++]=kCGLOGLPVersion_3_2_Core;
  attribs[attribCount++]=0;
  CGLPixelFormatObj pixelFormat;
  int npix;
  CGLChoosePixelFormat(attribs,&pixelFormat,&npix);
  if(!pixelFormat)
    {
    DROP_DEAD("CGLPixelFormat failure");
    }
  return pixelFormat;
  }
#else
  static
  NSOpenGLPixelFormat *
  CocoaView_choosePixelFormat(bool doubleBuffer,
                              bool stereo,
                              bool accum,
                              bool stencil)
  {
  NSOpenGLPixelFormatAttribute attribs[0x20];
  int attribCount=0;
  attribs[attribCount++]=NSOpenGLPFAColorSize;
  attribs[attribCount++]=1;
  attribs[attribCount++]=NSOpenGLPFAAlphaSize;
  attribs[attribCount++]=1;
  attribs[attribCount++]=NSOpenGLPFADepthSize;
  attribs[attribCount++]=1;
  if(doubleBuffer)
    {
    attribs[attribCount++]=NSOpenGLPFADoubleBuffer;
    }
  if(stereo)
    {
    attribs[attribCount++]=NSOpenGLPFAStereo;
    }
  if(accum)
    {
    attribs[attribCount++]=NSOpenGLPFAAccumSize;
    attribs[attribCount++]=1;
    }
  if(stencil)
    {
    attribs[attribCount++]=NSOpenGLPFAStencilSize;
    attribs[attribCount++]=1;
    }
  attribs[attribCount++]=0;
  NSOpenGLPixelFormat *pixelFormat=[[NSOpenGLPixelFormat alloc]
                                    initWithAttributes:attribs];
  if(!pixelFormat)
    {
    DROP_DEAD("NSOpenGLPixelFormat failure");
    }
  return pixelFormat;
  }
#endif

void
GlView_init(void)
{
if(!mgr)
  {
  (void)extSuffixes; // avoid ``unused variable'' warning
  (void)GlView_detectGlVersion; // avoid ``unused function'' warning
  NSAutoreleasePool *pool=[[NSAutoreleasePool alloc] init];
  mgr=[[CocoaViewManager alloc] init];
  mgr->pool=pool;
  mgr->initCpuCtx=Platform_cpuCtx();
  PtrVector_init(&mgr->views);
  mgr->processingEvents=false;
  mgr->eventCount=0;
  mgr->stereoMode=GlView_detectStereoMode();
  mgr->mainMenu=(NSMenu *)0;
  if(NSApp)
    {
    mgr->manageMenu=false;
    }
  else
    {
    mgr->manageMenu=true;
    [NSApplication sharedApplication];
    [NSApp finishLaunching];
    }
  NSImage *blankImg=[[NSImage alloc] initWithSize:NSMakeSize(1,1)];
  mgr->cursors[GFX_NO_CSR]=[[NSCursor alloc] initWithImage:blankImg
                                             hotSpot:NSMakePoint(0,0)];
  mgr->cursors[GFX_ARROW_CSR]=[NSCursor arrowCursor];
  mgr->cursors[GFX_TEXT_CSR]=[NSCursor IBeamCursor];
  mgr->cursors[GFX_HAND_CSR]=[NSCursor pointingHandCursor];
  mgr->cursors[GFX_CROSS_CSR]=[NSCursor crosshairCursor];
  mgr->cursors[GFX_WAIT_CSR]=[NSCursor disappearingItemCursor];
  mgr->cursors[GFX_NORTH_CSR]=[NSCursor resizeUpCursor];
  mgr->cursors[GFX_SOUTH_CSR]=[NSCursor resizeDownCursor];
  mgr->cursors[GFX_EAST_CSR]=[NSCursor resizeRightCursor];
  mgr->cursors[GFX_WEST_CSR]=[NSCursor resizeLeftCursor];
  CocoaView_open((GlView *)0,0,0,false);
  bool doubleBuffer=true;
  bool stereo=mgr->stereoMode==GFX_ACTIVE_STEREO ? true : false;
  bool accum=false;
  bool stencil=false;
  mgr->pixelFormat=CocoaView_choosePixelFormat(doubleBuffer,stereo,
                                               accum,stencil);
#if defined MAC_OS_X_VERSION_10_6
  CGLContextObj ctx;
  CGLCreateContext(mgr->pixelFormat,(CGLContextObj)0,&ctx);
  mgr->oglContext=[[NSOpenGLContext alloc]
                   initWithCGLContextObj:ctx];
#else
  mgr->oglContext=[[NSOpenGLContext alloc]
                   initWithFormat:mgr->pixelFormat
                   shareContext:(NSOpenGLContext *)nil];
#endif
  if(!mgr->oglContext)
    {
    DROP_DEAD("NSOpenGLContext failure");
    }
#if 0
  // FIXME: [NSOpenGLContext setView:] says 'invalid drawable' because
  //        the corresponding window has not been on the screen yet!
  [mgr->oglContext setView:mgr->commonView];
#endif
  [mgr->oglContext makeCurrentContext];
  CoreGl_init();
  if(mgr->stereoMode==GFX_ACTIVE_STEREO)
    {
    int glIsStereo=0;
    glGetIntegerv(GL_STEREO,&glIsStereo);
    if(!glIsStereo)
      {
      DROP_DEAD("OpenGL does not support active stereo");
      }
    }
  [mgr buildMenu];
  }
}

//----------------------------------------------------------------------------

const char * // specific OpenGL extensions
_CoreGl_querySpecificExtensions(void)
{
const char *extensions=(const char *)0;
// FIXME: isn't there any glXQueryExtensionsString/wglGetExtensionsString
//        equivalent?
return extensions;
}

//----------------------------------------------------------------------------
