//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransGl/glView.h"

#include "TransProg/ptrVector.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>
#define __gl_h_ // avoid inclusion of GL/gl.h
#include <GL/glx.h>

#include "glView.c"

#if !defined GLX_CONTEXT_MAJOR_VERSION_ARB
# define GLX_CONTEXT_MAJOR_VERSION_ARB 0x2091
#endif

#if !defined GLX_CONTEXT_MINOR_VERSION_ARB
# define GLX_CONTEXT_MINOR_VERSION_ARB 0x2092
#endif

#if !defined GLX_CONTEXT_FLAGS_ARB
# define GLX_CONTEXT_FLAGS_ARB 0x2094
#endif

#if !defined GLX_CONTEXT_PROFILE_MASK_ARB
# define GLX_CONTEXT_PROFILE_MASK_ARB 0x9126
#endif

#if !defined GLX_CONTEXT_DEBUG_BIT_ARB
# define GLX_CONTEXT_DEBUG_BIT_ARB 0x0001
#endif

#if !defined GLX_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB
# define GLX_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB 0x0002
#endif

#if !defined GLX_CONTEXT_CORE_PROFILE_BIT_ARB
# define GLX_CONTEXT_CORE_PROFILE_BIT_ARB 0x00000001
#endif

#if !defined GLX_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB
# define GLX_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB 0x00000002
#endif

static
GLXContext
(APIENTRY *createContextAttribsFnct)(Display *,
                                     void *,
                                     GLXContext,
                                     Bool,
                                     const int *);

typedef struct _X11View
{
GlView *glView;
Window window;
bool destroyFlag;
} X11View;

typedef struct _X11ViewManager
{
bool initDone;
int initCpuCtx;
X11View *commonView;
PtrVector views;
int eventCount;
GfxStereoMode stereoMode;
Display *display;
Cursor cursors[10];
GLXFBConfig fbConfig;
XVisualInfo *visual;
Colormap colormap;
GLXContext glxContext;
} X11ViewManager;

static X11ViewManager mgr;

//----------------------------------------------------------------------------

GfxStereoMode 
GlView_stereoMode(void)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
return mgr.stereoMode;
}

int // screen width
GlView_screenWidth(void)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
return DisplayWidth(mgr.display,DefaultScreen(mgr.display));
}

int // screen height
GlView_screenHeight(void)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
return DisplayHeight(mgr.display,DefaultScreen(mgr.display));
}

static
void
X11View_updateTitle(X11View *v)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
char *str=v->glView->title;
XTextProperty textProperty;
XStringListToTextProperty(&str,1,&textProperty);
XSetWMName(mgr.display,v->window,&textProperty);
XSetWMIconName(mgr.display,v->window,&textProperty);
XFree(textProperty.value);
}

static
void
X11View_updateCursor(X11View *v)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
XDefineCursor(mgr.display,v->window,mgr.cursors[v->glView->cursor]);
}

static
void
X11View_processEvent(X11View *v)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
if(v->destroyFlag)
  {
  return;
  }
GlView *view=v->glView;
XWindowAttributes attr;
XGetWindowAttributes(mgr.display,v->window,&attr);
view->state.width=attr.width;
view->state.height=attr.height;
Window r,c;
int sx,sy;
unsigned int modMask;
if(XQueryPointer(mgr.display,v->window,&r,&c,
                 &sx,&sy,&view->state.mouseX,&view->state.mouseY,
                 &modMask))
  {
  view->state.modifiers=GFX_NO_MDF;
  if(modMask&ControlMask) view->state.modifiers|=GFX_CONTROL_MDF;
  if(modMask&ShiftMask)   view->state.modifiers|=GFX_SHIFT_MDF;
  if(modMask&LockMask)    view->state.modifiers|=GFX_CAPSLOCK_MDF;
  if(modMask&Mod2Mask)    view->state.modifiers|=GFX_NUMLOCK_MDF;
  if(modMask&Button1Mask) view->state.modifiers|=GFX_LEFT_MDF;
  if(modMask&Button2Mask) view->state.modifiers|=GFX_MIDDLE_MDF;
  if(modMask&Button3Mask) view->state.modifiers|=GFX_RIGHT_MDF;
  }
if(!(*view->callback)(view))
  {
  v->destroyFlag=true;
  }
++mgr.eventCount;
}

int // events occured
GlView_processEvents(int64_t usecTimeout)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
mgr.eventCount=0;
if(!XEventsQueued(mgr.display,QueuedAlready)&&usecTimeout)
  {
  struct timeval tv={0,0};
  struct timeval *timeout=(struct timeval *)0;
  if(usecTimeout>0)
    {
    tv.tv_sec=(time_t)(usecTimeout/1000000);
    tv.tv_usec=(int)(usecTimeout%1000000);
    timeout=&tv;
    }
  fd_set rdSet;
  FD_ZERO(&rdSet);
  int xFd=ConnectionNumber(mgr.display);
  FD_SET(xFd,&rdSet);
  int r;
  RESTART_SYSCALL(r,select(xFd+1,&rdSet,(fd_set *)0,(fd_set *)0,timeout));
  if(r==-1)
    {
    SYS_DROP_DEAD("select() failure");
    }
  }
while(XPending(mgr.display))
  {
  XEvent evt;
  XNextEvent(mgr.display,&evt);
  GlView *view=(GlView *)0;
  for(int i=0,i_end=PtrVector_size(&mgr.views);i<i_end;++i)
    {
    X11View *v=(X11View *)PtrVector_get(&mgr.views,i);
    if(v->window==evt.xany.window)
      {
      view=v->glView;
      break;
      }
    }
  if(view)
    {
    X11View *v=(X11View *)view->specific;
    switch(evt.type)
      {
      case MotionNotify:
        {
        while(XCheckTypedWindowEvent(mgr.display,v->window,MotionNotify,&evt));
        view->state.event=GFX_MOTION_EVT;
        X11View_processEvent(v);
        break;
        }
      case Expose:
        {
        if(!glXMakeCurrent(mgr.display,v->window,mgr.glxContext))
          {
          DROP_DEAD("glXMakeCurrent() failure");
          }
        view->state.event=GFX_DRAW_EVT;
        X11View_processEvent(v);
        glXSwapBuffers(mgr.display,v->window); // implicit glFlush()
        while(XCheckTypedWindowEvent(mgr.display,v->window,Expose,&evt))
          {
          // OpenGL rendering seems to produce Expose events!
          }
        break;
        }
      case ButtonPress:
      case ButtonRelease:
        {
        GfxButton btn=GFX_NO_BTN;
        switch(evt.xbutton.button)
          {
          case Button1: btn=GFX_LEFT_BTN;                           break;
          case Button2: btn=GFX_MIDDLE_BTN;                         break;
          case Button3: btn=GFX_RIGHT_BTN;                          break;
          case Button4: if(evt.type==ButtonPress) btn=GFX_UP_BTN;   break;
          case Button5: if(evt.type==ButtonPress) btn=GFX_DOWN_BTN; break;
          }
        if(btn)
          {
          view->state.event=GFX_BUTTON_EVT;
          view->state.button=btn;
          view->state.pressed=evt.type==ButtonPress ? true : false;
          X11View_processEvent(v);
          }
        break;
        }
      case KeyPress:
      case KeyRelease:
        {
        static XComposeStatus compose={ (XPointer)0, 0 };
        if(!view->keyRepeat&&(evt.type==KeyRelease)&&XPending(mgr.display))
          {
          XEvent nextEvent;
          XPeekEvent(mgr.display,&nextEvent);
          if((nextEvent.type==KeyPress)&&
             (nextEvent.xkey.window==evt.xkey.window)&&
             (nextEvent.xkey.keycode==evt.xkey.keycode)&&
             (nextEvent.xkey.time==evt.xkey.time))
            {
            XNextEvent(mgr.display,&nextEvent);
            break; // ignore the KeyRelease/KeyPress pair
            }
          }
        char buffer[0x20];
        evt.xkey.state&=~ControlMask; // ignore [Control]
        KeySym keysym;
        int count=XLookupString(&evt.xkey,buffer,0x20,
                                &keysym,&compose);
        if((keysym==XK_Delete)&&   // Sometimes BackSpace is converted
           (evt.xkey.keycode==22)) // to Delete keycode
          {
          keysym=XK_BackSpace;
          buffer[0]='\b';
          count=1;
          }
        const char *key=(const char *)0;
        switch(keysym)
          {
          case XK_Up:
          case XK_KP_Up:     key="Up";       break;
          case XK_Down:
          case XK_KP_Down:   key="Down";     break;
          case XK_Left:
          case XK_KP_Left:   key="Left";     break;
          case XK_Right:
          case XK_KP_Right:  key="Right";    break;
          case XK_Prior:
          case XK_KP_Prior:  key="Prior";    break;
          case XK_Next:
          case XK_KP_Next:   key="Next";     break;
          case XK_Home:
          case XK_KP_Home:   key="Home";     break;
          case XK_End:
          case XK_KP_End:    key="End";      break;
          case XK_Begin:
          case XK_KP_Begin:  key="Begin";    break;
          case XK_Insert:
          case XK_KP_Insert: key="Insert";   break;
          case XK_Delete:
          case XK_KP_Delete: key="Delete";   break;
          case XK_F1:        key="F1";       break;
          case XK_F2:        key="F2";       break;
          case XK_F3:        key="F3";       break;
          case XK_F4:        key="F4";       break;
          case XK_F5:        key="F5";       break;
          case XK_F6:        key="F6";       break;
          case XK_F7:        key="F7";       break;
          case XK_F8:        key="F8";       break;
          case XK_F9:        key="F9";       break;
          case XK_F10:       key="F10";      break;
          case XK_F11:       key="F11";      break;
          case XK_F12:       key="F12";      break;
          }
        if(!key&&(count==1))
          {
          if(buffer[0]=='\r') buffer[0]='\n';
          buffer[1]='\0';
          key=buffer;
          }
        if(!key)
          {
          key=""; // generate an empty key event in order to update modifiers
          }
        view->state.event=GFX_KEY_EVT;
        int maxLen=sizeof(view->state.key)-1;
        strncpy(view->state.key,key,maxLen);
        view->state.key[maxLen]='\0';
        view->state.pressed=evt.type==KeyPress ? true : false;
        X11View_processEvent(v);
        break;
        }
      case ConfigureNotify:
        {
        while(XCheckTypedWindowEvent(mgr.display,v->window,
                                     ConfigureNotify,&evt));
        view->state.event=GFX_CONFIG_EVT;
        X11View_processEvent(v);
        break;
        }
      case MapNotify:
        {
        view->state.event=GFX_CONFIG_EVT;
        view->state.visible=true;
        X11View_processEvent(v);
        break;
        }
      case UnmapNotify:
        {
        view->state.event=GFX_CONFIG_EVT;
        view->state.visible=false;
        X11View_processEvent(v);
        break;
        }
      case VisibilityNotify:
        {
        view->state.event=GFX_CONFIG_EVT;
        view->state.visible=evt.xvisibility.state==VisibilityFullyObscured
                      ? false : true;
        X11View_processEvent(v);
        break;
        }
      case ClientMessage:
        {
        char *atomName=XGetAtomName(mgr.display,evt.xclient.data.l[0]);
        if(!strcmp(atomName,"WM_DELETE_WINDOW"))
          {
          view->state.event=GFX_CLOSE_EVT;
          X11View_processEvent(v);
          }
        XFree(atomName);
        break;
        }
      }
    }
  }
for(int i=PtrVector_size(&mgr.views);i--;)
  {
  X11View *v=(X11View *)PtrVector_get(&mgr.views,i);
  if(v->destroyFlag)
    {
    // FIXME: why not use i?
    PtrVector_erase(&mgr.views,
                    PtrVector_find(&mgr.views,(PtrVector_CmpFnct)0,v));
    GlView *view=v->glView;
    if(!glXMakeCurrent(mgr.display,mgr.commonView->window,mgr.glxContext))
      {
      DROP_DEAD("glXMakeCurrent() failure");
      }
    XDestroyWindow(mgr.display,v->window);
    Memory_free(v);
    Memory_free(view);
    }
  }
return mgr.eventCount;
}

void
GlView_sync(void)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
glFinish();
XSync(mgr.display,False);
}

static
void
X11View_open(GlView *view,
             int initX,
             int initY,
             bool borderless)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
X11View *v=Memory_alloc(1,X11View);
v->glView=view;
v->destroyFlag=false;
XSetWindowAttributes wAttr;
wAttr.colormap=mgr.colormap;
wAttr.background_pixel=WhitePixel(mgr.display,DefaultScreen(mgr.display));
wAttr.border_pixel=BlackPixel(mgr.display,DefaultScreen(mgr.display));
v->window=XCreateWindow(mgr.display,DefaultRootWindow(mgr.display),
                        initX,initY,
                        view ? view->state.width : 16,
                        view ? view->state.height : 16,0,
                        mgr.visual->depth,InputOutput,mgr.visual->visual,
                        CWColormap|CWBackPixel|CWBorderPixel,&wAttr);
if(!v->window)
  {
  DROP_DEAD("XCreateWindow() failure");
  }
if(view)
  {
  PtrVector_push(&mgr.views,v);
  view->specific=v;
  X11View_updateTitle(v);
  X11View_updateCursor(v);
  XSelectInput(mgr.display,v->window,
               StructureNotifyMask|VisibilityChangeMask|ExposureMask|
               PointerMotionMask|ButtonPressMask|ButtonReleaseMask|
               KeyPressMask|KeyReleaseMask);
  Atom protocol=XInternAtom(mgr.display,"WM_DELETE_WINDOW",False);
  XSetWMProtocols(mgr.display,v->window,&protocol,1);
  if(borderless)
    {
    Atom ha=XInternAtom(mgr.display,"_MOTIF_WM_HINTS",False);
    long hints[5]={ (1L<<1), // flags: MWM_HINTS_DECORATIONS
                    0L,      // functions:
                    0L,      // decorations: no decoration!
                    0L,      // inputmode:
                    0L };    // status:
    XChangeProperty(mgr.display,v->window,ha,ha,32,PropModeReplace,
                    (unsigned char *)hints,5);
    }
  XMapRaised(mgr.display,v->window);
  bool reparented=true;
  XEvent evt;
  while(!XCheckTypedWindowEvent(mgr.display,v->window,ReparentNotify,&evt))
    {
    if(XCheckTypedWindowEvent(mgr.display,v->window,Expose,&evt))
      {
      XPutBackEvent(mgr.display,&evt);
      reparented=false; // exposed without reparenting
      break;
      }
    }
  if(reparented) // try to fix the window manager offset
    {
    Window top,root,child,parent=v->window;
    Window *children;
    unsigned int childCount;
    do // find top window
      {
      top=parent;
      XQueryTree(mgr.display,top,&root,&parent,&children,&childCount);
      if(children) XFree(children);
      } while(parent!=DefaultRootWindow(mgr.display));
    // adjust x and y offset
    int rootX,rootY,topX,topY,winX,winY;
    unsigned int mask;
    XQueryPointer(mgr.display,top,&root,&child,
                  &rootX,&rootY,&topX,&topY,&mask);
    topX=rootX-topX;
    topY=rootY-topY;
    XQueryPointer(mgr.display,v->window,&root,&child,
                  &rootX,&rootY,&winX,&winY,&mask);
    winX=rootX-winX;
    winY=rootY-winY;
    initX+=winX-topX;
    initY+=winY-topY;
    // adjust width and height offset
    XWindowAttributes winAttr,topAttr;
    XGetWindowAttributes(mgr.display,top,&topAttr);
    XGetWindowAttributes(mgr.display,v->window,&winAttr);
    int dw=topAttr.width-winAttr.width;
    int dh=topAttr.height-winAttr.height;
    if(dw<view->state.width) view->state.width-=dw;
    if(dh<view->state.height) view->state.height-=dh;
    // apply new geometry
    XMoveResizeWindow(mgr.display,v->window,
                      initX,initY,view->state.width,view->state.height);
    }
  XWindowAttributes attr;
  XGetWindowAttributes(mgr.display,v->window,&attr);
  view->state.width=attr.width;
  view->state.height=attr.height;
  }
else
  {
  mgr.commonView=v;
  }
}

GlView * // created view
GlView_open(int initX,
            int initY,
            int initWidth,
            int initHeight,
            bool borderless,
            GlViewCallback callback,
            void *data)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
GlView *view=Memory_alloc(1,GlView);
strcpy(view->title,"GlView");
view->cursor=GFX_ARROW_CSR;
GfxState_init(&view->state,initWidth,initHeight);
view->keyRepeat=true;
X11View_open(view,initX,initY,borderless);
view->callback=callback;
view->data=data;
return view;
}

void
GlView_setTitle(GlView *view,
                 const char *title)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
int maxLen=(int)sizeof(view->title)-1;
strncpy(view->title,title,maxLen);
view->title[maxLen]='\0';
X11View_updateTitle((X11View *)view->specific);
}

void
GlView_setCursor(GlView *view,
                  GfxCursor cursor)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
view->cursor=cursor;
X11View_updateCursor((X11View *)view->specific);
}

void
GlView_postRedraw(GlView *view)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
X11View *v=(X11View *)view->specific;
XEvent evt;
evt.xexpose.type=Expose;
evt.xexpose.serial=0;
evt.xexpose.send_event=True;
evt.xexpose.display=mgr.display;
evt.xexpose.window=v->window;
evt.xexpose.x=0;
evt.xexpose.y=0;
evt.xexpose.width=view->state.width;
evt.xexpose.height=view->state.height;
evt.xexpose.count=0;
if(!XSendEvent(mgr.display,v->window,False,0,&evt))
  {
  DROP_DEAD("XSendEvent() failure");
  }
}

void
GlView_makeCurrentGlTarget(GlView *view)
{
ASSERT_CURRENT_CPUCTX(mgr.initCpuCtx);
X11View *v=(X11View *)view->specific;
if(!glXMakeCurrent(mgr.display,v->window,mgr.glxContext))
  {
  DROP_DEAD("glXMakeCurrent() failure");
  }
}

const intptr_t *
GlView_glContext(void)
{
static intptr_t params[2];
params[0]=(intptr_t)mgr.display;
params[1]=(intptr_t)mgr.glxContext;
return params;
}

//----------------------------------------------------------------------------

// FIXME: ugly hack!
// Even in C, the ``class'' member generates a warning!
// --> workaround accessing the ``class'' member right after ``depth''
#define XVisualInfo_class(info) \
        ((&((info)->depth))[1])

static
bool // success
checkVisual(Display *display,
            XVisualInfo *info,
            bool doubleBuffer,
            bool stereo,
            bool accum,
            bool stencil)
{
if((info->screen!=DefaultScreen(display))||
   ((XVisualInfo_class(info)!=TrueColor)&&
    (XVisualInfo_class(info)!=DirectColor))||
   (!info->red_mask)||(!info->green_mask)||(!info->blue_mask))
  {
  return false;
  }
int value;
if(glXGetConfig(display,info,GLX_USE_GL,&value)||!value)
  {
  return false;
  }
if(glXGetConfig(display,info,GLX_RGBA,&value)||!value)
  {
  return false;
  }
if(glXGetConfig(display,info,GLX_RED_SIZE,&value)||!value)
  {
  return false;
  }
if(glXGetConfig(display,info,GLX_GREEN_SIZE,&value)||!value)
  {
  return false;
  }
if(glXGetConfig(display,info,GLX_BLUE_SIZE,&value)||!value)
  {
  return false;
  }
#if 0 // sometimes value is 0 even if alpha is supported!
  if(glXGetConfig(display,info,GLX_ALPHA_SIZE,&value)||!value)
    {
    return false;
    }
#endif
if(glXGetConfig(display,info,GLX_DEPTH_SIZE,&value)||!value)
  {
  return false;
  }
if(doubleBuffer)
  {
  if(glXGetConfig(display,info,GLX_DOUBLEBUFFER,&value)||!value)
    {
    return false;
    }
  }
else
  {
  if(glXGetConfig(display,info,GLX_DOUBLEBUFFER,&value)||value)
    {
    return false;
    }
  }
if(stereo)
  {
  if(glXGetConfig(display,info,GLX_STEREO,&value)||!value)
    {
    return false;
    }
  }
else
  {
  if(glXGetConfig(display,info,GLX_STEREO,&value)||value)
    {
    return false;
    }
  }
if(accum)
  {
  if(glXGetConfig(display,info,GLX_ACCUM_RED_SIZE,&value)||!value)
    {
    return false;
    }
  if(glXGetConfig(display,info,GLX_ACCUM_GREEN_SIZE,&value)||!value)
    {
    return false;
    }
  if(glXGetConfig(display,info,GLX_ACCUM_BLUE_SIZE,&value)||!value)
    {
    return false;
    }
  if(glXGetConfig(display,info,GLX_ACCUM_ALPHA_SIZE,&value)||!value)
    {
    return false;
    }
  }
if(stencil)
  {
  if(glXGetConfig(display,info,GLX_STENCIL_SIZE,&value)||!value)
    {
    return false;
    }
  }
return true;
}

static
XVisualInfo *
chooseVisual(Display *display,
             bool doubleBuffer,
             bool stereo,
             bool accum,
             bool stencil)
{
int attribs[0x20];
int attribCount=0;
attribs[attribCount++]=GLX_RGBA;
attribs[attribCount++]=GLX_RED_SIZE;
attribs[attribCount++]=1;
attribs[attribCount++]=GLX_GREEN_SIZE;
attribs[attribCount++]=1;
attribs[attribCount++]=GLX_BLUE_SIZE;
attribs[attribCount++]=1;
attribs[attribCount++]=GLX_ALPHA_SIZE;
attribs[attribCount++]=1;
attribs[attribCount++]=GLX_DEPTH_SIZE;
attribs[attribCount++]=1;
if(doubleBuffer)
  {
  attribs[attribCount++]=GLX_DOUBLEBUFFER;
  }
if(stereo)
  {
  attribs[attribCount++]=GLX_STEREO;
  }
if(accum)
  {
  attribs[attribCount++]=GLX_ACCUM_RED_SIZE;
  attribs[attribCount++]=1;
  attribs[attribCount++]=GLX_ACCUM_GREEN_SIZE;
  attribs[attribCount++]=1;
  attribs[attribCount++]=GLX_ACCUM_BLUE_SIZE;
  attribs[attribCount++]=1;
  attribs[attribCount++]=GLX_ACCUM_ALPHA_SIZE;
  attribs[attribCount++]=1;
  }
if(stencil)
  {
  attribs[attribCount++]=GLX_STENCIL_SIZE;
  attribs[attribCount++]=1;
  }
attribs[attribCount++]=None;
XVisualInfo *best=glXChooseVisual(display,DefaultScreen(display),attribs);
if(best)
  {
  if(!checkVisual(display,best,doubleBuffer,stereo,accum,stencil))
    {
    XFree(best); // reject configurations we don't want to use
    best=(XVisualInfo *)0;
    }
  }
if(!best) // try our own visual selection algorithm
  {
  int infoCount;
  XVisualInfo *infos=XGetVisualInfo(display,VisualNoMask,
                                    (XVisualInfo *)0,&infoCount);
  if(infos)
    {
    for(int i=0;i<infoCount;++i)
      {
      if(checkVisual(display,infos+i,doubleBuffer,stereo,accum,stencil))
        {
        if(!best) // first match
          {
          best=infos+i;
          }
        else if(infos[i].depth>best->depth) // better depth
          {
          best=infos+i;
          }
        else if((infos[i].depth==best->depth)&&           // prefer TrueColor
                (XVisualInfo_class(infos+i)==TrueColor))  // to DirectColor
          {
          best=infos+i;
          }
        }
      }
    if(best)
      {
      XVisualInfo tmp;
      tmp.screen=best->screen;
      tmp.visualid=best->visualid; // get only this visual
      best=XGetVisualInfo(display,VisualScreenMask|VisualIDMask,
                          &tmp,&infoCount);
      }
    XFree(infos);
    }
  }
if(!best&&!doubleBuffer) // try double-buffer if single is not available
  {
  best=chooseVisual(display,true,stereo,accum,stencil);
  }
return best;
}

static
bool // success
checkFBConfig(Display *display,
              GLXFBConfig fbConfig,
              bool doubleBuffer,
              bool stereo,
              bool accum,
              bool stencil)
{
XVisualInfo *visual=glXGetVisualFromFBConfig(display,fbConfig);
if(!visual)
  {
  return false;
  }
if((visual->screen!=DefaultScreen(display))||
   ((XVisualInfo_class(visual)!=TrueColor)&&
    (XVisualInfo_class(visual)!=DirectColor))||
   (!visual->red_mask)||(!visual->green_mask)||(!visual->blue_mask))
  {
  XFree(visual);
  return false;
  }
XFree(visual);
int value;
if(glXGetFBConfigAttrib(display,fbConfig,GLX_RENDER_TYPE,&value)||
   !(value&GLX_RGBA_BIT))
  {
  return false;
  }
#if 0
if(glXGetFBConfigAttrib(display,fbConfig,GLX_SAMPLE_BUFFERS,&value)||
   !value)
  {
  return false;
  }
#endif
if(glXGetFBConfigAttrib(display,fbConfig,GLX_RED_SIZE,&value)||
   !value)
  {
  return false;
  }
if(glXGetFBConfigAttrib(display,fbConfig,GLX_GREEN_SIZE,&value)||
   !value)
  {
  return false;
  }
if(glXGetFBConfigAttrib(display,fbConfig,GLX_BLUE_SIZE,&value)||
   !value)
  {
  return false;
  }
#if 0 // sometimes value is 0 even if alpha is supported!
  if(glXGetFBConfigAttrib(display,fbConfig,GLX_ALPHA_SIZE,&value)||
     !value)
    {
    return false;
    }
#endif
if(glXGetFBConfigAttrib(display,fbConfig,GLX_DEPTH_SIZE,&value)||
   !value)
  {
  return false;
  }
if(doubleBuffer)
  {
  if(glXGetFBConfigAttrib(display,fbConfig,GLX_DOUBLEBUFFER,&value)||
     !value)
    {
    return false;
    }
  }
else
  {
  if(glXGetFBConfigAttrib(display,fbConfig,GLX_DOUBLEBUFFER,&value)||
     value)
    {
    return false;
    }
  }
if(stereo)
  {
  if(glXGetFBConfigAttrib(display,fbConfig,GLX_STEREO,&value)||
     !value)
    {
    return false;
    }
  }
else
  {
  if(glXGetFBConfigAttrib(display,fbConfig,GLX_STEREO,&value)||
     value)
    {
    return false;
    }
  }
if(accum)
  {
  if(glXGetFBConfigAttrib(display,fbConfig,GLX_ACCUM_RED_SIZE,&value)||
     !value)
    {
    return false;
    }
  if(glXGetFBConfigAttrib(display,fbConfig,GLX_ACCUM_GREEN_SIZE,&value)||
     !value)
    {
    return false;
    }
  if(glXGetFBConfigAttrib(display,fbConfig,GLX_ACCUM_BLUE_SIZE,&value)||
     !value)
    {
    return false;
    }
  if(glXGetFBConfigAttrib(display,fbConfig,GLX_ACCUM_ALPHA_SIZE,&value)||
     !value)
    {
    return false;
    }
  }
if(stencil)
  {
  if(glXGetFBConfigAttrib(display,fbConfig,GLX_STENCIL_SIZE,&value)||
     !value)
    {
    return false;
    }
  }
return true;
}

static
GLXFBConfig
chooseFBConfig(Display *display,
               bool doubleBuffer,
               bool stereo,
               bool accum,
               bool stencil)
{
int attribs[0x30];
int attribCount=0;
attribs[attribCount++]=GLX_X_RENDERABLE;
attribs[attribCount++]=True;
attribs[attribCount++]=GLX_DRAWABLE_TYPE;
attribs[attribCount++]=GLX_WINDOW_BIT;
attribs[attribCount++]=GLX_RENDER_TYPE;
attribs[attribCount++]=GLX_RGBA_BIT;
attribs[attribCount++]=GLX_X_VISUAL_TYPE;
attribs[attribCount++]=GLX_TRUE_COLOR;
attribs[attribCount++]=GLX_RED_SIZE;
attribs[attribCount++]=1;
attribs[attribCount++]=GLX_GREEN_SIZE;
attribs[attribCount++]=1;
attribs[attribCount++]=GLX_BLUE_SIZE;
attribs[attribCount++]=1;
attribs[attribCount++]=GLX_ALPHA_SIZE;
attribs[attribCount++]=1;
attribs[attribCount++]=GLX_DEPTH_SIZE;
attribs[attribCount++]=1;
if(doubleBuffer)
  {
  attribs[attribCount++]=GLX_DOUBLEBUFFER;
  attribs[attribCount++]=True;
  }
if(stereo)
  {
  attribs[attribCount++]=GLX_STEREO;
  attribs[attribCount++]=True;
  }
if(accum)
  {
  attribs[attribCount++]=GLX_ACCUM_RED_SIZE;
  attribs[attribCount++]=1;
  attribs[attribCount++]=GLX_ACCUM_GREEN_SIZE;
  attribs[attribCount++]=1;
  attribs[attribCount++]=GLX_ACCUM_BLUE_SIZE;
  attribs[attribCount++]=1;
  attribs[attribCount++]=GLX_ACCUM_ALPHA_SIZE;
  attribs[attribCount++]=1;
  }
if(stencil)
  {
  attribs[attribCount++]=GLX_STENCIL_SIZE;
  attribs[attribCount++]=1;
  }
// attribs[attribCount++]=GLX_SAMPLE_BUFFERS;
// attribs[attribCount++]=1;
// attribs[attribCount++]=GLX_SAMPLES;
// attribs[attribCount++]=1;
attribs[attribCount++]=None;
GLXFBConfig best=(GLXFBConfig)0;
int configCount;
GLXFBConfig *fbc=glXChooseFBConfig(display,DefaultScreen(display),
                                   attribs,&configCount);
if(fbc)
  {
  int bestSamples=-1;
  for(int i=0;i<configCount;++i)
    {
    if(checkFBConfig(display,fbc[i],doubleBuffer,stereo,accum,stencil))
      {
      int sampleBuffers;
      if(!glXGetFBConfigAttrib(display,fbc[i],GLX_SAMPLE_BUFFERS,
                               &sampleBuffers))
        {
        int samples=0;
        if(!glXGetFBConfigAttrib(display,fbc[i],GLX_SAMPLES,&samples))
          {
          if((!best)||(sampleBuffers&&(samples>bestSamples)))
            {
            best=fbc[i];
            bestSamples=samples;
            }
          }
        }
      }
    }
  XFree(fbc);
  }
if(!best&&!doubleBuffer) // try double-buffer if single is not available
  {
  best=chooseFBConfig(display,true,stereo,accum,stencil);
  }
return best;
}

void
GlView_init(void)
{
if(!mgr.initDone)
  {
  mgr.initDone=true;
  mgr.initCpuCtx=Platform_cpuCtx();
  PtrVector_init(&mgr.views);
  mgr.eventCount=0;
  mgr.stereoMode=GlView_detectStereoMode();
  // Change default value for vsync (see /usr/share/doc/nvidia/README)
  setenv("__GL_SYNC_TO_VBLANK","0",0);
  mgr.display=XOpenDisplay(getenv("DISPLAY") ? "" : ":0");
  if(!mgr.display)
    {
    DROP_DEAD("XOpenDisplay() failure");
    }
  // XSynchronize(mgr.display,True);
  char blankData[1]={0};
  XColor dummyColor;
  Pixmap blankPixmap=XCreateBitmapFromData(mgr.display,
                                           DefaultRootWindow(mgr.display),
                                           blankData,1,1);
  mgr.cursors[GFX_NO_CSR]=XCreatePixmapCursor(mgr.display,
                                              blankPixmap,blankPixmap,
                                              &dummyColor,&dummyColor,0,0);
  XFreePixmap(mgr.display,blankPixmap);
  mgr.cursors[GFX_ARROW_CSR]=XCreateFontCursor(mgr.display,XC_left_ptr);
  mgr.cursors[GFX_TEXT_CSR]=XCreateFontCursor(mgr.display,XC_xterm);
  mgr.cursors[GFX_HAND_CSR]=XCreateFontCursor(mgr.display,XC_hand2);
  mgr.cursors[GFX_CROSS_CSR]=XCreateFontCursor(mgr.display,XC_crosshair);
  mgr.cursors[GFX_WAIT_CSR]=XCreateFontCursor(mgr.display,XC_watch);
  mgr.cursors[GFX_NORTH_CSR]=XCreateFontCursor(mgr.display,XC_top_side);
  mgr.cursors[GFX_SOUTH_CSR]=XCreateFontCursor(mgr.display,XC_bottom_side);
  mgr.cursors[GFX_EAST_CSR]=XCreateFontCursor(mgr.display,XC_right_side);
  mgr.cursors[GFX_WEST_CSR]=XCreateFontCursor(mgr.display,XC_left_side);
  CoreGl_extensionFunctionWithSuffix(&createContextAttribsFnct,
                                     "glXCreateContextAttribs",
                                     extSuffixes);
  bool doubleBuffer=true;
  bool stereo=mgr.stereoMode==GFX_ACTIVE_STEREO ? true : false;
  bool accum=false;
  bool stencil=false;
  if(createContextAttribsFnct)
    {
    mgr.fbConfig=chooseFBConfig(mgr.display,
                                doubleBuffer,stereo,accum,stencil);
    }
  if(mgr.fbConfig)
    {
    mgr.visual=glXGetVisualFromFBConfig(mgr.display,mgr.fbConfig);
    if(!mgr.visual)
      {
      DROP_DEAD("glXGetVisualFromFBConfig() failure");
      }
    }
  if(!mgr.visual)
    {
    mgr.visual=chooseVisual(mgr.display,doubleBuffer,stereo,accum,stencil);
    if(!mgr.visual)
      {
      DROP_DEAD("chooseVisual() failure");
      }
    }
  mgr.colormap=XCreateColormap(mgr.display,DefaultRootWindow(mgr.display),
                               mgr.visual->visual,AllocNone);
  if(!mgr.colormap)
    {
    DROP_DEAD("XCreateColormap() failure");
    }
  X11View_open((GlView *)0,0,0,false);
  GLXContext tmpContext=glXCreateContext(mgr.display,mgr.visual,
                                         (GLXContext)0,True);
  if(!tmpContext)
    {
    DROP_DEAD("glXCreateContext() failure");
    }
  if(!glXMakeCurrent(mgr.display,mgr.commonView->window,tmpContext))
    {
    DROP_DEAD("glXMakeCurrent() failure");
    }
  if(mgr.fbConfig)
    {
    int gl3Attribs[]={GLX_CONTEXT_MAJOR_VERSION_ARB,
                      3,
                      GLX_CONTEXT_MINOR_VERSION_ARB,
                      2,
                      GLX_CONTEXT_PROFILE_MASK_ARB,
                      GLX_CONTEXT_CORE_PROFILE_BIT_ARB,
                      // GLX_CONTEXT_FLAGS_ARB,
                      // GLX_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB,
                      0};
    int *useGl3Attribs=(int *)0;
    int major=0,minor=0;
    if(GlView_detectGlVersion(&major,&minor)&&
       ((major>3)||((major==3)&&(minor>=2))))
      {
      gl3Attribs[1]=major;
      gl3Attribs[3]=minor;
      useGl3Attribs=gl3Attribs;
      }
    mgr.glxContext=createContextAttribsFnct(mgr.display,mgr.fbConfig,
                                            (GLXContext)0,True,useGl3Attribs);
    if(!mgr.glxContext)
      {
      DROP_DEAD("glXCreateContextAttribs() failure");
      }
    }
  else
    {
    mgr.glxContext=tmpContext;
    tmpContext=(GLXContext)0;
    }
  if(tmpContext)
    {
    glXMakeCurrent(mgr.display,(Window)0,(GLXContext)0);
    glXDestroyContext(mgr.display,tmpContext);
    }
  if(!glXMakeCurrent(mgr.display,mgr.commonView->window,mgr.glxContext))
    {
    DROP_DEAD("glXMakeCurrent() failure");
    }
  CoreGl_init();
  if(mgr.stereoMode==GFX_ACTIVE_STEREO)
    {
    int glIsStereo=0;
    glGetIntegerv(GL_STEREO,&glIsStereo);
    if(!glIsStereo)
      {
      DROP_DEAD("OpenGL context does not support active stereo");
      }
    }
  }
}

//----------------------------------------------------------------------------

const char * // specific OpenGL extensions
_CoreGl_querySpecificExtensions(void)
{
const char *extensions=(const char *)0;
Display *dpy=mgr.display ? mgr.display
                         : XOpenDisplay(getenv("DISPLAY") ? "" : ":0");

if(dpy)
  {
  extensions=glXQueryExtensionsString(dpy,DefaultScreen(dpy));
  if(!mgr.display)
    {
    XCloseDisplay(dpy);
    }
  }
return extensions;
}

//----------------------------------------------------------------------------
