//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/math.h"

TRANSPROG_API_VAR uint16_t _Math_halfB[512];
TRANSPROG_API_VAR uint8_t _Math_halfS[512];

TRANSPROG_API_VAR uint32_t _Math_halfM[2048];
TRANSPROG_API_VAR uint16_t _Math_halfO[64];
TRANSPROG_API_VAR uint32_t _Math_halfE[64];

// half-float conversions as described in Jeroen van der Zijp's
// article ``Fast Half Float Conversions'', November 2008

void _Math_init(void)
{
int i;
// populate _Math_halfB and _Math_halfS
for(i=0;i<256;++i)
  {
  int e=i-127;
  if(e<-24) // very small numbers map to zero
    {
    _Math_halfB[i|0x000]=0x0000;
    _Math_halfB[i|0x100]=0x8000;
    _Math_halfS[i|0x000]=24;
    _Math_halfS[i|0x100]=24;
    }
  else if(e<-14) // small numbers map to denorms
    {
    _Math_halfB[i|0x000]=(uint16_t)(0x0400>>(18-e));
    _Math_halfB[i|0x100]=(uint16_t)((0x0400>>(18-e))|0x8000);
    _Math_halfS[i|0x000]=(uint8_t)(-e-1);
    _Math_halfS[i|0x100]=(uint8_t)(-e-1);
    }
  else if(e<=15) // normal numbers just lose precision
    {
    _Math_halfB[i|0x000]=(uint16_t)(((e+15)<<10));
    _Math_halfB[i|0x100]=(uint16_t)(((e+15)<<10)|0x8000);
    _Math_halfS[i|0x000]=13;
    _Math_halfS[i|0x100]=13;
    }
  else if(e<128) // large numbers map to infinity
    {
    _Math_halfB[i|0x000]=0x7C00;
    _Math_halfB[i|0x100]=0xFC00;
    _Math_halfS[i|0x000]=24;
    _Math_halfS[i|0x100]=24;
    }
  else // infinity and nan's stay infinity and nan's
    {
    _Math_halfB[i|0x000]=0x7C00;
    _Math_halfB[i|0x100]=0xFC00;
    _Math_halfS[i|0x000]=13;
    _Math_halfS[i|0x100]=13;
    }
  }
// populate _Math_halfE
_Math_halfE[0]=0;
for(i=1;i<31;++i)
  {
  _Math_halfE[i]=i<<23;
  }
_Math_halfE[31]=0x47800000;
_Math_halfE[32]=0x80000000;
for(i=33;i<63;++i)
  {
  _Math_halfE[i]=0x80000000+((i-32)<<23);
  }
_Math_halfE[63]=0xC7800000;
// populate _Math_halfO
for(i=0;i<64;++i)
  {
  _Math_halfO[i]=1024;
  }
_Math_halfO[32]=0;
_Math_halfO[0]=0;
// populate _Math_halfM
_Math_halfM[0]=0;
for(i=1;i<1024;++i)
  {
  uint32_t m=i<<13;      // zero pad mantissa bits
  uint32_t e=0;          // zero exponent
  while(!(m&0x00800000)) // while not normalized
    {
    e-=0x00800000;       // decrement exponent (1<<23)
    m<<=1;               // shift mantissa
    }
  m&=~0x00800000;        // clear leading 1 bit
  e+=0x38800000;         // adjust bias ((127-14)<<23)
  _Math_halfM[i]=m|e;
  }
for(i=1024;i<2048;++i)
  {
  _Math_halfM[i]=0x38000000+((i-1024)<<13);
  }
}

//-------- Misc --------------------------------------------------------------

#define GAUSSIAN_ELIMINATION_BODY(realType,absFnct,varCount,eq,vars) \
        {                                                            \
        int width=varCount+1;                                        \
        for(int k=0;k<varCount;++k)                                  \
          {                                                          \
          /*-- find row with max column k --*/                       \
          int kMax=k;                                                \
          for(int l=k+1;l<varCount;++l)                              \
            {                                                        \
            if(absFnct(eq[k+l*width])>absFnct(eq[k+kMax*width]))     \
              {                                                      \
              kMax=l;                                                \
              }                                                      \
            }                                                        \
          if(!eq[k+kMax*width])                                      \
            {                                                        \
            return false; /* singular matrix */                      \
            }                                                        \
          /*-- swap rows --*/                                        \
          if(k!=kMax)                                                \
            {                                                        \
            for(int l=0;l<width;++l)                                 \
              {                                                      \
              realType tmp=eq[l+k*width];                            \
              eq[l+k*width]=eq[l+kMax*width];                        \
              eq[l+kMax*width]=tmp;                                  \
              }                                                      \
            }                                                        \
          /*-- cancel variable k on rows below --*/                  \
          for(int i=k+1;i<varCount;++i)                              \
            {                                                        \
            realType factor=eq[k+i*width]/eq[k+k*width];             \
            for(int j=k;j<width;++j)                                 \
              {                                                      \
              eq[j+i*width]-=eq[j+k*width]*factor;                   \
              }                                                      \
            }                                                        \
          }                                                          \
        /*-- compute variables --*/                                  \
        for(int k=varCount;k--;)                                     \
          {                                                          \
          vars[k]=eq[varCount+k*width];                              \
          for(int i=k+1;i<varCount;++i)                              \
            {                                                        \
            vars[k]-=eq[i+k*width]*vars[i];                          \
            }                                                        \
          vars[k]/=eq[k+k*width];                                    \
          }                                                          \
        return true;                                                 \
        }

TRANSPROG_API
bool // success
gaussianElimination32(int varCount,
                      real32_t *inout_equationSystem, // [varCount][varCount+1]
                      real32_t *out_vars)
GAUSSIAN_ELIMINATION_BODY(real32_t,fabsf,
                          varCount,inout_equationSystem,out_vars)

TRANSPROG_API
bool // success
gaussianElimination64(int varCount,
                      real64_t *inout_equationSystem, // [varCount][varCount+1]
                      real64_t *out_vars)
GAUSSIAN_ELIMINATION_BODY(real64_t,fabs,
                          varCount,inout_equationSystem,out_vars)

//----------------------------------------------------------------------------
