//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/path.h"

#include "TransProg/memory.h"
#include "TransProg/random.h"

#define FIND_PATH(path)                             \
        char foundPath[PATH_MAX+1];                 \
        if(Path_lookFor(foundPath,PATH_MAX+1,path)) \
          { path=foundPath; }

PathKind
Path_kind(const char *path)
{
#if defined _WIN32
  DWORD r=GetFileAttributes(path);
  if(r==INVALID_FILE_ATTRIBUTES)
    {
    return PATH_ERROR;
    }
  if(r&FILE_ATTRIBUTE_DIRECTORY)
    {
    return PATH_DIR;
    }
  if(r&FILE_ATTRIBUTE_DEVICE)
    {
    return PATH_SPECIAL;
    }
  return PATH_FILE;
#else
  struct stat st;
  int r;
  RESTART_SYSCALL(r,stat(path,&st));
  if(r==-1)
    {
    return PATH_ERROR;
    }
  if(S_ISREG(st.st_mode))
    {
    return PATH_FILE;
    }
  if(S_ISDIR(st.st_mode))
    {
    return PATH_DIR;
    }
  return PATH_SPECIAL;
#endif
}

int64_t // file size or -1
Path_fileSize(const char *fileName)
{
FIND_PATH(fileName)
#if defined _WIN32
  WIN32_FILE_ATTRIBUTE_DATA info;
  if(GetFileAttributesEx(fileName,GetFileExInfoStandard,&info))
    {
    if(info.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY)
      {
      return -1;
      }
    else
      {
      int64_t sz=info.nFileSizeHigh;
      sz<<=32;
      sz+=info.nFileSizeLow;
      return sz;
      }
    }
  else
    {
    return -1;
    }
#else
  struct stat st;
  int r;
  RESTART_SYSCALL(r,stat(fileName,&st));
  return r==-1 ? -1 : st.st_size;
#endif
}

intptr_t // amount read or -1
Path_readFile(const char *fileName,
              void *out_buffer,
              intptr_t size,
              int64_t offset)
{
FIND_PATH(fileName)
intptr_t amount=-1;
#if defined _WIN32
  SECURITY_ATTRIBUTES secAttr;
  secAttr.nLength=sizeof(secAttr);
  secAttr.bInheritHandle=FALSE;
  secAttr.lpSecurityDescriptor=NULL;
  HANDLE fd=CreateFile(filename,GENERIC_READ,FILE_SHARE_READ|FILE_SHARE_WRITE,
                       &secAttr,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
  if(fd!=INVALID_HANDLE_VALUE)
    {
    LARGE_INTEGER pos;
    LARGE_INTEGER newPos;
    pos.QuadPart=offset;
    if(SetFilePointerEx(fd,pos,&newPos,FILE_BEGIN)&&
       (newPos.QuadPart==pos.QuadPart))
      {
      char *ptr=(char *)out_buffer;
      intptr_t remaining=size;
      while(remaining)
        {
        DWORD count;
        if(ReadFile(fd,out_buffer,(DWORD)remaining,&count,NULL)&&(count>0))
          {
          ptr+=count;
          remaining-=count;
          }
        else
          {
          break;
          }
        }
      amount=size-remaining;
      }
    CloseHandle(fd);
    }
#else
  int fd;
  RESTART_SYSCALL(fd,open(fileName,O_RDONLY));
  if(fd!=-1)
    {
    int64_t pos;
    RESTART_SYSCALL(pos,lseek(fd,offset,SEEK_SET));
    if(offset==pos)
      {
      char *ptr=(char *)out_buffer;
      intptr_t remaining=size;
      while(remaining)
        {
        intptr_t r;
        RESTART_SYSCALL(r,read(fd,ptr,remaining));
        if(r>0)
          {
          ptr+=r;
          remaining-=r;
          }
        else
          {
          break;
          }
        }
      amount=size-remaining;
      }
    int r;
    RESTART_SYSCALL(r,close(fd));
    }
#endif
return amount;
}

intptr_t // amount written or -1
Path_writeFile(const char *fileName,
               const void *buffer,
               intptr_t size,
               bool append)
{
intptr_t amount=-1;
#if defined _WIN32
  SECURITY_ATTRIBUTES secAttr;
  secAttr.nLength=sizeof(secAttr);
  secAttr.bInheritHandle=FALSE;
  secAttr.lpSecurityDescriptor=NULL;
  HANDLE fd=CreateFile(fileName,GENERIC_WRITE,
                       FILE_SHARE_READ|FILE_SHARE_WRITE,&secAttr,
                       (append ? OPEN_ALWAYS : CREATE_ALWAYS),
                       FILE_ATTRIBUTE_NORMAL,NULL);
  if(fd!=INVALID_HANDLE_VALUE)
    {
    if(append)
      {
      LARGE_INTEGER pos;
      pos.QuadPart=0;
      SetFilePointerEx(fd,pos,NULL,FILE_END);
      }
    const char *ptr=(const char *)buffer;
    intptr_t remaining=size;
    while(remaining)
      {
      DWORD count;
      if(WriteFile(fd,ptr,(DWORD)remaining,&count,NULL)&&(count>0))
        {
        ptr+=count;
        remaining-=count;
        }
      else
        {
        break;
        }
      }
    amount=size-remaining;
    CloseHandle(fd);
    }
#else
  int fd;
  int flags=O_WRONLY|O_CREAT|(append ? O_APPEND : O_TRUNC);
  RESTART_SYSCALL(fd,open(fileName,flags,0644));
  if(fd!=-1)
    {
    const char *ptr=(const char *)buffer;
    intptr_t remaining=size;
    while(remaining)
      {
      intptr_t r;
      RESTART_SYSCALL(r,write(fd,ptr,remaining));
      if(r>0)
        {
        ptr+=r;
        remaining-=r;
        }
      else
        {
        break;
        }
      }
    amount=size-remaining;
    int r;
    RESTART_SYSCALL(r,close(fd));
    }
#endif
return amount;
}

FILE * // open file or 0
Path_fopen(const char *fileName,
           const char *mode)
{
if(strchr(mode,'r')) // file must exist to be read
  {
  FIND_PATH(fileName)
  return fopen(fileName,mode);
  }
else
  {
  return fopen(fileName,mode);
  }
}

INLINE
int // new path length
Path_skip(char *path,
          int pathLength,
          char *pos,
          int skip)
{
memmove(pos,pos+skip,pathLength+1-(pos+skip-path));
return pathLength-skip;
}

void
Path_normalize(char *inout_path)
{
#if defined _WIN32
  if(inout_path[0]&&(inout_path[1]==':'))
    {
    inout_path+=2; // preserve drive letter
    }
#endif
int len=0;
char *p=inout_path;
while(*p)
  {
  if((*p=='\\')||(*p=='/'))
    {
    if(!len||(inout_path[len-1]!=Path_dirSep()))
      {
      inout_path[len++]=Path_dirSep();
      }
    }
  else
    {
    inout_path[len++]=*p;
    }
  ++p;
  }
inout_path[len]='\0';
bool isAbs=inout_path[0]==Path_dirSep() ?  true : false;
for(p=inout_path;*p;++p)
  {
  if((p==inout_path)&&(p[0]=='.')&&(p[1]==Path_dirSep())) // starts with ./
    {
    len=Path_skip(inout_path,len,p,2);
    --p;
    }
  else if((p[0]==Path_dirSep())&&(p[1]=='.'))
    {
    if((p[2]==Path_dirSep())||!p[2]) // contains /./ or ends with /.
      {
      len=Path_skip(inout_path,len,p,2);
      --p;
      }
    else if((p[2]=='.')&&                   // contains /../
            ((p[3]==Path_dirSep())||!p[3])) // or ends with /..
      {
      char *prev;
      if(p==inout_path)
        {
        prev=p;
        }
      else
        {
        for(prev=p-1;(prev!=inout_path)&&(*prev!=Path_dirSep());--prev);
        }
      if(*prev==Path_dirSep())
        {
        ++prev;
        }
      if(p[3])
        {
        len=Path_skip(inout_path,len,prev,(int)(p+4-prev));
        }
      else
        {
        len=Path_skip(inout_path,len,prev,(int)(p+3-prev));
        }
      p=prev>inout_path ? prev-2 : prev-1;
      }
    }
  }
if(len&&(inout_path[len-1]==Path_dirSep()))
  {
  inout_path[--len]='\0';
  }
if(!len)
  {
  inout_path[0]=(char)(isAbs ? Path_dirSep() : '.');
  inout_path[1]='\0';
  }
}

bool // success
Path_join(char *out_path,
          int pathSize,
          const char *path1,
          const char *path2)
{
int len1=(int)strlen(path1);
int len2=(int)strlen(path2);
if(pathSize<len1+len2+2)
  {
  return false;
  }
if(len1)
  {
  memcpy(out_path,path1,len1);
  out_path[len1]=Path_dirSep();
  memcpy(out_path+len1+1,path2,len2);
  out_path[len1+len2+1]='\0';
  }
else
  {
  memcpy(out_path,path2,len2);
  out_path[len2]='\0';
  }
Path_normalize(out_path);
return true;
}

const char * // basename in path or 0
Path_findBasename(const char *path)
{
int len=(int)strlen(path);
if(!len)
  {
  return (const char *)0;
  }
const char *begin=path;
const char *end=begin+len-1;
while(*end==Path_dirSep())
  {
  if(end==begin)
    {
    return (const char *)0;
    }
  --end;
  }
const char *limit=end;
while((*limit!=Path_dirSep())&&(limit!=begin))
  {
  --limit;
  }
return *limit==Path_dirSep() ? limit+1 : limit;
}

const char * // extension in path or 0
Path_findExtension(const char *path)
{
int len=(int)strlen(path);
if(!len)
  {
  return (const char *)0;
  }
const char *begin=path;
const char *end=begin+len-1;
while(*end==Path_dirSep())
  {
  if(end==begin)
    {
    return (const char *)0;
    }
  --end;
  }
const char *limit=end;
while((*limit!='.')&&(*limit!=Path_dirSep())&&(limit!=begin))
  {
  --limit;
  }
return *limit=='.' ? limit : (const char *)0;
}

bool // success
Path_currentDir(char *out_currentDir,
                int currentDirSize)
{
#if defined _WIN32
  return GetCurrentDirectory(currentDirSize,out_currentDir) ? true : false;
#else
  return getcwd(out_currentDir,currentDirSize) ? true : false;
#endif
}

bool // success (lookedFor was absolute, relative or relative to executable)
Path_lookFor(char *out_path,
             int pathSize,
             const char *lookedFor)
{
strncpy(out_path,lookedFor,pathSize);
out_path[pathSize-1]='\0';
Path_normalize(out_path);
if(Path_kind(out_path)!=PATH_ERROR)
  {
  return true;
  }
char execPath[PATH_MAX+1];
strncpy(execPath,Platform_executable(),PATH_MAX);
execPath[PATH_MAX]='\0';
const char *base=Path_findBasename(execPath);
if(base)
  {
  execPath[base-execPath]='\0';
  }
Path_join(out_path,pathSize,execPath,lookedFor);
return Path_kind(out_path)!=PATH_ERROR;
}

#if defined _WIN32
  struct _PathDir
  {
  WIN32_FIND_DATA findData;
  HANDLE findHandle;
  bool first;
  bool mustClose;
  };
#endif

PathDir * // directory or 0
Path_openDir(const char *path)
{
FIND_PATH(path)
#if defined _WIN32
  int len=path ? (int)strlen(path) : 0;
  char *pattern=Memory_alloc(len+6,char);
  if(len)
    {
    memcpy(pattern,path,len);
    }
  else
    {
    pattern[len++]='.';
    }
  if(pattern[len-1]!='\\')
    {
    pattern[len++]='\\';
    }
  pattern[len++]='*';
  pattern[len++]='.';
  pattern[len++]='*';
  pattern[len]='\0';
  PathDir *d=Memory_alloc(1,PathDir);
  d->first=true;
  d->mustClose=true;
  d->findHandle=FindFirstFile(pattern,&d->findData);
  Memory_free(pattern);
  if(d->findHandle==INVALID_HANDLE_VALUE)
    {
    d->mustClose=false;
    if(GetLastError()!=ERROR_FILE_NOT_FOUND)
      {
      Memory_free(d);
      return (PathDir *)0;
      }
    }
  return d;
#else
  DIR *dir;
  do
    {
    dir=opendir(path&&path[0] ? path : ".");
    } while(!dir&&(errno==EINTR));
  return (PathDir *)dir;
#endif
}

const char * // next element in directory or 0
Path_readDir(PathDir *dir)
{
#if defined _WIN32
  for(;;)
    {
    const char *result=(const char *)0;
    if(dir->first)
      {
      dir->first=false;
      if(dir->mustClose)
        {
        result=dir->findData.cFileName;
        }
      }
    else
      {
      if(FindNextFile(dir->findHandle,&dir->findData))
        {
        result=dir->findData.cFileName;
        }
      }
    if(!result||(strcmp(result,".")&&strcmp(result,"..")))
      {
      return result;
      }
    }
#else
  for(;;)
    {
    struct dirent *e=readdir((DIR *)dir);
    if(!e)
      {
      return (const char *)0;
      }
    if(strcmp(e->d_name,".")&&strcmp(e->d_name,".."))
      {
      return e->d_name;
      }
    }
#endif
}

void
Path_closeDir(PathDir *dir)
{
#if defined _WIN32
  if(dir->mustClose)
    {
    FindClose(dir->findHandle);
    }
  Memory_free(dir);
#else
  closedir((DIR *)dir);
#endif
}

bool // success
Path_createDir(const char *path)
{
#if defined _WIN32
  return CreateDirectory(path,NULL) ? true : false;
#else
  int r;
  RESTART_SYSCALL(r,mkdir(path,0755));
  return r==-1 ? false : true;
#endif
}

static
bool // success
Path_tmpName(char *out_name,
             int nameSize,
             const char *path,
             const char *pattern)
{
if(pattern)
  {
  if(!strcmp(pattern,".")) return false;
  if(!strcmp(pattern,"..")) return false;
  if(strchr(pattern,Path_dirSep())) return false;
  }
if(!pattern||!pattern[0])
  {
  pattern="tmp_??????";
  }
int pathLen=(int)strlen(path);
while(pathLen&&(path[pathLen-1]==Path_dirSep()))
  {
  --pathLen;
  }
int patternLen=(int)strlen(pattern);
if(nameSize<pathLen+patternLen+2)
  {
  return false;
  }
memcpy(out_name,path,pathLen);
out_name[pathLen++]=Path_dirSep();
const char *pool="ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                 "0123456789";
int cpuCtx=Platform_cpuCtx();
for(int i=0;i<patternLen;++i)
  {
  char c=pattern[i];
  out_name[pathLen++]=(char)((c=='?') ? pool[Random_int(cpuCtx,36)] : c);
  }
out_name[pathLen]='\0';
return true;
}

bool // success
Path_createTmpFile(char *out_fileName,
                   int fileNameSize,
                   const char *path,
                   const char *pattern)
{
for(int attempt=0;attempt<10000;++attempt)
  {
  if(!Path_tmpName(out_fileName,fileNameSize,path,pattern))
    {
    break;
    }
#if defined _WIN32
  SECURITY_ATTRIBUTES secAttr;
  secAttr.nLength=sizeof(secAttr);
  secAttr.bInheritHandle=TRUE;
  secAttr.lpSecurityDescriptor=NULL;
  HANDLE fd=CreateFile(out_fileName,
                       GENERIC_READ|GENERIC_WRITE,
                       FILE_SHARE_READ|FILE_SHARE_WRITE,
                       &secAttr,
                       CREATE_NEW,
                       FILE_ATTRIBUTE_NORMAL,
                       NULL);
  if(fd==INVALID_HANDLE_VALUE)
    {
    if(GetLastError()!=ERROR_ALREADY_EXISTS)
      {
      break;
      }
    }
  else
    {
    CloseHandle(fd);
    return true;
    }
#else
  int fd;
  RESTART_SYSCALL(fd,open(out_fileName,O_CREAT|O_RDWR|O_EXCL,0644));
  if(fd==-1)
    {
    if(errno!=EEXIST)
      {
      break;
      }
    }
  else
    {
    int r;
    RESTART_SYSCALL(r,close(fd));
    return true;
    }
#endif
  }
out_fileName[0]='\0';
return false;
}

bool // success
Path_createTmpDir(char *out_dirName,
                  int dirNameSize,
                  const char *path,
                  const char *pattern)
{
for(int attempt=0;attempt<10000;++attempt)
  {
  if(!Path_tmpName(out_dirName,dirNameSize,path,pattern))
    {
    break;
    }
#if defined _WIN32
  if(CreateDirectory(out_dirName,NULL))
    {
    return true;
    }
  else
    {
    if(GetLastError()!=ERROR_ALREADY_EXISTS)
      {
      break;
      }
    }
#else
  int r;
  RESTART_SYSCALL(r,mkdir(out_dirName,0755));
  if(r==-1)
    {
    if(errno!=EEXIST)
      {
      break;
      }
    }
  else
    {
    return true;
    }
#endif
  }
out_dirName[0]='\0';
return false;
}

bool // success
Path_remove(const char *path)
{
if(Path_kind(path)==PATH_DIR)
  {
  PathDir *dir=Path_openDir(path);
  if(dir)
    {
    char *s=Memory_alloc(PATH_MAX+1,char);
    for(;;)
      {
      const char *entry=Path_readDir(dir);
      if(!entry)
        {
        break;
        }
      Path_join(s,PATH_MAX+1,path,entry);
      Path_remove(s);
      }
    Memory_free(s);
    Path_closeDir(dir);
    }
#if defined _WIN32
  return RemoveDirectory(path) ? true : false;
#else
  int r;
  RESTART_SYSCALL(r,rmdir(path));
  return r==-1 ? false : true;
#endif
  }
else
  {
#if defined _WIN32
  return DeleteFile(path) ? true : false;
#else
  int r;
  RESTART_SYSCALL(r,unlink(path));
  return r==-1 ? false : true;
#endif
  }
}

//----------------------------------------------------------------------------
