//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/platform.h"

#include "TransProg/synchro.h"

TRANSPROG_API_VAR char *_Platform_executable;
TRANSPROG_API_VAR int _Platform_argc;
TRANSPROG_API_VAR char **_Platform_argv;

TRANSPROG_API_VAR struct _CpuMemInfo _Platform_cpuMemInfo;

#if defined THREAD_STORAGE
  TRANSPROG_API_VAR THREAD_STORAGE void *_Platform_cpuData;
#elif defined _WIN32
  TRANSPROG_API_VAR DWORD _Platform_cpuData;
#else
  TRANSPROG_API_VAR pthread_key_t _Platform_cpuData;
#endif

static void (* volatile _Platform_threadFunction)(int,void *);
static void * volatile _Platform_threadData;
static volatile int _Platform_runningThreadCount;
static volatile int _Platform_threadStep; // no problem if it loops
static SystemSynchro _Platform_threadSync;

struct _PlatformThread
{
#if defined _WIN32
  HANDLE systemId;
  HANDLE suspendEvent;
#else
  pthread_t systemId;
#endif
};

static struct _PlatformThread _Platform_threads[ASSUMED_MAX_CPU_COUNT];

#if defined _WIN32
  TRANSPROG_API_VAR struct _PlatformW32Fncts _Platform_w32Fncts;
# define LOAD_W32_FNCT(lib,name)                           \
         do                                                \
           {                                               \
           *((FARPROC *)&_Platform_w32Fncts.name)=         \
             GetProcAddress(_Platform_w32Fncts.lib,#name); \
           } while(0)

  HANDLE
  _Platform_suspendEvent(int cpuCtx)
  {
  return _Platform_threads[cpuCtx].suspendEvent;
  }
#endif

#if defined _WIN32

static bool _hideWindowsConsole=false;

TRANSPROG_API
int
_windowsMainFunction(int (*mainFnct)(int,char **))
{
_hideWindowsConsole=true;
char *cmdLine=GetCommandLine();
if(cmdLine)
  {
  // cf: MSDN Parsing C Command-Line Arguments
  int sz=(int)strlen(cmdLine);
  char *tmp=(char *)malloc((sz+1)*sizeof(char));
  int argc=0;
  bool quote=false;
  bool escape=false;
  bool inArg=false;
  int i,j;
  for(i=0,j=0;i<sz;++i)
    {
    char c=cmdLine[i];
    switch(c)
      {
      case ' ':
      case '\t':
        {
        if(escape) { tmp[j++]='\\'; escape=false; }
        if(quote) { tmp[j++]=c; }
        else if(inArg) { tmp[j++]='\0'; ++argc; inArg=false; }
        break;
        }
      case '\\':
        {
        inArg=true;
        if(escape) { tmp[j++]=c; escape=false; }
        else { escape=true; }
        break;
        }
      case '\"':
        {
        inArg=true;
        if(escape) { tmp[j++]=c; escape=false; }
        else { quote=!quote; }
        break;
        }
      default:
        {
        inArg=true;
        if(escape) { tmp[j++]='\\'; escape=false; }
        tmp[j++]=c;
        }
      }
    }
  if(escape) { tmp[j++]='\\'; }
  if(inArg) { tmp[j++]='\0'; ++argc; }
  char **argv=(char **)malloc((argc+1)*sizeof(char *));
  for(i=0,j=0;i<argc;++i)
    {
    argv[i]=tmp+j;
    while(tmp[j++]);
    }
  argv[argc]=(char *)0;
  int status=(*mainFnct)(argc,argv);
  free(argv);
  free(tmp);
  return status;
  }
else
  {
  char *argv[2]={"unknown",(char *)0};
  return (*mainFnct)(1,argv);
  }
}

#endif

static
int
_Platform_setCpuData(void *cpuData)
{
#if defined THREAD_STORAGE
  _Platform_cpuData=cpuData;
#elif defined _WIN32
  if(!TlsSetValue(_Platform_cpuData,cpuData))
    {
    SYS_DROP_DEAD("TlsSetValue() failure");
    }
#else
  int r=pthread_setspecific(_Platform_cpuData,cpuData);
  if(r)
    {
    SYS_DROP_DEAD2("pthread_setspecific() failure",r);
    }
#endif
return (int)(intptr_t)cpuData;
}

static
void
_Platform_setupThread(void *cpuData)
{
int cpuCtx=_Platform_setCpuData(cpuData);
struct _PlatformThread *th=_Platform_threads+cpuCtx;
#if defined _WIN32
  if(!DuplicateHandle(GetCurrentProcess(),GetCurrentThread(),
                      GetCurrentProcess(),&th->systemId,
                      0,FALSE,DUPLICATE_SAME_ACCESS))
    {
    SYS_DROP_DEAD("DuplicateHandle() failure");
    }
#else
  th->systemId=pthread_self();
#endif
if(_Platform_cpuMemInfo.attachCpu)
  {
  int cpuId=Platform_cpuId(cpuCtx);
#if defined __FreeBSD__
  // FIXME: ugly hack!
  //        ``tid'' is the first member of the opaque thread structure
  //        described in /usr/src/lib/libthr/thread/thr_private.h
  int32_t tid=*((int32_t *)th->systemId);
  cpuset_t cpuset;
  CPU_ZERO(&cpuset);
  CPU_SET(cpuId,&cpuset);
  int r=cpuset_setaffinity(CPU_LEVEL_WHICH,CPU_WHICH_TID,tid,
                           sizeof(cpuset),&cpuset);
  if(r)
    {
    SYS_DROP_DEAD2("cpuset_setaffinity() failure",r);
    }
#elif defined __NetBSD__
  cpuset_t *cset=cpuset_create();
  if(!cset)
    {
    SYS_DROP_DEAD("cpuset_create() failure");
    }
  cpuset_set(cpuId,cset);
  int r=pthread_setaffinity_np(th->systemId,cpuset_size(cset),cset))
  if(r)
    {
    SYS_DROP_DEAD2("pthread_setaffinity_np() failure",r);
    }
  cpuset_destroy(cset);
#elif defined __linux__
  cpu_set_t cpuset;
  CPU_ZERO(&cpuset);
  CPU_SET(cpuId,&cpuset);
  int r=pthread_setaffinity_np(th->systemId,sizeof(cpuset),&cpuset);
  if(r)
    {
    SYS_DROP_DEAD2("pthread_setaffinity_np() failure",r);
    }
#elif defined THREAD_AFFINITY_POLICY
  // FIXME: under MacOsX, affinity does not mean attachment to a specific CPU
  thread_affinity_policy_data_t policy;
  policy.affinity_tag=cpuId;
  if(thread_policy_set(pthread_mach_thread_np(th->systemId),
                       THREAD_AFFINITY_POLICY,
                       (thread_policy_t)&policy,
                       THREAD_AFFINITY_POLICY_COUNT)!=KERN_SUCCESS)
    {
    DROP_DEAD("thread_policy_set() failure");
    }
#elif defined _WIN32
  if(_Platform_w32Fncts.SetThreadGroupAffinity)
    {
    GROUP_AFFINITY aff;
    memset(&aff,0,sizeof(aff));
    aff.Mask=1ULL<<(cpuId&0x0000003F);
    aff.Group=cpuId>>6;
    if(!(*_Platform_w32Fncts.SetThreadGroupAffinity)(th->systemId,
                                                     &aff,NULL))
      {
      SYS_DROP_DEAD("SetThreadGroupAffinity() failure");
      }
    }
  else
    {
    if(!SetThreadAffinityMask(th->systemId,1ULL<<cpuId))
      {
      SYS_DROP_DEAD("SetThreadAffinityMask() failure");
      }
    }
  if(_Platform_w32Fncts.InitializeConditionVariable&&
     _Platform_w32Fncts.SleepConditionVariableCS&&
     _Platform_w32Fncts.WakeAllConditionVariable)
    {
    th->suspendEvent=(HANDLE)0;
    }
  else
    {
    th->suspendEvent=CreateEvent(NULL,FALSE,FALSE,NULL);
    if(!th->suspendEvent)
      {
      SYS_DROP_DEAD("CreateEvent() failure");
      }
    }
#else
  (void)cpuId;
#endif
  }
}

static
#if defined _WIN32
DWORD WINAPI // unused
#else
void * // unused
#endif
_Platform_thread(void *cpuData)
{
_Platform_setupThread(cpuData);
int cpuCtx=Platform_cpuCtx();
int lastStep=0;
for(;;)
  {
  int threadStep=_Platform_threadStep;
  if(lastStep!=threadStep)
    {
    lastStep=threadStep;
    (*_Platform_threadFunction)(cpuCtx,_Platform_threadData);
    Atomic_preDecrementInt(&_Platform_runningThreadCount);
    }
  else if(_Platform_threadFunction)
    {
    Platform_pause(); // nb: pause does not seem to help much
    }
  else
    {
    SystemSynchro_enter(&_Platform_threadSync);
    while(!_Platform_threadFunction)
      {
      SystemSynchro_wait(&_Platform_threadSync);
      }
    SystemSynchro_leave(&_Platform_threadSync);
    }
  }
return 0;
}

void
Platform_init(int argc,
              char **argv)
{
//---- Store command line ----
  {
  _Platform_argc=0;
  _Platform_argv=(char **)malloc((argc+1)*sizeof(char *));
  for(int i=0;i<argc;++i)
    {
#if defined __APPLE__
    // when launching a MacOsX application with the Finder,
    // a process serial number is given as argv[1]
    if(i==1)
      {
      int64_t high,low;
      if(sscanf(argv[i],"-psn_%"F64"d_%"F64"d",&high,&low)==2)
        {
        continue; // just ignore this argument
        }
      }
#endif
    // FIXME: MSC complains about strdup() as well as _strdup()...
    int len=(int)strlen(argv[i]);
    char *arg=(char *)malloc((len+1)*sizeof(char));
    memcpy(arg,argv[i],len+1);
    _Platform_argv[_Platform_argc++]=arg;
    }
  _Platform_argv[_Platform_argc]=(char *)0;
  }
//---- Find current executable ----
  {
  _Platform_executable=(char *)malloc((PATH_MAX+1)*sizeof(char));
#if defined _WIN32
  if(GetModuleFileName(NULL,_Platform_executable,PATH_MAX))
    {
    _Platform_executable[PATH_MAX]='\0';
    }
  else
    {
    _Platform_executable[0]='\0';
    }
#else
  if(argv[0][0]=='/')
    {
    strncpy(_Platform_executable,argv[0],PATH_MAX);
    _Platform_executable[PATH_MAX]='\0';
    }
  else if(getcwd(_Platform_executable,PATH_MAX))
    {
    strncat(_Platform_executable,"/",PATH_MAX);
    _Platform_executable[PATH_MAX]='\0';
    strncat(_Platform_executable,argv[0],PATH_MAX);
    _Platform_executable[PATH_MAX]='\0';
    }
  else
    {
    _Platform_executable[0]='\0';
    }
  if(_Platform_executable[0])
    {
    char *real=(char *)malloc((PATH_MAX+1)*sizeof(char));
    if(realpath(_Platform_executable,real))
      {
      strncpy(_Platform_executable,real,PATH_MAX);
      _Platform_executable[PATH_MAX]='\0';
      }
    free(real);
    }
#endif
  }
#if defined _WIN32
//---- initialize Win32 libraries ----
  {
  SetErrorMode(SEM_FAILCRITICALERRORS|SEM_NOOPENFILEERRORBOX);
  WSADATA wsaData; 
  if(WSAStartup(MAKEWORD(2,2),&wsaData))
    {
    SYS_DROP_DEAD("cannot initialize Winsock");
    }
  _Platform_w32Fncts.mainModule=GetModuleHandle((LPCTSTR)0);
  _Platform_w32Fncts.kernel32=GetModuleHandle("kernel32.dll");
  if(_Platform_w32Fncts.kernel32)
    {
    LOAD_W32_FNCT(kernel32,GetLargePageMinimum);
    LOAD_W32_FNCT(kernel32,GetLogicalProcessorInformation);
    LOAD_W32_FNCT(kernel32,GetLogicalProcessorInformationEx);
    LOAD_W32_FNCT(kernel32,SetThreadGroupAffinity);
    LOAD_W32_FNCT(kernel32,InitializeConditionVariable);
    LOAD_W32_FNCT(kernel32,SleepConditionVariableCS);
    LOAD_W32_FNCT(kernel32,WakeAllConditionVariable);
    }
  }
#endif
//---- Check assumed size relations ----
  {
  if(sizeof(void *)!=sizeof(void (*)(void)))
    {
    DROP_DEAD("pointers to objects and to functions "
              "don't have the same size");
    }
  if(sizeof(void *)!=sizeof(intptr_t))
    {
    DROP_DEAD("pointers and intptr_t "
              "don't have the same size");
    }
  if(sizeof(void *)!=sizeof(size_t))
    {
    DROP_DEAD("pointers and size_t "
              "don't have the same size");
    }
  if(sizeof(int)!=sizeof(int32_t))
    {
    DROP_DEAD("int and int32_t "
              "don't have the same size");
    }
  }
//---- Check little/big endian ----
  {
  bool fail=true;
  uint16_t test=0x1234;
  switch(*((uint8_t *)&test))
    {
    case 0x12:
      if(BYTE_ORDER==BIG_ENDIAN) fail=false;
      break;
    case 0x34:
      if(BYTE_ORDER==LITTLE_ENDIAN) fail=false;
      break;
    }
  if(fail)
    {
    DROP_DEAD("problem in little/big endian detection");
    }
  }
//---- Check real32_t/real64_t representation ----
  {
  uint32_t test[2]={ 0xDEADBEEF , 0xDEADBEEF };
  real32_t test32=-6.25985339871e+18F;
  real64_t test64=-1.1885959257070704e+148;
  if(memcmp(test,&test32,sizeof(test32)))
    {
    DROP_DEAD("problem in real32_t format detection");
    }
  if(memcmp(test,&test64,sizeof(test64)))
    {
    DROP_DEAD("problem in real64_t format detection");
    }
  }
//---- Check real64_t alignment ----
  {
  // nb: real64_t operations are much slower when unaligned
  struct { uint8_t ui8; real64_t r64; } test64;
  int offset=(int)(((uint8_t *)&test64.r64)-&test64.ui8);
  if(offset!=8)
    {
    Platform_msg("warning: real64_t alignment is %d instead of 8\n",
                 offset);
    }
  }
//---- Detect hardware ----
  {
  extern void _Platform_detectCpuMem(void);
  _Platform_detectCpuMem();
  if(Platform_cpuCtxCount()>ASSUMED_MAX_CPU_COUNT)
    {
    char buffer[0x100];
    sprintf(buffer,"ASSUMED_MAX_CPU_COUNT should be at least %d\n",
            Platform_cpuCtxCount());
    DROP_DEAD(buffer);
    }
  if(Platform_numaCtxCount()>ASSUMED_MAX_NUMA_COUNT)
    {
    char buffer[0x100];
    sprintf(buffer,"ASSUMED_MAX_NUMA_COUNT should be at least %d\n",
            Platform_numaCtxCount());
    DROP_DEAD(buffer);
    }
  int sz=Platform_maxCacheLineSize();
  if(sz)
    {
    if(sz%ASSUMED_CACHE_ALIGNMENT)
      {
      Platform_msg("warning: ASSUMED_CACHE_ALIGNMENT "
                   "should be a multiple of %d\n",
                   sz);
      }
    }
  else
    {
    Platform_msg("warning: no cache line size detected\n");
    }
  _Platform_cpuMemInfo.maxCpuDistance=0;
  for(int cpu=0,cpu_end=Platform_cpuCtxCount();cpu<cpu_end;++cpu)
    {
    uint8_t *distances=_Platform_cpuMemInfo.cpuDistances[cpu].distances;
    for(int other=0;other<cpu_end;++other)
      {
      int dist=0,ctx1=cpu,ctx2=other;
      while(ctx1!=ctx2)
        {
        ctx1=Platform_memParent(ctx1);
        ctx2=Platform_memParent(ctx2);
        ++dist;
        }
      if(dist>_Platform_cpuMemInfo.maxCpuDistance)
        {
        _Platform_cpuMemInfo.maxCpuDistance=dist;
        }
      if(dist>=(1<<(8*sizeof(*distances))))
        {
        char buffer[0x100];
        sprintf(buffer,"_Platform_cpuMemInfo.cpuDistances[][] is not"
                       " wide enough for %d\n",
                       dist);
        DROP_DEAD(buffer);
        }
      distances[other]=(uint8_t)dist;
      }
    }
  for(int cpu=0,cpu_end=Platform_cpuCtxCount();cpu<cpu_end;++cpu)
    {
    uint8_t *proximities=_Platform_cpuMemInfo.cpuProximities[cpu].proximities;
    for(int other=0;other<cpu_end;++other)
      {
      int prox=(2<<Platform_maxCpuDistance())-
               (1<<Platform_cpuDistance(cpu,other));
      if(prox>=(1<<(8*sizeof(*proximities))))
        {
        char buffer[0x100];
        sprintf(buffer,"_Platform_cpuMemInfo.cpuProximities[][] is not"
                       " wide enough for %d\n",
                       prox);
        DROP_DEAD(buffer);
        }
      proximities[other]=(uint8_t)prox;
      }
    }
  for(int cpu=0,cpu_end=Platform_cpuCtxCount();cpu<cpu_end;++cpu)
    {
    NarrowCpuCtx *trip=_Platform_cpuMemInfo.cpuRoundTrips[cpu].trip;
    trip[0]=(NarrowCpuCtx)cpu;
    int tripCount=1;
    for(int parent,ctx=cpu;(parent=Platform_memParent(ctx))!=-1;ctx=parent)
      {
      int begin=Platform_memFirstCpu(parent);
      int end=begin+Platform_memCpuCount(parent);
      int skipBegin=Platform_memFirstCpu(ctx);
      int skipEnd=skipBegin+Platform_memCpuCount(ctx);
      if(skipEnd==end)
        {
        skipEnd=begin;
        }
      int c=skipEnd;
      while(c!=skipBegin)
        {
        trip[tripCount++]=(NarrowCpuCtx)c;
        if(++c==end)
          {
          c=begin;
          }
        }
      }
    }
#if !defined THREAD_STORAGE
# if defined _WIN32
    _Platform_cpuData=TlsAlloc();
    if(_Platform_cpuData==TLS_OUT_OF_INDEXES)
      {
      SYS_DROP_DEAD("TlsAlloc() failure");
      }
# else
    int r=pthread_key_create(&_Platform_cpuData,(void (*)(void *))0);
    if(r)
      {
      SYS_DROP_DEAD2("pthread_key_create() failure",r);
      }
# endif
#endif
  int cpuCtx=0;
  _Platform_setCpuData((void *)(intptr_t)cpuCtx);
  }
//---- Initialise various services ----
  {
  extern void _Memory_init(void);
  _Memory_init();
  extern void _Math_init(void);
  _Math_init();
  extern void _Random_init(void);
  _Random_init();
  }
//---- Setup threads ----
  {
  _Platform_threadFunction=(void (*)(int,void *))0;
  _Platform_threadData=(void *)0;
  _Platform_runningThreadCount=0;
  _Platform_threadStep=0;
  SystemSynchro_init(&_Platform_threadSync);
  for(int cpu=0,cpu_end=Platform_cpuCtxCount();cpu<cpu_end;++cpu)
    {
    void *cpuData=(void *)(intptr_t)cpu;
    if(cpu)
      {
#if defined _WIN32
      HANDLE t=CreateThread(NULL,0,&_Platform_thread,cpuData,0,NULL);
      if(!t)
        {
        SYS_DROP_DEAD("CreateThread() failure");
        }
#else
      pthread_t t;
      int r=pthread_create(&t,(pthread_attr_t *)0,&_Platform_thread,cpuData);
      if(r)
        {
        SYS_DROP_DEAD2("pthread_create() failure",r);
        }
#endif
      }
    else
      {
      _Platform_setupThread(cpuData);
      }
    }
  }
}

void
Platform_errorMsg(int errorCode,
                  char *msg,
                  int msgSize)
{
#if defined _WIN32
  int r;
  char *err;
  if(FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_SYSTEM,
                   (LPCVOID)0,errorCode,
                   0,// MAKELANGID(LANG_NEUTRAL,SUBLANG_NEUTRAL),
                   (LPTSTR)&err,msgSize,(va_list *)0))
    {
    for(int i=(int)strlen(err);i--;)
      {
      switch(err[i])
        {
        case '\r':
        case '\n':
          err[i]='\0';
          break;
        default:
          i=0;
        }
      }
    r=snprintf(msg,msgSize,"%s",err);
    LocalFree((HLOCAL)err);
    }
  else
    {
    r=snprintf(msg,msgSize,"???");
    }
  snprintf(msg+r,msgSize-r," (%d)",errorCode);
#else
  snprintf(msg,msgSize,"%s (%d)",strerror(errorCode),errorCode);
#endif
msg[msgSize-1]='\0'; // ensure string termination
}

int
Platform_msg(const char *format,
             ...)
{
static bool firstCall=true;
if(firstCall)
  {
  firstCall=false;
  FILE *logFile=(FILE *)0;
#if defined _WIN32
  if(_hideWindowsConsole)
    {
    const char *logName="\\TransProgMsg.txt"; // FIXME: variable name?
    char path[PATH_MAX+1];
    if(GetCurrentDirectory(PATH_MAX,path))
      {
      strncat(path,logName,PATH_MAX);
      path[PATH_MAX]='\0';
      logFile=freopen(path,"a",stderr);
      }
    if(!logFile&&Platform_environ("USERPROFILE",path,PATH_MAX))
      {
      strncat(path,logName,PATH_MAX);
      path[PATH_MAX]='\0';
      logFile=freopen(path,"a",stderr);
      }
    if(logFile)
      {
      char buffer[PATH_MAX+0x100];
      snprintf(buffer,PATH_MAX+0x100,
               "Saving debug messages to file:\n    %s\n",path);
      buffer[PATH_MAX+0x100-1]='\0';
      MessageBox(NULL,buffer,"Debug Message",MB_OK|MB_ICONWARNING);
      }
    }
#elif defined __APPLE__
  char path[PATH_MAX+1];
  if(getcwd(path,PATH_MAX)&&!strcmp(path,"/")&&
     Platform_environ("HOME",path,PATH_MAX))
    {
    const char *logName="/TransProgMsg.txt"; // FIXME: variable name?
    strncat(path,logName,PATH_MAX);
    path[PATH_MAX]='\0';
    logFile=freopen(path,"a",stderr);
    }
#endif
#if !defined _MSC_VER // FIXME: crashes with cl.exe
  // fprintf() on unbuffered streams uses a huge temporary buffer on stack
  setvbuf(stderr,NULL,_IOLBF,0);
#endif
  if(logFile)
    {
    Platform_msg("---------------------------------------"
                 "---------------------------------------\n");
    Platform_msg("executable: %s\n",Platform_executable());
    for(int i=1,i_end=Platform_argc();i<i_end;++i)
      {
      Platform_msg("argument %d: %s\n",i,Platform_argv(i));
      }
    Platform_msg("---------------------------------------"
                 "---------------------------------------\n");
    }
  }
va_list args;
va_start(args,format);
int r=vfprintf(stderr,format,args);
va_end(args);
fflush(stderr);
return r;
}

void
Platform_dropDead(void)
{
#if defined _WIN32
  ExitProcess(1);
#else
  abort();
#endif
}

bool // success
Platform_environ(const char *varName,
                 char *out_value,
                 int size)
{
#if defined _WIN32
  // FIXME: check required storage
  return GetEnvironmentVariable(varName,out_value,size) ? true : false;
#else
  char *var=getenv(varName);
  if(var)
    {
    strncpy(out_value,var,size);
    out_value[size-1]='\0';
    return true;
    }
  else
    {
    return false;
    }
#endif
}

int // keyword index
Platform_environAsKeyword(const char *varName,
                          const char * const *keywords,
                          int keywordCount,
                          int defaultIndex)
{
char option[0x100];
int index=defaultIndex;
if(Platform_environ(varName,option,0x100)&&strlen(option))
  {
  int i;
  for(i=0;i<keywordCount;++i)
    {
    if(!strcasecmp(option,keywords[i]))
      {
      index=i;
      break;
      }
    }
  if(i==keywordCount)
    {
    Platform_msg("warning: ignoring %s (%s) environment variable\n"
                 "         (should be",
                 varName,option);
    for(i=0;i<keywordCount;++i)
      {
      if(i)
        {
        Platform_msg("%s",i==keywordCount-1 ? " or" : ",");
        }
      Platform_msg(" %s",keywords[i]);
      }
    Platform_msg(")\n");
    }
  }
return index;
}

bool // boolean value extracted from environment variable
Platform_environAsBool(const char *varName,
                       bool defaultValue)
{
char option[0x100];
bool value=defaultValue;
if(Platform_environ(varName,option,0x100)&&strlen(option))
  {
  if(!strcasecmp(option,"yes")||
     !strcasecmp(option,"y")||
     !strcasecmp(option,"true")||
     !strcasecmp(option,"on")||
     !strcasecmp(option,"1"))
    {
    value=true;
    }
  else if(!strcasecmp(option,"no")||
          !strcasecmp(option,"n")||
          !strcasecmp(option,"false")||
          !strcasecmp(option,"off")||
          !strcasecmp(option,"0"))
    {
    value=false;
    }
  else
    {
    Platform_msg("warning: ignoring %s (%s) environment variable\n"
                 "         (should be yes, y, true, on, 1, no, n, "
                 "false, off or 0)\n",
                 varName,option);
    }
  }
return value;
}

int // integer value extracted from environment variable
Platform_environAsInt(const char *varName,
                      int defaultValue)
{
char option[0x100];
int value=defaultValue;
if(Platform_environ(varName,option,0x100)&&strlen(option))
  {
  if(sscanf(option," %d ",&value)!=1)
    {
    Platform_msg("warning: ignoring %s (%s) environment variable\n"
                 "         (should be an integer)\n",
                 varName,option);
    }
  }
return value;
}

int64_t // microseconds since 1970/01/01 UTC
Platform_time(void)
{
#if defined _WIN32
  FILETIME ft;
  GetSystemTimeAsFileTime(&ft);
  int64_t t=(((int64_t)ft.dwHighDateTime)<<32)+(int64_t)ft.dwLowDateTime;
  return (t/10LL)-11644473600000000LL; // 100ns to 1us, 1601 to 1970
#else
  struct timeval tv;
  gettimeofday(&tv,(struct timezone *)0);
  return tv.tv_sec*1000000LL+tv.tv_usec;
#endif
}

void
Platform_sleep(int64_t microseconds)
{
#if defined _WIN32
  Sleep((DWORD)(microseconds/1000LL));
#else
  if(microseconds)
    {
    struct timespec ts;
    ts.tv_sec=microseconds/1000000LL;
    ts.tv_nsec=1000LL*(microseconds%1000000LL);
    int r;
    RESTART_SYSCALL(r,nanosleep(&ts,&ts));
    if(r)
      {
      SYS_DROP_DEAD("nanosleep() failure");
      }
    }
  else
    {
    int r;
    RESTART_SYSCALL(r,sched_yield());
    if(r)
      {
      SYS_DROP_DEAD("sched_yield() failure");
      }
    }
#endif
}

void
Platform_parallel(void (*fnct)(int,void *), 
                  void *data)
{
ASSERT_CURRENT_CPUCTX(0);
if(_Platform_threadFunction&&fnct)
  {
  _Platform_threadFunction=fnct;
  _Platform_threadData=data;
  }
else
  {
  SystemSynchro_enter(&_Platform_threadSync);
  _Platform_threadFunction=fnct;
  _Platform_threadData=data;
  SystemSynchro_broadcast(&_Platform_threadSync);
  SystemSynchro_leave(&_Platform_threadSync);
  }
if(_Platform_threadFunction)
  {
  _Platform_runningThreadCount=Platform_cpuCtxCount();
  ++_Platform_threadStep;
  (*_Platform_threadFunction)(0,_Platform_threadData);
  Atomic_preDecrementInt(&_Platform_runningThreadCount);
  while(_Platform_runningThreadCount)
    {
    Platform_pause(); // nb: pause does not seem to help much
    }
  }
}

//----------------------------------------------------------------------------
