//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/skeleton.h"

#include "TransProg/path.h"

#define HAS_CHANNEL_HINT 0x00008000

#include "skeletonHints.c"

void
Skeleton_init(Skeleton *s)
{
s->nodeCount=0;
s->nodes=(struct _SkelNode *)0;
s->names=(char **)0;
s->channelCount=0;
s->frameTime=0.0;
s->frameCount=0;
s->frames=(real64_t *)0;
}

Skeleton *
Skeleton_new(void)
{
Skeleton *s=Memory_alloc(1,Skeleton);
Skeleton_init(s);
return s;
}

void
Skeleton_destroy(Skeleton *s)
{
if(s->nodes)
  {
  for(int i=0,i_end=s->nodeCount;i<i_end;++i)
    {
    Memory_free(s->names[i]);
    }
  Memory_free(s->names);
  Memory_free(s->nodes);
  }
if(s->frames)
  {
  Memory_free(s->frames);
  }
}

void
Skeleton_delete(Skeleton *s)
{
Skeleton_destroy(s);
Memory_free(s);
}

int // node or -1
Skeleton_findNode(const Skeleton *s,
                  const char *name)
{
for(int i=0,i_end=s->nodeCount;i<i_end;++i)
  {
  if(!strcmp(name,s->names[i]))
    {
    return i;
    }
  }
return -1;
}

INLINE
void
_Skeleton_applyFrameToNode(const struct _SkelNode *sn,
                           const real64_t *frames,
                           SpatialBasis *inout_basis)
{
const real64_t *nodeValues=sn->values;
if(sn->hint&HAS_CHANNEL_HINT)
  {
  const int *channels=sn->channels;
  real64_t values[6];
  for(int i=0;i<6;++i)
    {
    values[i]=nodeValues[i];
    int ch=channels[i];
    if(ch!=-1)
      {
      values[i]+=frames[ch];
      }
    }
  (*sn->applyHint)(values,inout_basis);
  }
else
  {
  (*sn->applyHint)(nodeValues,inout_basis);
  }
}

void
Skeleton_applyFrameToNode(const Skeleton *s,
                          int node,
                          int frame,
                          SpatialBasis *inout_basis)
{
_Skeleton_applyFrameToNode(s->nodes+node,
                           s->frames+frame*s->channelCount,
                           inout_basis);
}

INLINE
void
_Skeleton_applyFrameToNodes(const struct _SkelNode *nodes,
                            const real64_t *frames,
                            int startNode,
                            SpatialBasis *inout_bases)
{
const struct _SkelNode *sn=nodes+startNode;
_Skeleton_applyFrameToNode(sn,frames,inout_bases+startNode);
for(int i=sn->children,i_end=i+sn->childCount;i<i_end;++i)
  {
  SpatialBasis_setMatrix(inout_bases+i,inout_bases[startNode].matrix);
  _Skeleton_applyFrameToNodes(nodes,frames,i,inout_bases);
  }
}

void
Skeleton_applyFrameToNodes(const Skeleton *s,
                           int startNode,
                           int frame,
                           SpatialBasis *inout_bases)
{
_Skeleton_applyFrameToNodes(s->nodes,
                            s->frames+frame*s->channelCount,
                            startNode,inout_bases);
}

INLINE
void
_Skeleton_applyTimeToNode(const struct _SkelNode *sn,
                          const real64_t *prevFrames,
                          const real64_t *nextFrames,
                          real64_t nextPrevRatio,
                          real64_t prevNextRatio,
                          SpatialBasis *inout_basis)
{
const real64_t *nodeValues=sn->values;
if(sn->hint&HAS_CHANNEL_HINT)
  {
  const int *channels=sn->channels;
  real64_t values[6];
  for(int i=0;i<3;++i)
    {
    real64_t value=nodeValues[i];
    int ch=channels[i];
    if(ch!=-1)
      {
      value+=prevFrames[ch]*prevNextRatio+nextFrames[ch]*nextPrevRatio;
      }
    values[i]=value;
    }
  for(int i=3;i<6;++i)
    {
    real64_t value=nodeValues[i];
    int ch=channels[i];
    if(ch!=-1)
      {
      real64_t prevValue=prevFrames[ch];
      real64_t nextValue=nextFrames[ch];
      if(nextValue<prevValue-M_PI) nextValue+=M_2PI;
      else if(nextValue>prevValue+M_PI) prevValue+=M_2PI;
      value+=prevValue*prevNextRatio+nextValue*nextPrevRatio;
      }
    values[i]=value;
    }
  (*sn->applyHint)(values,inout_basis);
  }
else
  {
  (*sn->applyHint)(nodeValues,inout_basis);
  }
}

void
Skeleton_applyTimeToNode(const Skeleton *s,
                         int node,
                         real64_t time,
                         SpatialBasis *inout_basis)
{
int frameCount=s->frameCount;
int channelCount=s->channelCount;
real64_t period=(frameCount-1)*s->frameTime;
real64_t nextPrevRatio;
int prevIndex;
interpolationIndex64(frameCount,fmod(time/period,1.0),
                     &prevIndex,&nextPrevRatio);
const real64_t *prevFrames=s->frames+prevIndex*channelCount;
const real64_t *nextFrames=prevFrames+channelCount;
_Skeleton_applyTimeToNode(s->nodes+node,
                          prevFrames,nextFrames,
                          nextPrevRatio,1.0-nextPrevRatio,
                          inout_basis);
}

INLINE
void
_Skeleton_applyTimeToNodes(const struct _SkelNode *nodes,
                           const real64_t *prevFrames,
                           const real64_t *nextFrames,
                           real64_t nextPrevRatio,
                           real64_t prevNextRatio,
                           int startNode,
                           SpatialBasis *inout_bases)

{
const struct _SkelNode *sn=nodes+startNode;
_Skeleton_applyTimeToNode(sn,prevFrames,nextFrames,
                          nextPrevRatio,prevNextRatio,inout_bases+startNode);
for(int i=sn->children,i_end=i+sn->childCount;i<i_end;++i)
  {
  SpatialBasis_setMatrix(inout_bases+i,inout_bases[startNode].matrix);
  _Skeleton_applyTimeToNodes(nodes,prevFrames,nextFrames,
                             nextPrevRatio,prevNextRatio,i,inout_bases);
  }
}

void
Skeleton_applyTimeToNodes(const Skeleton *s,
                          int startNode,
                          real64_t time,
                          SpatialBasis *inout_bases)
{
int frameCount=s->frameCount;
int channelCount=s->channelCount;
real64_t period=(frameCount-1)*s->frameTime;
real64_t nextPrevRatio;
int prevIndex;
interpolationIndex64(frameCount,fmod(time/period,1.0),
                     &prevIndex,&nextPrevRatio);
const real64_t *prevFrames=s->frames+prevIndex*channelCount;
const real64_t *nextFrames=prevFrames+channelCount;
_Skeleton_applyTimeToNodes(s->nodes,prevFrames,nextFrames,
                           nextPrevRatio,1.0-nextPrevRatio,
                           startNode,inout_bases);
}

bool // success
Skeleton_load(Skeleton *s,
              FILE *input)
{
real64_t dt;
int frameCount;
int channelCount;
int nodeCount;
struct _SkelNode *nodes=(struct _SkelNode *)0;
char **names=(char **)0;
real64_t *frames=(real64_t *)0;
if(fscanf(input," dt %lg ",&dt)!=1)
  {
  Platform_msg("`dt' expected in skeleton\n");
  goto failure;
  }
if(fscanf(input," frames %d ",&frameCount)!=1)
  {
  Platform_msg("`frames' expected in skeleton\n");
  goto failure;
  }
if(fscanf(input," channels %d ",&channelCount)!=1)
  {
  Platform_msg("`channels' expected in skeleton\n");
  goto failure;
  }
if(fscanf(input," nodes %d ",&nodeCount)!=1)
  {
  Platform_msg("`nodes' expected in skeleton\n");
  goto failure;
  }
if(nodeCount)
  {
  nodes=Memory_alloc(nodeCount,struct _SkelNode);
  names=Memory_initAlloc(nodeCount,char *);
  if(frameCount&&channelCount)
    {
    frames=Memory_alloc(frameCount*channelCount,real64_t);
    }
  for(int i=0;i<nodeCount;++i)
    {
    struct _SkelNode *node=nodes+i;
    char name[0x100];
    int parent;
    int childCount;
    int children;
    int transformCount;
    if(fscanf(input," name %s ",name)!=1)
      {
      Platform_msg("`name' expected in skeleton node %d\n",i);
      goto failure;
      }
    names[i]=Memory_allocStr(name);
    if(fscanf(input," parent %d ",&parent)!=1)
      {
      Platform_msg("`parent' expected in skeleton node %d\n",i);
      goto failure;
      }
    node->parent=parent;
    if(fscanf(input," children %d %d ",&childCount,&children)!=2)
      {
      Platform_msg("`children' expected in skeleton node %d\n",i);
      goto failure;
      }
    node->childCount=childCount;
    node->children=children;
    if(fscanf(input," transformations %d ",&transformCount)!=1)
      {
      Platform_msg("`transformations' expected in skeleton node %d\n",i);
      goto failure;
      }
    for(int j=0;j<6;++j)
      {
      node->values[j]=0.0;
      node->channels[j]=-1;
      }
    bool hasChannels=false;
    int lastHint=0;
    int hintCount=0;
    node->hint=0;
    for(int j=0;j<transformCount;++j)
      {
      real64_t value;
      int channel;
      if(fscanf(input," kind %s value %lg channel %d ",
                      name,&value,&channel)!=3)
        {
        Platform_msg("transformation %d expected in skeleton node %d\n",j,i);
        goto failure;
        }
      int kind=0x0F;
      const char *kinds[]={"TX","TY","TZ","RX","RY","RZ"};
      int k;
      for(k=0;k<6;++k) { if(!strcmp(name,kinds[k])) { kind=k; break; } }
      if(kind==0x0F)
        {
        Platform_msg("bad transformation %d in skeleton node %d\n",j,i);
        goto failure;
        }
      node->values[kind]=value;
      node->channels[kind]=channel;
      if(channel!=-1)
        {
        hasChannels=true;
        for(int k=0;k<frameCount;++k)
          {
          if(fscanf(input," %lg ",frames+k*channelCount+channel)!=1)
            {
            Platform_msg("missing transformation %d frames"
                         " in skeleton node %d\n",j,i);
            goto failure;
            }
          }
        }
      if((channel!=-1)||value)
        {
        if(kind<3) // TX, TY or TX
          {
          if((!hintCount)||lastHint) // last hint was not T
            {
            lastHint=0;
            ++hintCount;
            }
          }
        else
          {
          lastHint=kind-2; // RX, RY or RZ
          node->hint=node->hint|(lastHint<<(hintCount<<1));
          ++hintCount;
          }
        }
      }
    node->hint=node->hint|(hintCount<<8);
    if(hasChannels)
      {
      node->hint|=HAS_CHANNEL_HINT;
      }
    node->applyHint=Skeleton_chooseApplyHint(node->hint);
    }
  }
Skeleton_destroy(s);
Skeleton_init(s);
s->nodeCount=nodeCount;
s->nodes=nodes;
s->names=names;
s->channelCount=channelCount;
s->frameTime=dt;
s->frameCount=frameCount;
s->frames=frames;
return true;
failure:
if(nodes)
  {
  for(int i=0;i<nodeCount;++i)
    {
    if(names[i])
      {
      Memory_free(names[i]);
      }
    }
  Memory_free(names);
  Memory_free(nodes);
  }
if(frames)
  {
  Memory_free(frames);
  }
return false;
}

bool // success
Skeleton_loadFile(Skeleton *s,
                  const char *fileName)
{
bool result;
FILE *input=Path_fopen(fileName,"r");
if(input)
  {
  result=Skeleton_load(s,input);
  fclose(input);
  }
else
  {
  Platform_msg("cannot open file `%s'\n",fileName);
  result=false;
  }
return result;
}

//----------------------------------------------------------------------------
