//----------------------------------------------------------------------------
// Copyright (c) 2008-2016 Fabrice HARROUET (ENIB)
//
// This file is part of TransProg.
//   http://www.enib.fr/~harrouet/transprog.html
//
// TransProg is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// TransProg is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with TransProg.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "TransProg/socket.h"

bool // success
Socket_parseIpv4Address(Ipv4Address *out_address,
                        const char *txt)
{
Ipv4Address b3,b2,b1,b0;
if((sscanf(txt,"%u.%u.%u.%u",&b3,&b2,&b1,&b0)!=4)||
   (b3>0x000000FF)||(b2>0x000000FF)||(b1>0x000000FF)||(b0>0x000000FF))
  {
  return false;
  }
*out_address=((b3<<24)|(b2<<16)|(b1<<8)|(b0<<0));
return true;
}

void
Socket_formatIpv4Address(char *out_txt,
                         Ipv4Address address)
{
sprintf(out_txt,"%u.%u.%u.%u",
               (address>>24)&0x000000FF,
               (address>>16)&0x000000FF,
               (address>>8)&0x000000FF,
               (address>>0)&0x000000FF);
}

bool // success
Socket_hostName(char *out_hostName,
                int capacity)
{
int r;
RESTART_SYSCALL(r,gethostname(out_hostName,capacity));
if(r==SOCKET_ERROR)
  {
  return false;
  }
return true;
}

Ipv4Address // IPV4 address of host name or 0
Socket_ipv4Address(const char *hostName)
{
struct hostent *host=gethostbyname(hostName);
if(host)
  {
  return ntohl(*((Ipv4Address *)(host->h_addr)));
  }
Ipv4Address addr=0;
Socket_parseIpv4Address(&addr,hostName); // try dotted-decimal notation
return addr;
}

SocketHandle // socket or INVALID_SOCKET
Socket_open(int domain,
            int type,
            int protocol)
{
SocketHandle s;
RESTART_SYSCALL(s,socket(domain,type,protocol));
return s;
}

bool // success
Socket_close(SocketHandle s)
{
int r;
RESTART_SYSCALL(r,closesocket(s));
return r==SOCKET_ERROR ? false : true;
}

bool // success
Socket_shutdown(SocketHandle s,
                int how)
{
int r;
RESTART_SYSCALL(r,shutdown(s,how));
return r==SOCKET_ERROR ? false : true;
}

bool // success
Socket_setReuseAddrOption(SocketHandle s)
{
#if defined _WIN32
  BOOL option=TRUE;
#else
  int option=1;
#endif
int r;
RESTART_SYSCALL(r,setsockopt(s,SOL_SOCKET,SO_REUSEADDR,
                             (const char *)&option,sizeof(option)));
return r==SOCKET_ERROR ? false : true;
}

bool // success
Socket_setBroadcastOption(SocketHandle s)
{
#if defined _WIN32
  BOOL option=TRUE;
#else
  int option=1;
#endif
int r;
RESTART_SYSCALL(r,setsockopt(s,SOL_SOCKET,SO_BROADCAST,
                             (const char *)&option,sizeof(option)));
return r==SOCKET_ERROR ? false : true;
}

bool // success
Socket_setRcvBufOption(SocketHandle s,
                       int rcvBuf)
{
int r;
RESTART_SYSCALL(r,setsockopt(s,SOL_SOCKET,SO_RCVBUF,
                             (const char *)&rcvBuf,sizeof(rcvBuf)));
return r==SOCKET_ERROR ? false : true;
}

bool // success
Socket_bindIpv4(SocketHandle s,
                Ipv4Address address,
                uint16_t port)
{
struct sockaddr_in addr;
socklen_t addrLen=(socklen_t)sizeof(addr);
memset(&addr,0,addrLen);
addr.sin_family=AF_INET;
addr.sin_port=htons(port);
addr.sin_addr.s_addr=htonl(address);
int r;
RESTART_SYSCALL(r,bind(s,(const struct sockaddr *)&addr,addrLen));
return r==SOCKET_ERROR ? false : true;
}

bool // success
Socket_localIpv4(SocketHandle s,
                 Ipv4Address *out_address,
                 uint16_t *out_port)
{
struct sockaddr_in addr;
socklen_t addrLen=(socklen_t)sizeof(addr);
int r;
RESTART_SYSCALL(r,getsockname(s,(struct sockaddr *)&addr,&addrLen));
if(r==SOCKET_ERROR)
  {
  return false;
  }
*out_address=ntohl(addr.sin_addr.s_addr);
*out_port=ntohs(addr.sin_port);
return true;
}

bool // success
Socket_distantIpv4(SocketHandle s,
                   Ipv4Address *out_address,
                   uint16_t *out_port)
{
struct sockaddr_in addr;
socklen_t addrLen=(socklen_t)sizeof(addr);
int r;
RESTART_SYSCALL(r,getpeername(s,(struct sockaddr *)&addr,&addrLen));
if(r==SOCKET_ERROR)
  {
  return false;
  }
*out_address=ntohl(addr.sin_addr.s_addr);
*out_port=ntohs(addr.sin_port);
return true;
}

bool // success
Socket_listen(SocketHandle s,
              int backlog)
{
int r;
RESTART_SYSCALL(r,listen(s,backlog));
return r==SOCKET_ERROR ? false : true;
}

SocketHandle // dialog socket or INVALID_SOCKET
Socket_acceptIpv4(SocketHandle listenSocket,
                  Ipv4Address *out_address,
                  uint16_t *out_port)
{
struct sockaddr_in addr;
socklen_t addrLen=(socklen_t)sizeof(addr);
SocketHandle s;
RESTART_SYSCALL(s,accept(listenSocket,(struct sockaddr *)&addr,&addrLen));
if(s==INVALID_SOCKET)
  {
  return INVALID_SOCKET;
  }
if(out_address)
  {
  *out_address=ntohl(addr.sin_addr.s_addr);
  }
if(out_port)
  {
  *out_port=ntohs(addr.sin_port);
  }
return s;
}

bool // success
Socket_connectIpv4(SocketHandle s,
                   Ipv4Address address,
                   uint16_t port)
{
struct sockaddr_in addr;
socklen_t addrLen=(socklen_t)sizeof(addr);
memset(&addr,0,addrLen);
addr.sin_family=AF_INET;
addr.sin_port=htons(port);
addr.sin_addr.s_addr=htonl(address);
int r;
RESTART_SYSCALL(r,connect(s,(const struct sockaddr *)&addr,addrLen));
return r==SOCKET_ERROR ? false : true;
}

int // sent bytes or SOCKET_ERROR
Socket_send(SocketHandle s,
            const void *data,
            int size,
            int flags)
{
int r;
RESTART_SYSCALL(r,send(s,(const char *)data,size,flags));
return r;
}

int // sent bytes or SOCKET_ERROR
Socket_sendToIpv4(SocketHandle s,
                  const void *data,
                  int size,
                  int flags,
                  Ipv4Address address,
                  uint16_t port)
{
struct sockaddr_in addr;
socklen_t addrLen=(socklen_t)sizeof(addr);
memset(&addr,0,addrLen);
addr.sin_family=AF_INET;
addr.sin_port=htons(port);
addr.sin_addr.s_addr=htonl(address);
int r;
RESTART_SYSCALL(r,sendto(s,(const char *)data,size,flags,
                         (const struct sockaddr *)&addr,addrLen));
return r;
}

int // received bytes or 0 (EOF) or SOCKET_ERROR
Socket_recv(SocketHandle s,
            void *buffer,
            int capacity,
            int flags)
{
int r;
RESTART_SYSCALL(r,recv(s,(char *)buffer,capacity,flags));
return r;
}

int // received bytes or 0 (EOF) or SOCKET_ERROR
Socket_recvFromIpv4(SocketHandle s,
                    void *buffer,
                    int capacity,
                    int flags,
                    Ipv4Address *out_address,
                    uint16_t *out_port)
{
struct sockaddr_in addr;
socklen_t addrLen=(socklen_t)sizeof(addr);
int r;
RESTART_SYSCALL(r,recvfrom(s,(char *)buffer,capacity,flags,
                           (struct sockaddr *)&addr,&addrLen));
if(r!=SOCKET_ERROR)
  {
  if(out_address)
    {
    *out_address=ntohl(addr.sin_addr.s_addr);
    }
  if(out_port)
    {
    *out_port=ntohs(addr.sin_port);
    }
  }
return r;
}

int // number of socket in ready-state or SOCKET_ERROR
Socket_select(SocketHandle *inout_rdSet,
              SocketHandle *inout_wrSet,
              int64_t usecTimeout)
{
fd_set rdSet,wrSet;
SocketHandle maxHandle=-1;
int rdCount=0;
int wrCount=0;
if(inout_rdSet)
  {
  FD_ZERO(&rdSet);
  for(;;)
    {
    SocketHandle s=inout_rdSet[rdCount];
    if(s==INVALID_SOCKET) break;
    ++rdCount;
    FD_SET(s,&rdSet);
    if(s>=maxHandle) maxHandle=s;
    }
  }
if(inout_wrSet)
  {
  FD_ZERO(&wrSet);
  for(;;)
    {
    SocketHandle s=inout_wrSet[wrCount];
    if(s==INVALID_SOCKET) break;
    ++wrCount;
    FD_SET(s,&wrSet);
    if(s>=maxHandle) maxHandle=s;
    }
  }
struct timeval tv;
if(usecTimeout>=0)
  {
#if defined _WIN32
  tv.tv_sec=(long)(usecTimeout/1000000);
  tv.tv_usec=(long)(usecTimeout%1000000);
#else
  tv.tv_sec=(time_t)(usecTimeout/1000000);
  tv.tv_usec=(int)(usecTimeout%1000000);
#endif
  }
int r;
RESTART_SYSCALL(r,select((int)maxHandle+1,
                         rdCount ? &rdSet : (fd_set *)0,
                         wrCount ? &wrSet : (fd_set *)0,
                         (fd_set *)0,
                         usecTimeout>=0 ? &tv : (struct timeval *)0));
if(r!=SOCKET_ERROR)
  {
  for(int i=rdCount;i--;)
    {
    SocketHandle s=inout_rdSet[i];
    if(!FD_ISSET(s,&rdSet))
      {
      inout_rdSet[i]=inout_rdSet[--rdCount];
      inout_rdSet[rdCount]=INVALID_SOCKET;
      }
    }
  for(int i=wrCount;i--;)
    {
    SocketHandle s=inout_wrSet[i];
    if(!FD_ISSET(s,&wrSet))
      {
      inout_wrSet[i]=inout_wrSet[--wrCount];
      inout_wrSet[wrCount]=INVALID_SOCKET;
      }
    }
  }
return r;
}

//----------------------------------------------------------------------------
